//--------------------------------------------------------------------------
//  Checkers Gameboard (British Draughts)
//  Copyright (C) 2005-2009 Kambiz R. Khojasteh
//--------------------------------------------------------------------------

{$B-} // Complete Boolean Evaluation Off
{$R-} // Range Checking Off
{$Q-} // Overflow Checking Off
{$V-} // Var String Checking Off
{$T-} // Typed @ operator Off
{$X+} // Extended Syntax On
{$P+} // Open String Params On
{$O+} // Optimization On
{$H+} // Use long strings by default

unit CheckersBoard;

interface

uses
  Windows, Messages, Classes, Graphics, Controls, SysUtils, Gameboard;

const
  EMPTY      = $00;
  KING       = $04;
  BLACK      = DARK_PLAYER;
  WHITE      = LIGHT_PLAYER;
  BLACK_KING = BLACK or KING;
  WHITE_KING = WHITE or KING;

type

  TPosition = (   P01,  P02,  P03,  P04,
               P05,  P06,  P07,  P08,
                  P09,  P10,  P11,  P12,
               P13,  P14,  P15,  P16,
                  P17,  P18,  P19,  P20,
               P21,  P22,  P23,  P24,
                  P25,  P26,  P27,  P28,
               P29,  P30,  P31,  P32);

  TBoard = record // MSB = P01; LSB = P32
    Blacks: DWORD;
    Whites: DWORD;
    Kings: DWORD;
  end;

  PMovement = ^TMovement;
  TMovement = record
    Count: Byte;
    Positions: array[0..9] of TPosition;
    Attacked: DWORD;
  end;

  TMoveList = class(TObject)
  private
    fList: PMovement;
    fCount: Integer;
    fCapacity: Integer;
    procedure SetCapacity(Value: Integer);
    function GetMovePtr(Index: Integer): PMovement; inline;
    function GetMoves(Index: Integer): TMovement;
    procedure Grow; inline;
  protected
    procedure Clear; inline;
    procedure Add(const Move: TMovement); overload; inline;
    procedure Add(SrcPos, DstPos: TPosition); overload; inline;
    procedure Filter(const Move: TMovement); overload;
    procedure Filter(SrcPos, DstPos: TPosition); overload;
    procedure Filter(Pos: TPosition); overload;
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    property Capacity: Integer read fCapacity write SetCapacity;
    property MovePtr[Index: Integer]: PMovement read GetMovePtr;
    property FirstMove: PMovement read fList;
  public
    destructor Destroy; override;
    function IndexOf(const Move: TMovement): Integer;
    function MoveExists(const Move: TMovement): Boolean; overload;
    function MoveExists(SrcPos, DstPos: TPosition): Boolean; overload;
    function CompleteMove(var Move: TMovement): Boolean;
    function NextMatchedMove(const Move: TMovement; StartIndex: Integer = -1): Integer;
    property Moves[Index: Integer]: TMovement read GetMoves; default;
    property Count: Integer read fCount;
  end;

  TPositionDragOverEvent = procedure(Sender: TObject; Source, Target: TPosition; var Accept: Boolean) of object;
  TPositionDragDropEvent = procedure(Sender: TObject; Source, Target: TPosition) of object;
  TPositionInfoTipEvent = procedure(Sender: TObject; Pos: TPosition; var InfoTip: String) of object;
  TPositionActionEvent = procedure(Sender: TObject; Pos: TPosition; var Handled: Boolean) of object;
  TPositionEvent = procedure(Sender: TObject; Pos: TPosition) of object;
  TSimpleMoveEvent = procedure(Sender: TObject; SrcPos, DstPos: TPosition) of object;
  TAttackMoveEvent = procedure(Sender: TObject; SrcPos, DstPos, JmpPos: TPosition) of object;

  TCheckersBoard = class(TGameboard)
  private
    fBoard: TBoard;
    fLegalMoves: TMoveList;
    fMoving: Boolean;
    fKey: TPositionKey;
    fPrepared: Boolean;
    fLastPosition: Integer;
    fActivePosition: Integer;
    fOnPromote: TPositionEvent;
    fOnSimpleMove: TSimpleMoveEvent;
    fOnAttackMove: TAttackMoveEvent;
    procedure SetBoard(const Value: TBoard);
    function GetPieces(Pos: TPosition): TPiece;
    procedure SetPieces(Pos: TPosition; Value: TPiece);
    function GetLegalMoves: TMoveList;
  protected
    procedure DoTurn; override;
    procedure DoChange; override;
    function GetDistinctPieceCount: Byte; override;
    function GetLegalMovesCount: Integer; override;
    function GetBoardAsText: String; override;
    procedure SetBoardAsText(const Value: String); override;
  protected
    function CalculateKey: TPositionKey;
    function GetKey: TPositionKey; override;
    function GetScore: Integer; override;
    function GetAvgMoveCount: Cardinal; override;
    class function GetPositionTable: TPositionTable; override;
    procedure InitializeSearch; override;
    procedure Transpose(Gameboard: TGameboard; TargetMove: Integer); override;
  protected
    procedure DoPromote(Pos: TPosition); virtual;
    procedure DoSimpleMove(SrcPos, DstPos: TPosition); virtual;
    procedure DoAttackMove(SrcPos, DstPos, JmpPos: TPosition); virtual;
    procedure PrepareLegalMoves;
    function MovePiece(SrcPos, DstPos: TPosition): Boolean; virtual;
    property Prepared: Boolean read fPrepared;
    property Moving: Boolean read fMoving;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure SaveToStream(Stream: TStream); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure GetTextPresentation(out Board, Moves: String); override;
    procedure StartNewGame; override;
    function ApplyLegalMove(Index: Integer): Boolean; override;
    function ApplyMove(Src, Dst: TPosition): Boolean; overload;
    function ApplyMove(Move: TMovement): Boolean; overload;
    property LegalMoves: TMoveList read GetLegalMoves;
    property Board: TBoard read fBoard write SetBoard;
    property Pieces[Pos: TPosition]: TPiece read GetPieces write SetPieces; default;
    property OnPromote: TPositionEvent read fOnPromote write fOnPromote;
    property OnSimpleMove: TSimpleMoveEvent read fOnSimpleMove write fOnSimpleMove;
    property OnAttackMove: TAttackMoveEvent read fOnAttackMove write fOnAttackMove;
  end;

  TCheckers = class(TGameboardControl)
  private
    fPiecesBitmap: TBitmap;
    fTwoDigitsPositions: Boolean;
    fOnPositionClick: TPositionActionEvent;
    fOnPositionDragOver: TPositionDragOverEvent;
    fOnPositionDragStart: TNotifyEvent;
    fOnPositionDragEnd: TNotifyEvent;
    fOnPositionDragDrop: TPositionDragDropEvent;
    fOnPositionInfoTip: TPositionInfoTipEvent;
    fOnSimpleMove: TSimpleMoveEvent;
    fOnAttackMove: TAttackMoveEvent;
    fOnPromote: TPositionEvent;
    DefaultPieceSize: Integer;
    function GetBoard: TBoard;
    procedure SetBoard(const Value: TBoard);
    function GetPieces(Pos: TPosition): TPiece;
    procedure SetPieces(Pos: TPosition; Value: TPiece);
    procedure SetPiecesBitmap(Value: TBitmap);
    function GetPosRect(Pos: TPosition): TRect;
    function GetSelectedMove: TMovement;
    function GetLegalMoves: TMoveList;
    function GetCheckersBoard: TCheckersBoard;
    function GetCoordinatesDisplay: TSquareLabel;
    procedure SetCoordinatesDisplay(Value: TSquareLabel);
    function GetCoordinatesOffset: Integer;
    procedure SetCoordinatesOffset(Value: Integer);
    procedure PiecesBitmapChange(Sender: TObject);
    procedure PiecePromote(Sender: TObject; Pos: TPosition);
    procedure PieceSimpleMove(Sender: TObject; SrcPos, DstPos: TPosition);
    procedure PieceAttackMove(Sender: TObject; SrcPos, DstPos, JmpPos: TPosition);
  protected
    procedure DoSquareResize; override;
    procedure DoSquareClick(Square: TSquare; var Handled: Boolean); override;
    procedure DoPositionClick(Pos: TPosition; var Handled: Boolean); virtual;
    procedure DoSquareInfoTip(Square: TSquare; var InfoTip: String); override;
    procedure DoPositionInfoTip(Pos: TPosition; var InfoTip: String); virtual;
    procedure DoSquareDragOver(Source, Target: TSquare; var Accept: Boolean); override;
    procedure DoPositionDragOver(Source, Target: TPosition; var Accept: Boolean); virtual;
    procedure DoSquareDragDrop(Source, Target: TSquare); override;
    procedure DoPositionDragDrop(Source, Target: TPosition); virtual;
    procedure DoSquareDragStart; override;
    procedure DoPositionDragStart; virtual;
    procedure DoSquareDragEnd; override;
    procedure DoPositionDragEnd; virtual;
    procedure DoSimpleMove(SrcPos, DstPos: TPosition); virtual;
    procedure DoAttackMove(SrcPos, DstPos, JmpPos: TPosition); virtual;
    procedure DoPromote(Pos: TPosition); virtual;
    function GetGameboardClass: TGameboardClass; override;
    procedure SetImageIndexOfSquare(Square: TSquare; ImageIndex: Integer); override;
    function GetImageIndexOfSquare(Square: TSquare): Integer; override;
    function GetSquareLabel(Square: TSquare): String; override;
    function IsPathComplete(const Path: array of TSquare): Boolean; override;
    function CanPlayPath(const Path: array of TSquare): Boolean; override;
    function PlayPath(const Path: array of TSquare): Boolean; override;
    function CompletePath(var Path: TMovementPath): Boolean; override;
    function IsLegalSquare(Square: TSquare): Boolean; override;
    function CanSelectSquare(Square: TSquare): Boolean; override;
    function UpdatePiecesImages: Boolean; virtual;
    function PosToSquare(Pos: TPosition): TSquare;
    function SquareToPos(Square: TSquare; out Pos: TPosition): Boolean;
    function MoveToPath(const Move: TMovement): TMovementPath;
    function PathToMove(const Path: array of TSquare; out Move: TMovement): Boolean;
    property CheckersBoard: TCheckersBoard read GetCheckersBoard;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function SelectMove(const Move: TMovement): Boolean; overload;
    function SelectMove(Index: Integer): Boolean; overload;
    function SelectPosition(Pos: TPosition): Boolean;
    function CanSelectPosition(Pos: TPosition): Boolean;
    function IsPositionSelected(Pos: TPosition): Boolean;
    function IsLegalMove(const Move: TMovement): Boolean;
    function CanPlayMove(const Move: TMovement): Boolean;
    function Play(const Move: TMovement): Boolean;
    function MoveOf(Player: TPlayer; TurnNo: Integer): String; override;
    function ClientToPos(X, Y: Integer; out Pos: TPosition): Boolean;
    property PosRect[Pos: TPosition]: TRect read GetPosRect;
    property SelectedMove: TMovement read GetSelectedMove;
    property LegalMoves: TMoveList read GetLegalMoves;
    property Board: TBoard read GetBoard write SetBoard;
    property Pieces[Pos: TPosition]: TPiece read GetPieces write SetPieces;
  published
    property Align;
    property Anchors;
    property AnimateDelay;
    property AnimateMoves;
    property AnimateStep;
    property AutoDragDrop;
    property AutoSize;
    property BoardAngle;
    property BoardAsText;
    property BoardBorderColor;
    property BoardMarginBottom;
    property BoardMarginColor;
    property BoardMarginLeft;
    property BoardMarginRight;
    property BoardMarginTop;
    property BoardMode;
    property Center;
    property Color;
    property ComputerPlayer default WHITE;
    property Constraints;
    property CoordinatesDisplay: TSquareLabel read GetCoordinatesDisplay write SetCoordinatesDisplay default slNone;
    property CoordinatesOffset: Integer read GetCoordinatesOffset write SetCoordinatesOffset default 4;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FixedSquareSize;
    property Font;
    property Hint;
    property InvertSquares;
    property MarkerOffset;
    property MarkerSize;
    property MarkSelectableSquares;
    property MaxMemory;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property Paused;
    property PiecesBitmap: TBitmap read fPiecesBitmap write SetPiecesBitmap;
    property RecordFullMovesOnly;
    property SearchDepth;
    property SelectionOptions;
    property ShowHint;
    property SquareBorderColor;
    property SquareDarkColor default $00008000;
    property SquareDarkMarkColor default clLime;
    property SquareDarkMarkPattern;
    property SquareDarkSelectColor default clLime;
    property SquareDarkSelectPattern;
    property SquareDarkPattern;
    property SquareLightColor default $00AAFFFF;
    property SquareLightMarkColor default clRed;
    property SquareLightMarkPattern;
    property SquareLightSelectColor default clRed;
    property SquareLightSelectPattern;
    property SquareLightPattern;
    property SquareSize;
    property TabOrder;
    property TabStop;
    property ThinkCursor;
    property ThinkPriority;
    property TwoDigitsPositions: Boolean read fTwoDigitsPositions write fTwoDigitsPositions default True;
    property Visible;
    property OnAnimateBegin;
    property OnAnimateEnd;
    property OnAttackMove: TAttackMoveEvent read fOnAttackMove write fOnAttackMove;
    property OnBoardChange;
    property OnBoardModeChange;
    property OnCanResize;
    property OnClick;
    property OnComputerError;
    property OnComputerMove;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGameEnded;
    property OnHistoryChange;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMoveBegin;
    property OnMoveEnd;
    property OnPauseStateChange;
    property OnPlayingStateChange;
    property OnPositionClick: TPositionActionEvent read fOnPositionClick write fOnPositionClick;
    property OnPositionDragDrop: TPositionDragDropEvent read fOnPositionDragDrop write fOnPositionDragDrop;
    property OnPositionDragEnd: TNotifyEvent read fOnPositionDragEnd write fOnPositionDragEnd;
    property OnPositionDragOver: TPositionDragOverEvent read fOnPositionDragOver write fOnPositionDragOver;
    property OnPositionDragStart: TNotifyEvent read fOnPositionDragStart write fOnPositionDragStart;
    property OnPositionInfoTip: TPositionInfoTipEvent read fOnPositionInfoTip write fOnPositionInfoTip;
    property OnPromote: TPositionEvent read fOnPromote write fOnPromote;
    property OnResize;
    property OnSelectionChange;
    property OnSimpleMove: TSimpleMoveEvent read fOnSimpleMove write fOnSimpleMove;
    property OnSquareResize;
    property OnStartDock;
    property OnStartDrag;
    property OnThinkBegin;
    property OnThinkEnd;
    property OnTurnChange;
  end;

function PosToStr(Pos: TPosition; TwoDigits: Boolean = True): String;
function StrToPos(const S: String; out Pos: TPosition): Boolean;

function MoveToStr(const Move: TMovement; TwoDigits: Boolean = True): String;
function StrToMove(const S: String; out Move: TMovement): Boolean;

procedure PosToCoordinate(Pos: TPosition; out Row, Col: Integer);
function CoordinateToPos(Row, Col: Integer; out Pos: TPosition): Boolean;

procedure FindLegalMoves(Moves: TMoveList; const Board: TBoard; Player: TPlayer);

implementation

{$R *.res}

var
  Zobrist: array[TPosition] of array[WHITE..BLACK] of array[Boolean] of TPositionKey;

var
  PositionTable: TPositionTable;

{ Helper Functions }

function PosToStr(Pos: TPosition; TwoDigits: Boolean): String;
begin
  if TwoDigits then
    Result := Format('%2.2u', [Ord(Pos) + 1])
  else
    Result := Format('%u', [Ord(Pos) + 1]);
end;

function StrToPos(const S: String; out Pos: TPosition): Boolean;
var
  V: Integer;
begin
  Result := False;
  V := StrToIntDef(S, 0);
  if (V >= 1) and (V <= 32) then
  begin
    Pos := TPosition(V - 1);
    Result := True;
  end;
end;

function MoveToStr(const Move: TMovement; TwoDigits: Boolean): String;
var
  I: Integer;
begin
  if Move.Count > 0 then
  begin
    Result := PosToStr(Move.Positions[0], TwoDigits);
    for I := 1 to Move.Count - 1 do
      Result := Result + '-' + PosToStr(Move.Positions[I], TwoDigits);
  end
  else
    Result := '';
end;

function StrToMove(const S: String; out Move: TMovement): Boolean;
var
  B, E: Integer;
begin
  Result := False;
  with Move do
  begin
    Count := 0;
    B := 1;
    E := 1;
    while E <= Length(S) do
    begin
      if S[E] = '-' then
      begin
        if not StrToPos(Copy(S, B, E - B + 1), Positions[Count]) then Exit;
        Inc(Count);
        if Count = Length(Positions) then Exit;
        B := E + 1;
      end;
      Inc(E);
    end;
    if not StrToPos(Copy(S, B, E - B + 1), Positions[Count]) then Exit;
  end;
  Result := True;
end;

procedure PosToCoordinate(Pos: TPosition; out Row, Col: Integer);
begin
  Row := Ord(Pos) shr 2;
  Col := ((Ord(Pos) mod 4) shl 1) + ((Row and 1) xor 1);
end;

function CoordinateToPos(Row, Col: Integer; out Pos: TPosition): Boolean;
begin
  if (Cardinal(Row) <= 7) and (Cardinal(Col) <= 7) and (Odd(Row) xor Odd(Col)) then
  begin
    Pos := TPosition(Row shl 2 + (Col - ((Row and 1) xor 1)) shr 1);
    Result := True;
  end
  else
    Result := False;
end;

function IsJump(SrcPos, DstPos: TPosition; out JumpedPos: TPosition): Boolean;
var
  SrcRow, SrcCol: Integer;
  DstRow, DstCol: Integer;
begin
  Result := False;
  PosToCoordinate(SrcPos, SrcRow, SrcCol);
  PosToCoordinate(DstPos, DstRow, DstCol);
  if Abs(SrcRow - DstRow) <> 2 then Exit;
  if Abs(SrcCol - DstCol) <> 2 then Exit;
  if SrcRow < DstRow then
    Inc(SrcRow)
  else
    Dec(SrcRow);
  if SrcCol < DstCol then
    Inc(SrcCol)
  else
    Dec(SrcCol);
  Result := CoordinateToPos(SrcRow, SrcCol, JumpedPos);
end;

function BitCountOf(Value: DWORD): Integer; inline;
begin
  Result := 0;
  while Value <> 0 do
  begin
    while not ByteBool(Value) do
      Value := Value shr 8;
    if ByteBool(Value and 1) then
      Inc(Result);
    Value := Value shr 1;
  end;
end;

procedure FindLegalMoves(Moves: TMoveList; const Board: TBoard; Player: TPlayer);

  function CanAttack(SideCells, EnemyCells, EmptyCells: DWORD; Direction: Integer): Boolean;
  var
    Cells: DWORD;
  begin
    Result := True;
    if Direction >= 0 then
    begin
      Cells := (SideCells shr 4) and EnemyCells;
      if Cells <> 0 then
      begin
        Cells := Cells shl 4;
        // Does any ODD row attack its BOTTOM-LEFT cell?
        if LongBool((((Cells and $F0F0F0F0)) shr 7) and $F0F0F0F0 and EmptyCells) then Exit;
        // Does any EVEN row attack its BOTTOM-RIGHT cell?
        if LongBool((((Cells and $0F0F0F0F)) shr 9) and $0F0F0F0F and EmptyCells) then Exit;
      end;
      // Does any ODD row attack its BOTTOM-RIGHT cell?
      Cells := ((SideCells and $F0F0F0F0) shr 5) and $0F0F0F0F and EnemyCells;
      if (Cells <> 0) and LongBool((Cells shr 4) and $F0F0F0F0 and EmptyCells) then Exit;
      // Does any EVEN row attack its BOTTOM-LEFT cell?
      Cells := ((SideCells and $0F0F0F0F) shr 3) and $F0F0F0F0 and EnemyCells;
      if (Cells <> 0) and LongBool((Cells shr 4) and $0F0F0F0F and EmptyCells) then Exit;
    end;
    if Direction <= 0 then
    begin
      Cells := (SideCells shl 4) and EnemyCells;
      if Cells <> 0 then
      begin
        Cells := Cells shr 4;
        // Does any EVEN row attack its TOP-RIGHT cell?
        if LongBool((((Cells and $0F0F0F0F)) shl 7) and $0F0F0F0F and EmptyCells) then Exit;
        // Does any ODD row attack its TOP-LEFT cell?
        if LongBool((((Cells and $F0F0F0F0)) shl 9) and $F0F0F0F0 and EmptyCells) then Exit;
      end;
      // Does any EVEN row attack its TOP-LEFT cell?
      Cells := ((SideCells and $0F0F0F0F) shl 5) and $F0F0F0F0 and EnemyCells;
      if (Cells <> 0) and LongBool((Cells shl 4) and $0F0F0F0F and EmptyCells) then Exit;
      // Does any ODD row attack its TOP-RIGHT cell?
      Cells := ((SideCells and $F0F0F0F0) shl 3) and $0F0F0F0F and EnemyCells;
      if (Cells <> 0) and LongBool((Cells shl 4) and $F0F0F0F0 and EmptyCells) then Exit;
    end;
    Result := False;
  end;

  procedure AddAttackMoves(SideCells, EnemyCells, EmptyCells: DWORD; Direction: Integer);
  var
    TargetCells: DWORD;

    procedure FollowAttack(SrcOffset, JmpOffset: Integer);
    var
      Move: TMovement;
      LastPos: ^TPosition;

      procedure CompleteAttack(AttackingCell, EnemyCells, EmptyCells: DWORD);
      var
        AttackedCell: DWORD;
        AttackTargetCell: DWORD;

        procedure AddAttack(Offset: Integer);
        var
          Pos: TPosition;
        begin
          Pos := TPosition(Ord(LastPos^) + Offset);
          Inc(LastPos);
          LastPos^ := Pos;
          with Move do
          begin
            Inc(Count);
            Attacked := Attacked or AttackedCell;
          end;
          CompleteAttack(AttackTargetCell, EnemyCells and not AttackedCell,
            (EmptyCells or AttackingCell or AttackedCell) and not AttackTargetCell);
          with Move do
          begin
            Dec(Count);
            Attacked := Attacked and not AttackedCell;
          end;
          Dec(LastPos);
        end;

      var
        IsAttackComplete: Boolean;
      begin
        IsAttackComplete := True;
        // ODD
        if LongBool(AttackingCell and $F0F0F0F0) then
        begin
          EmptyCells := EmptyCells and $F0F0F0F0;
          EnemyCells := EnemyCells and $0F0F0F0F;
          if Direction >= 0 then
          begin
            // BOTTOM-LEFT
            AttackedCell := (AttackingCell shr 4) and EnemyCells;
            if AttackedCell <> 0 then
            begin
              AttackTargetCell := (AttackedCell shr 3) and EmptyCells;
              if AttackTargetCell <> 0 then
              begin
                IsAttackComplete := False;
                AddAttack(7);
              end;
            end;
            // BOTTOM-RIGHT
            AttackedCell := (AttackingCell shr 5) and EnemyCells;
            if AttackedCell <> 0 then
            begin
              AttackTargetCell := (AttackedCell shr 4) and EmptyCells;
              if AttackTargetCell <> 0 then
              begin
                IsAttackComplete := False;
                AddAttack(9);
              end;
            end;
          end;
          if Direction <= 0 then
          begin
            // TOP-LEFT
            AttackedCell := (AttackingCell shl 4) and EnemyCells;
            if AttackedCell <> 0 then
            begin
              AttackTargetCell := (AttackedCell shl 5) and EmptyCells;
              if AttackTargetCell <> 0 then
              begin
                IsAttackComplete := False;
                AddAttack(-9);
              end;
            end;
            // TOP-RIGHT
            AttackedCell := (AttackingCell shl 3) and EnemyCells;
            if AttackedCell <> 0 then
            begin
              AttackTargetCell := (AttackedCell shl 4) and EmptyCells;
              if AttackTargetCell <> 0 then
              begin
                IsAttackComplete := False;
                AddAttack(-7);
              end;
            end;
          end;
        end
        else // EVEN
        begin
          EmptyCells := EmptyCells and $0F0F0F0F;
          EnemyCells := EnemyCells and $F0F0F0F0;
          if Direction >= 0 then
          begin
            // BOTTOM-LEFT
            AttackedCell := (AttackingCell shr 3) and EnemyCells;
            if AttackedCell <> 0 then
            begin
              AttackTargetCell := (AttackedCell shr 4) and EmptyCells;
              if AttackTargetCell <> 0 then
              begin
                IsAttackComplete := False;
                AddAttack(7);
              end;
            end;
            // BOTTOM-RIGHT
            AttackedCell := (AttackingCell shr 4) and EnemyCells;
            if AttackedCell <> 0 then
            begin
              AttackTargetCell := (AttackedCell shr 5) and EmptyCells;
              if AttackTargetCell <> 0 then
              begin
                IsAttackComplete := False;
                AddAttack(9);
              end;
            end;
          end;
          if Direction <= 0 then
          begin
            // TOP-LEFT
            AttackedCell := (AttackingCell shl 5) and EnemyCells;
            if AttackedCell <> 0 then
            begin
              AttackTargetCell := (AttackedCell shl 4) and EmptyCells;
              if AttackTargetCell <> 0 then
              begin
                IsAttackComplete := False;
                AddAttack(-9);
              end;
            end;
            // EVEN: TOP-RIGHT
            AttackedCell := (AttackingCell shl 4) and EnemyCells;
            if AttackedCell <> 0 then
            begin
              AttackTargetCell := (AttackedCell shl 3) and EmptyCells;
              if AttackTargetCell <> 0 then
              begin
                IsAttackComplete := False;
                AddAttack(-7);
              end;
            end;
          end;
        end;
        if IsAttackComplete then Moves.Add(Move);
      end;

    var
      AttackingCell, AttackedCell, AttackTargetCell: DWORD;
      SrcPos, DstPos: TPosition;
    begin
      DstPos := High(TPosition);
      repeat
        while Byte(TargetCells) = 0 do
        begin
          TargetCells := TargetCells shr 8;
          Dec(DstPos, 8);
        end;
        if LongBool(TargetCells and $00000001) then
        begin
          SrcPos := TPosition(Ord(DstPos) + SrcOffset);
          AttackingCell := $80000000 shr Ord(SrcPos);
          AttackedCell := $80000000 shr (Ord(DstPos) + JmpOffset);
          AttackTargetCell := $80000000 shr Ord(DstPos);
          with Move do
          begin
            Count := 2;
            LastPos := @Positions;
            Attacked := AttackedCell;
          end;
          LastPos^ := SrcPos;
          Inc(LastPos);
          LastPos^ := DstPos;
          CompleteAttack(AttackTargetCell, EnemyCells and not AttackedCell,
            (EmptyCells or AttackingCell or AttackedCell) and not AttackTargetCell);
        end;
        TargetCells := TargetCells shr 1;
        Dec(DstPos);
      until (TargetCells = 0);
    end;

  var
    SideCellsOdd, EnemyCellsOdd, EmptyCellsOdd: DWORD;
    SideCellsEven, EnemyCellsEven, EmptyCellsEven: DWORD;
  begin
    SideCellsOdd := SideCells and $F0F0F0F0;
    SideCellsEven := SideCells and $0F0F0F0F;
    EnemyCellsOdd := EnemyCells and $F0F0F0F0;
    EnemyCellsEven := EnemyCells and $0F0F0F0F;
    EmptyCellsOdd := EmptyCells and $F0F0F0F0;
    EmptyCellsEven := EmptyCells and $0F0F0F0F;
    if Direction >= 0 then
    begin
      // ODD: BOTTOM-LEFT
      TargetCells := (SideCellsOdd shr 4) and EnemyCellsEven;
      if TargetCells <> 0 then
      begin
        TargetCells := (TargetCells shr 3) and EmptyCellsOdd;
        if TargetCells <> 0 then FollowAttack(-7, -3);
      end;
      // ODD: BOTTOM-RIGHT
      TargetCells := (SideCellsOdd shr 5) and EnemyCellsEven;
      if TargetCells <> 0 then
      begin
        TargetCells := (TargetCells shr 4) and EmptyCellsOdd;
        if TargetCells <> 0 then FollowAttack(-9, -4);
      end;
      // EVEN: BOTTOM-LEFT
      TargetCells := (SideCellsEven shr 3) and EnemyCellsOdd;
      if TargetCells <> 0 then
      begin
        TargetCells := (TargetCells shr 4) and EmptyCellsEven;
        if TargetCells <> 0 then FollowAttack(-7, -4);
      end;
      // EVEN: BOTTOM-RIGHT
      TargetCells := (SideCellsEven shr 4) and EnemyCellsOdd;
      if TargetCells <> 0 then
      begin
        TargetCells := (TargetCells shr 5) and EmptyCellsEven;
        if TargetCells <> 0 then FollowAttack(-9, -5);
      end;
    end;
    if Direction <= 0 then
    begin
      // ODD: TOP-LEFT
      TargetCells := (SideCellsOdd shl 4) and EnemyCellsEven;
      if TargetCells <> 0 then
      begin
        TargetCells := (TargetCells shl 5) and EmptyCellsOdd;
        if TargetCells <> 0 then FollowAttack(9, 5);
      end;
      // ODD: TOP-RIGHT
      TargetCells := (SideCellsOdd shl 3) and EnemyCellsEven;
      if TargetCells <> 0 then
      begin
        TargetCells := (TargetCells shl 4) and EmptyCellsOdd;
        if TargetCells <> 0 then FollowAttack(7, 4);
      end;
      // EVEN: TOP-LEFT
      TargetCells := (SideCellsEven shl 5) and EnemyCellsOdd;
      if TargetCells <> 0 then
      begin
        TargetCells := (TargetCells shl 4) and EmptyCellsEven;
        if TargetCells <> 0 then FollowAttack(9, 4);
      end;
      // EVEN: TOP-RIGHT
      TargetCells := (SideCellsEven shl 4) and EnemyCellsOdd;
      if TargetCells <> 0 then
      begin
        TargetCells := (TargetCells shl 3) and EmptyCellsEven;
        if TargetCells <> 0 then FollowAttack(7, 3);
      end;
    end;
  end;

  procedure AddSimpleMoves(SideCells, EmptyCells: DWORD; Direction: Integer);
  var
    TargetCells: DWORD;

    procedure AddMoves(SrcOffset: Integer);
    var
      Pos: TPosition;
    begin
      Pos := High(TPosition);
      repeat
        while Byte(TargetCells) = 0 do
        begin
          TargetCells := TargetCells shr 8;
          Dec(Pos, 8);
        end;
        if LongBool(TargetCells and $00000001) then
          Moves.Add(TPosition(Ord(Pos) + SrcOffset), Pos);
        TargetCells := TargetCells shr 1;
        Dec(Pos);
      until (TargetCells = 0);
    end;

  var
    SideCellsOdd, EmptyCellsOdd: DWORD;
    SideCellsEven, EmptyCellsEven: DWORD;
  begin
    SideCellsOdd := SideCells and $F0F0F0F0;
    SideCellsEven := SideCells and $0F0F0F0F;
    EmptyCellsOdd := EmptyCells and $F0F0F0F0;
    EmptyCellsEven := EmptyCells and $0F0F0F0F;
    if Direction >= 0 then
    begin
      // ODD: BOTTOM-LEFT; EVEN: BOTTOM-RIGHT
      TargetCells := (SideCells shr 4) and EmptyCells;
      if TargetCells <> 0 then AddMoves(-4);
      // ODD: BOTTOM-RIGHT
      TargetCells := (SideCellsOdd shr 5) and EmptyCellsEven;
      if TargetCells <> 0 then AddMoves(-5);
      // EVEN: BOTTOM-LEFT
      TargetCells := (SideCellsEven shr 3) and EmptyCellsOdd;
      if TargetCells <> 0 then AddMoves(-3);
    end;
    if Direction <= 0 then
    begin
      // ODD: TOP-LEFT; EVEN: TOP-RIGHT
      TargetCells := (SideCells shl 4) and EmptyCells;
      if TargetCells <> 0 then AddMoves(4);
      // ODD: TOP-RIGHT
      TargetCells := (SideCellsOdd shl 3) and EmptyCellsEven;
      if TargetCells <> 0 then AddMoves(3);
      // EVEN: TOP-LEFT
      TargetCells := (SideCellsEven shl 5) and EmptyCellsOdd;
      if TargetCells <> 0 then AddMoves(5);
    end;
  end;

var
 EmptyCells: DWORD;
 KingCells: DWORD;
 SimpleCells: DWORD;
begin
  Moves.Clear;
  with Board do
  begin
    EmptyCells := not (Blacks or Whites);
    case Player of
      BLACK:
      begin
        KingCells := Blacks and Kings;
        if CanAttack(Blacks, Whites, EmptyCells, +1) or
          ((KingCells <> 0) and CanAttack(KingCells, Whites, EmptyCells, -1)) then
        begin
          if KingCells <> 0 then AddAttackMoves(KingCells, Whites, EmptyCells, 0);
          SimpleCells := Blacks and not Kings;
          if SimpleCells <> 0 then AddAttackMoves(SimpleCells, Whites, EmptyCells, +1);
        end
        else
        begin
          AddSimpleMoves(Blacks, EmptyCells, +1);
          AddSimpleMoves(KingCells, EmptyCells, -1);
        end;
      end;
      WHITE:
      begin
        KingCells := Whites and Kings;
        if CanAttack(Whites, Blacks, EmptyCells, -1) or
          ((KingCells <> 0) and CanAttack(KingCells, Blacks, EmptyCells, +1)) then
        begin
          if KingCells <> 0 then AddAttackMoves(KingCells, Blacks, EmptyCells, 0);
          SimpleCells := Whites and not Kings;
          if SimpleCells <> 0 then AddAttackMoves(SimpleCells, Blacks, EmptyCells, -1);
        end
        else
        begin
          AddSimpleMoves(Whites, EmptyCells, -1);
          AddSimpleMoves(KingCells, EmptyCells, +1);
        end;
      end;
    end;
  end;
end;

{ TMoveList }

destructor TMoveList.Destroy;
begin
  SetCapacity(0);
  inherited Destroy;
end;

procedure TMoveList.SetCapacity(Value: Integer);
begin
  if fCapacity <> Value then
  begin
    fCapacity := Value;
    ReallocMem(fList, fCapacity * SizeOf(TMovement));
    if fCount > fCapacity then
      fCount := fCapacity;
  end;
end;

function TMoveList.GetMovePtr(Index: Integer): PMovement;
begin
  Result := fList;
  Inc(Result, Index);
end;

function TMoveList.GetMoves(Index: Integer): TMovement;
begin
  if (Index >= 0) and (Index < Count) then
    Result := MovePtr[Index]^
  else
    Result.Count := 0;
end;

procedure TMoveList.Grow;
begin
  Inc(fCapacity, 32);
  ReallocMem(fList, fCapacity * SizeOf(TMovement));
end;

procedure TMoveList.Clear;
begin
  fCount := 0;
end;

procedure TMoveList.Add(const Move: TMovement);
var
  Item, NextItem: PMovement;
  Index: Integer;
begin
  if fCount = fCapacity then Grow;
  Item := fList;
  Index := 0;
  while (Index < fCount) and (Item^.Count >= Move.Count) do
  begin
    Inc(Index);
    Inc(Item);
  end;
  if Index < fCount then
  begin
    NextItem := Item;
    Inc(NextItem);
    System.Move(Item^, NextItem^, (fCount - Index) * SizeOf(TMovement));
  end;
  Item^ := Move;
  Inc(fCount);
end;

procedure TMoveList.Add(SrcPos, DstPos: TPosition);
var
  Item: PMovement;
begin
  if fCount = fCapacity then Grow;
  Item := fList;
  Inc(Item, fCount);
  with Item^ do
  begin
    Count := 2;
    Positions[0] := SrcPos;
    Positions[1] := DstPos;
    Attacked := 0;
  end;
  Inc(fCount);
end;

procedure TMoveList.Filter(const Move: TMovement);
var
  Index: Integer;
  Item: PMovement;
  FreeItem: PMovement;
  RelocateItem: PMovement;
  RelocateCount: Integer;
begin
  FreeItem := nil;
  RelocateItem := nil;
  RelocateCount := 0;
  Item := fList;
  Inc(Item, fCount - 1);
  for Index := fCount - 1 downto 0 do
  begin
    if (Item^.Count > Move.Count) and
       CompareMem(@Item^.Positions, @Move.Positions, Move.Count * SizeOf(TPosition)) then
    begin
      if (RelocateCount > 0) and Assigned(FreeItem) then
      begin
        System.Move(RelocateItem^, FreeItem^, RelocateCount * SizeOf(TMovement));
        FreeItem := nil;
      end;
      Dec(Item^.Count, Move.Count - 1);
      System.Move(Item^.Positions[Move.Count - 1], Item^.Positions, (Item^.Count - Move.Count + 2) * SizeOf(TPosition));
      RelocateItem := Item;
      Inc(RelocateCount);
    end
    else
    begin
      FreeItem := Item;
      Dec(fCount);
    end;
    Dec(Item);
  end;
  if (RelocateCount > 0) and Assigned(FreeItem) then
    System.Move(RelocateItem^, FreeItem^, RelocateCount * SizeOf(TMovement));
end;

procedure TMoveList.Filter(SrcPos, DstPos: TPosition);
var
  Move: TMovement;
begin
  Move.Count := 2;
  Move.Positions[0] := SrcPos;
  Move.Positions[1] := DstPos;
  Filter(Move);
end;

procedure TMoveList.Filter(Pos: TPosition);
var
  Move: TMovement;
begin
  Move.Count := 1;
  Move.Positions[0] := Pos;
  Filter(Move);
end;

function TMoveList.IndexOf(const Move: TMovement): Integer;
var
  Item: PMovement;
  Index: Integer;
begin
  Result := -1;
  Item := fList;
  for Index := 0 to fCount - 1 do
  begin
    if CompareMem(Item, @Move, SizeOf(Move.Count) + Move.Count * SizeOf(TPosition)) then
    begin
      Result := Index;
      Exit;
    end;
    Inc(Item);
  end;
end;

function TMoveList.MoveExists(const Move: TMovement): Boolean;
begin
  Result := (IndexOf(Move) >= 0);
end;

function TMoveList.MoveExists(SrcPos, DstPos: TPosition): Boolean;
var
  Move: TMovement;
begin
  Move.Count := 0;
  Move.Positions[0] := SrcPos;
  Move.Positions[1] := DstPos;
  Result := (NextMatchedMove(Move) >= 0);
end;

function TMoveList.CompleteMove(var Move: TMovement): Boolean;
var
  Item: PMovement;
  Index: Integer;
  NextPos: TPosition;
  NextPosFound: Boolean;
begin
  Result := False;
  NextPos := P01;
  repeat
    NextPosFound := False;
    Item := fList;
    for Index := 0 to fCount - 1 do
    begin
      if (Move.Count <= Item^.Count) and
         CompareMem(@Move.Positions, @Item^.Positions, Move.Count * SizeOf(TPosition)) then
      begin
        if Move.Count = Item^.Count then
        begin
          Result := True;
          Exit;
        end;
        if NextPosFound and (Item^.Positions[Move.Count] <> NextPos) then
          Exit;
        NextPos := Item^.Positions[Move.Count];
        NextPosFound := True;
      end;
      Inc(Item);
    end;
    if NextPosFound then
    begin
      Move.Positions[Move.Count] := NextPos;
      Inc(Move.Count);
    end;
  until not NextPosFound;
end;

function TMoveList.NextMatchedMove(const Move: TMovement; StartIndex: Integer): Integer;
var
  Item: PMovement;
  Index: Integer;
begin
  Result := -1;
  Item := fList;
  Inc(Item, StartIndex + 1);
  for Index := StartIndex + 1 to fCount - 1 do
  begin
    if (Move.Count <= Item^.Count) and
       CompareMem(@Move.Positions, @Item^.Positions, Move.Count * SizeOf(TPosition)) then
    begin
      Result := Index;
      Exit;
    end;
    Inc(Item);
  end;
end;

procedure TMoveList.LoadFromStream(Stream: TStream);
begin
  Stream.Read(fCount, SizeOf(fCount));
  if fCount > 0 then
  begin
    if fCount > fCapacity then
      SetCapacity(fCount);
    Stream.Read(fList^, fCount * SizeOf(TMovement));
  end;
end;

procedure TMoveList.SaveToStream(Stream: TStream);
begin
  Stream.Write(fCount, SizeOf(fCount));
  if fCount > 0 then
    Stream.Write(fList^, fCount * SizeOf(TMovement));
end;

{ TCheckersBoard }

constructor TCheckersBoard.Create;
begin
  inherited Create;
  fLegalMoves := TMoveList.Create;
end;

destructor TCheckersBoard.Destroy;
begin
  fLegalMoves.Free;
  inherited Destroy;
end;

procedure TCheckersBoard.SetBoard(const Value: TBoard);
begin
  if not CompareMem(@fBoard, @Value, SizeOf(TBoard)) then
  begin
    fBoard := Value;
    DoChange;
  end;
end;

function TCheckersBoard.GetPieces(Pos: TPosition): TPiece;
var
  Mask: DWORD;
begin
  Mask := $80000000 shr Ord(Pos);
  with fBoard do
  begin
    if LongBool(Blacks and Mask) then
      if LongBool(Kings and Mask) then
        Result := BLACK_KING
      else
        Result := BLACK
    else if LongBool(Whites and Mask) then
      if LongBool(Kings and Mask) then
        Result := WHITE_KING
      else
        Result := WHITE
    else
      Result := EMPTY;
  end;
end;

procedure TCheckersBoard.SetPieces(Pos: TPosition; Value: TPiece);
var
  Mask: DWORD;
begin
  if Pieces[Pos] <> Value then
  begin
    Mask := $80000000 shr Ord(Pos);
    with fBoard do
    begin
      Blacks := Blacks and not Mask;
      Whites := Whites and not Mask;
      Kings := Kings and not Mask;
      if Value in [BLACK, BLACK_KING] then
        Blacks := Blacks or Mask
      else if Value in [WHITE, WHITE_KING] then
        Whites := Whites or Mask;
      if Value in [BLACK_KING, WHITE_KING] then
        Kings := Kings or Mask;
    end;
    DoChange;
  end;
end;

function TCheckersBoard.GetLegalMoves: TMoveList;
begin
  if not Prepared then
    PrepareLegalMoves;
  Result := fLegalMoves;
end;

function TCheckersBoard.GetLegalMovesCount: Integer;
begin
  Result := LegalMoves.Count;
end;

function TCheckersBoard.GetBoardAsText: String;
var
  Pos: TPosition;
  S: PChar;
begin
  SetLength(Result, 32);
  S := PChar(Result);
  for Pos := Low(TPosition) to High(TPosition) do
  begin
    case Pieces[Pos] of
      BLACK:      S^ := 'b';
      BLACK_KING: S^ := 'B';
      WHITE:      S^ := 'w';
      WHITE_KING: S^ := 'W';
    else
      S^ := ' ';
    end;
    Inc(S);
  end;
end;

procedure TCheckersBoard.SetBoardAsText(const Value: String);
var
  Pos: TPosition;
  S: PChar;
begin
  BeginUpdate;
  try
    S := PChar(Value);
    for Pos := Low(TPosition) to High(TPosition) do
      if S^ <> #0 then
      begin
        case S^ of
          'b': Pieces[Pos] := BLACK;
          'B': Pieces[Pos] := BLACK_KING;
          'w': Pieces[Pos] := WHITE;
          'W': Pieces[Pos] := WHITE_KING;
        else
          Pieces[Pos] := EMPTY;
        end;
        Inc(S);
      end
      else
        Pieces[Pos] := EMPTY;
  finally
    EndUpdate;
  end;
end;

function TCheckersBoard.GetDistinctPieceCount: Byte;
begin
  Result := 2;
end;

function TCheckersBoard.GetAvgMoveCount: Cardinal;
begin
  Result := 4;
end;

function TCheckersBoard.GetKey: TPositionKey;
begin
  Result := fKey;
end;

function TCheckersBoard.GetScore: Integer;
const
  SimplePieceRank = 500;
  KingPieceRank   = 750;
  SimplePieceRanks: array[WHITE..BLACK] of array[TPosition] of Integer = (
    (SimplePieceRank + 8, SimplePieceRank + 8, SimplePieceRank + 8, SimplePieceRank + 8,
     SimplePieceRank + 7, SimplePieceRank + 7, SimplePieceRank + 7, SimplePieceRank + 7,
     SimplePieceRank + 6, SimplePieceRank + 6, SimplePieceRank + 6, SimplePieceRank + 6,
     SimplePieceRank + 5, SimplePieceRank + 5, SimplePieceRank + 5, SimplePieceRank + 5,
     SimplePieceRank + 4, SimplePieceRank + 4, SimplePieceRank + 4, SimplePieceRank + 4,
     SimplePieceRank + 3, SimplePieceRank + 3, SimplePieceRank + 3, SimplePieceRank + 3,
     SimplePieceRank + 2, SimplePieceRank + 2, SimplePieceRank + 2, SimplePieceRank + 2,
     SimplePieceRank + 1, SimplePieceRank + 1, SimplePieceRank + 1, SimplePieceRank + 1),
    (SimplePieceRank + 1, SimplePieceRank + 1, SimplePieceRank + 1, SimplePieceRank + 1,
     SimplePieceRank + 2, SimplePieceRank + 2, SimplePieceRank + 2, SimplePieceRank + 2,
     SimplePieceRank + 3, SimplePieceRank + 3, SimplePieceRank + 3, SimplePieceRank + 3,
     SimplePieceRank + 4, SimplePieceRank + 4, SimplePieceRank + 4, SimplePieceRank + 4,
     SimplePieceRank + 5, SimplePieceRank + 5, SimplePieceRank + 5, SimplePieceRank + 5,
     SimplePieceRank + 6, SimplePieceRank + 6, SimplePieceRank + 6, SimplePieceRank + 6,
     SimplePieceRank + 7, SimplePieceRank + 7, SimplePieceRank + 7, SimplePieceRank + 7,
     SimplePieceRank + 8, SimplePieceRank + 8, SimplePieceRank + 8, SimplePieceRank + 8));
var
  Mask: DWORD;
  Pos: TPosition;
begin
  Result := 0;
  Mask := $00000001;
  for Pos := High(TPosition) downto Low(TPosition) do
  begin
    with fBoard do
    begin
      if LongBool(Blacks and Mask) then
        if LongBool(Kings and Mask) then
          Inc(Result, KingPieceRank)
        else
          Inc(Result, SimplePieceRank) // Inc(Result, SimplePieceRanks[BLACK][Pos])
      else if LongBool(Whites and Mask) then
        if LongBool(Kings and Mask) then
          Dec(Result, KingPieceRank)
        else
          Dec(Result, SimplePieceRank) // Dec(Result, SimplePieceRanks[WHITE][Pos])
    end;
    Mask := Mask shl 1;
  end;
  if Player = WHITE then
    Result := -Result;
end;

class function TCheckersBoard.GetPositionTable: TPositionTable;
begin
  Result := PositionTable;
end;

procedure TCheckersBoard.Transpose(Gameboard: TGameboard; TargetMove: Integer);
var
  SrcPos, DstPos, JumpedPos: TPosition;
  SrcCell, DstCell, JumpedCell: DWORD;
  AttackedCells: DWORD;
begin
  fPrepared := False;
  with TCheckersBoard(Gameboard) do
  begin
    Self.fBoard := fBoard;
    Self.ActivePlayer := Player xor BOTH_PLAYERS;
    Self.fKey := fKey xor PlayerZobristMask;
    with fLegalMoves.MovePtr[TargetMove]^ do
    begin
      SrcPos := Positions[0];
      DstPos := Positions[Count - 1];
      AttackedCells := Attacked;
    end;
  end;
  SrcCell := $80000000 shr Ord(SrcPos);
  DstCell := $80000000 shr Ord(DstPos);
  with fBoard do
  begin
    if Player = WHITE then
    begin
      Blacks := (Blacks and not SrcCell) or DstCell;
      if LongBool(Kings and SrcCell) then
      begin
        Kings := (Kings and not SrcCell) or DstCell;
        fKey := fKey xor Zobrist[SrcPos][BLACK][True] xor Zobrist[DstPos][BLACK][True];
      end
      else if LongBool(DstCell and $0000000F) then
      begin
        Kings := Kings or DstCell;
        fKey := fKey xor Zobrist[SrcPos][BLACK][False] xor Zobrist[DstPos][BLACK][True];
      end
      else
        fKey := fKey xor Zobrist[SrcPos][BLACK][False] xor Zobrist[DstPos][BLACK][False];
      if AttackedCells <> 0 then
      begin
        JumpedPos := High(TPosition);
        repeat
          while Byte(AttackedCells) = 0 do
          begin
            AttackedCells := AttackedCells shr 8;
            Dec(JumpedPos, 8);
          end;
          if LongBool(AttackedCells and $00000001) then
          begin
            JumpedCell := $80000000 shr Ord(JumpedPos);
            Whites := Whites and not JumpedCell;
            if LongBool(Kings and JumpedCell) then
            begin
              Kings := Kings and not JumpedCell;
              fKey := fKey xor Zobrist[JumpedPos][WHITE][True];
            end
            else
              fKey := fKey xor Zobrist[JumpedPos][WHITE][False];
          end;
          AttackedCells := AttackedCells shr 1;
          Dec(JumpedPos);
        until AttackedCells = 0;
      end;
    end
    else
    begin
      Whites := (Whites and not SrcCell) or DstCell;
      if LongBool(Kings and SrcCell) then
      begin
        Kings := (Kings and not SrcCell) or DstCell;
        fKey := fKey xor Zobrist[SrcPos][WHITE][True] xor Zobrist[DstPos][WHITE][True];
      end
      else if LongBool(DstCell and $F0000000) then
      begin
        Kings := Kings or DstCell;
        fKey := fKey xor Zobrist[SrcPos][WHITE][False] xor Zobrist[DstPos][WHITE][True];
      end
      else
        fKey := fKey xor Zobrist[SrcPos][WHITE][False] xor Zobrist[DstPos][WHITE][False];
      if AttackedCells <> 0 then
      begin
        JumpedPos := High(TPosition);
        repeat
          while Byte(AttackedCells) = 0 do
          begin
            AttackedCells := AttackedCells shr 8;
            Dec(JumpedPos, 8);
          end;
          if LongBool(AttackedCells and $00000001) then
          begin
            JumpedCell := $80000000 shr Ord(JumpedPos);
            Blacks := Blacks and not JumpedCell;
            if LongBool(Kings and JumpedCell) then
            begin
              Kings := Kings and not JumpedCell;
              fKey := fKey xor Zobrist[JumpedPos][BLACK][True];
            end
            else
              fKey := fKey xor Zobrist[JumpedPos][BLACK][False];
          end;
          AttackedCells := AttackedCells shr 1;
          Dec(JumpedPos);
        until AttackedCells = 0;
      end;
    end;
  end;
end;

function TCheckersBoard.CalculateKey: TPositionKey;
var
  Pos: TPosition;
  Mask: DWORD;
begin
  Result := 0;
  Mask := $00000001;
  for Pos := High(TPosition) downto Low(TPosition) do
  begin
    with fBoard do
    begin
      if LongBool(Blacks and Mask) then
        Result := Result xor Zobrist[Pos][BLACK][LongBool(Kings and Mask)]
      else if LongBool(Whites and Mask) then
        Result := Result xor Zobrist[Pos][WHITE][LongBool(Kings and Mask)];
    end;
    Mask := Mask shl 1;
  end;
  if Player = BLACK then
    Result := Result xor PlayerZobristMask;
end;

procedure TCheckersBoard.InitializeSearch;
begin
  fKey := CalculateKey;
  inherited InitializeSearch;
end;

procedure TCheckersBoard.GetTextPresentation(out Board, Moves: String);
var
  R, C, M: Integer;
  Pos: TPosition;
begin
  Board := '';
  for R := 0 to 7 do
  begin
    if R <> 0 then
      Board := Board + #13#10;
    for C := 0 to 7 do
    begin
      if C <> 0 then
        Board := Board + ' ';
      if CoordinateToPos(R, C, Pos) then
        case Pieces[Pos] of
          BLACK:      Board := Board + 'b';
          BLACK_KING: Board := Board + 'B';
          WHITE:      Board := Board + 'w';
          WHITE_KING: Board := Board + 'W';
        else
          Board := Board + '-';
        end
      else
        Board := Board + '.';
    end;
    Board := Board + ^I^I;
    for C := 0 to 7 do
    begin
      if CoordinateToPos(R, C, Pos) then
        Board := Board + PosToStr(Pos)
      else
        Board := Board + '  ';
    end;
  end;
  Moves := '';
  for M := 0 to LegalMoves.Count - 1 do
  begin
    if M <> 0 then
      Moves := Moves + #13#10;
    Moves := Moves + Format('%d) %s', [M + 1, MoveToStr(LegalMoves[M])]);
  end;
end;

procedure TCheckersBoard.DoTurn;
begin
  fLastPosition := 0;
  fPrepared := False;
  inherited DoTurn;
end;

procedure TCheckersBoard.DoChange;
begin
  if not fMoving then
  begin
    fLastPosition := 0;
    fPrepared := False;
  end;
  inherited DoChange;
end;

procedure TCheckersBoard.DoPromote(Pos: TPosition);
begin
  if Assigned(OnPromote) then
    OnPromote(Self, Pos);
end;

procedure TCheckersBoard.DoSimpleMove(SrcPos, DstPos: TPosition);
begin
  if Assigned(OnSimpleMove) then
    OnSimpleMove(Self, SrcPos, DstPos);
end;

procedure TCheckersBoard.DoAttackMove(SrcPos, DstPos, JmpPos: TPosition);
begin
  if Assigned(OnAttackMove) then
    OnAttackMove(Self, SrcPos, DstPos, JmpPos);
end;

procedure TCheckersBoard.PrepareLegalMoves;
begin
  if not Prepared then
  begin
    fPrepared := True;
    FindLegalMoves(fLegalMoves, fBoard, Player);
    if fActivePosition <> 0 then
    begin
      fLegalMoves.Filter(TPosition(fActivePosition));
      fActivePosition := 0;
    end;
  end;
end;

function TCheckersBoard.MovePiece(SrcPos, DstPos: TPosition): Boolean;
var
  JmpPos: TPosition;
  Promote: Boolean;
begin
  Result := False;
  if LegalMoves.MoveExists(SrcPos, DstPos) then
  begin
    fMoving := True;
    try
      Promote := False;
      if IsJump(SrcPos, DstPos, JmpPos) then
      begin
        DoAttackMove(SrcPos, DstPos, JmpPos);
        Pieces[JmpPos] := EMPTY;
      end
      else
        DoSimpleMove(SrcPos, DstPos);
      if (DstPos <= P04) and (Pieces[SrcPos] = WHITE) then
      begin
        Pieces[DstPos] := WHITE_KING;
        Promote := True;
      end
      else if (DstPos >= P29) and (Pieces[SrcPos] = BLACK) then
      begin
        Pieces[DstPos] := BLACK_KING;
        Promote := True;
      end
      else
        Pieces[DstPos] := Pieces[SrcPos];
      Pieces[SrcPos] := EMPTY;
      fLegalMoves.Filter(SrcPos, DstPos);
      fLastPosition := Ord(DstPos);
      DoChange;
    finally
      fMoving := False;
    end;
    if Promote then
      DoPromote(DstPos);
    Result := True;
  end;
end;

procedure TCheckersBoard.StartNewGame;
var
  Pos: TPosition;
begin
  BeginUpdate;
  try
    Player := BLACK;
    for Pos := P01 to P12 do
      Pieces[Pos] := BLACK;
    for Pos := P13 to P20 do
      Pieces[Pos] := EMPTY;
    for Pos := P21 to P32 do
      Pieces[Pos] := WHITE;
  finally
    EndUpdate;
  end;
end;

procedure TCheckersBoard.SaveToStream(Stream: TStream);
begin
  inherited SaveToStream(Stream);
  Stream.Write(fBoard, SizeOf(fBoard));
  Stream.Write(fLastPosition, SizeOf(fLastPosition));
end;

procedure TCheckersBoard.LoadFromStream(Stream: TStream);
begin
  BeginUpdate;
  try
    inherited LoadFromStream(Stream);
    Stream.Read(fBoard, SizeOf(fBoard));
    Stream.Read(fActivePosition, SizeOf(fActivePosition));
    DoChange;
  finally
    EndUpdate;
  end;
end;

function TCheckersBoard.ApplyLegalMove(Index: Integer): Boolean;
begin
  Result := False;
  if (Index >= 0) and (Index < LegalMoves.Count) then
    Result := ApplyMove(LegalMoves[Index]);
end;

function TCheckersBoard.ApplyMove(Move: TMovement): Boolean;
var
  I: Integer;
begin
  Result := False;
  if Move.Count > 1 then
  begin
    for I := 0 to Move.Count - 2 do
      if not ApplyMove(Move.Positions[I], Move.Positions[I + 1]) then
        Exit;
    Result := True;
  end;
end;

function TCheckersBoard.ApplyMove(Src, Dst: TPosition): Boolean;
begin
  Result := False;
  if MovePiece(Src, Dst) then
  begin
    if LegalMoves.Count = 0 then
      ChangeSide;
    Result := True;
  end;
end;

{ TCheckers }

constructor TCheckers.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  CheckersBoard.OnPromote := PiecePromote;
  CheckersBoard.OnSimpleMove := PieceSimpleMove;
  CheckersBoard.OnAttackMove := PieceAttackMove;
  fPiecesBitmap := TBitmap.Create;
  fPiecesBitmap.OnChange := PiecesBitmapChange;
  fTwoDigitsPositions := True;
  ComputerPlayer := WHITE;
  SquareDarkColor := $00008000;
  SquareDarkSelectColor := clLime;
  SquareDarkMarkColor := clLime;
  SquareLightColor := $00AAFFFF;
  SquareLightSelectColor := clRed;
  SquareLightMarkColor := clRed;
end;

destructor TCheckers.Destroy;
begin
  fPiecesBitmap.Free;
  inherited Destroy;
end;

function TCheckers.GetCheckersBoard: TCheckersBoard;
begin
  Result := TCheckersBoard(Gameboard);
end;

function TCheckers.GetBoard: TBoard;
begin
  Result := CheckersBoard.Board;
end;

procedure TCheckers.SetBoard(const Value: TBoard);
begin
  CheckersBoard.Board := Value;
end;

function TCheckers.GetPieces(Pos: TPosition): TPiece;
begin
  Result := CheckersBoard.Pieces[Pos];
end;

procedure TCheckers.SetPieces(Pos: TPosition; Value: TPiece);
begin
  CheckersBoard.Pieces[Pos] := Value;
end;

function TCheckers.GetLegalMoves: TMoveList;
begin
  Result := CheckersBoard.LegalMoves;
end;

procedure TCheckers.SetPiecesBitmap(Value: TBitmap);
begin
  PiecesBitmap.Assign(Value);
end;

function TCheckers.GetPosRect(Pos: TPosition): TRect;
begin
  Result := SquareBounds[PosToSquare(Pos)];
end;

function TCheckers.GetSelectedMove: TMovement;
begin
  if not PathToMove(Selection, Result) then
    Result.Count := 0;
end;

function TCheckers.GetCoordinatesDisplay: TSquareLabel;
begin
  Result := SquareLabel;
end;

procedure TCheckers.SetCoordinatesDisplay(Value: TSquareLabel);
begin
  SquareLabel := Value;
end;

function TCheckers.GetCoordinatesOffset: Integer;
begin
  Result := SquareLabelOffset;
end;

procedure TCheckers.SetCoordinatesOffset(Value: Integer);
begin
  SquareLabelOffset := Value;
end;

procedure TCheckers.DoSquareResize;
begin
  if PiecesBitmap.Empty then
    UpdatePiecesImages;
  inherited DoSquareResize;
end;

procedure TCheckers.DoSquareClick(Square: TSquare; var Handled: Boolean);
var
  Pos: TPosition;
begin
  if SquareToPos(Square, Pos) then
    DoPositionClick(Pos, Handled)
  else
    Handled := True;
  inherited DoSquareClick(Square, Handled);
end;

procedure TCheckers.DoPositionClick(Pos: TPosition; var Handled: Boolean);
begin
  if Assigned(OnPositionClick) then
    OnPositionClick(Self, Pos, Handled);
end;

procedure TCheckers.DoSquareInfoTip(Square: TSquare; var InfoTip: String);
var
  Pos: TPosition;
begin
  if SquareToPos(Square, Pos) then
    DoPositionInfoTip(Pos, InfoTip);
  inherited DoSquareInfoTip(Square, InfoTip);
end;

procedure TCheckers.DoPositionInfoTip(Pos: TPosition; var InfoTip: String);
begin
  if Assigned(OnPositionInfoTip) then
    OnPositionInfoTip(Self, Pos, InfoTip);
end;

procedure TCheckers.DoSquareDragOver(Source, Target: TSquare; var Accept: Boolean);
var
  SrcPos, TrgPos: TPosition;
begin
  if SquareToPos(Source, SrcPos) and SquareToPos(Target, TrgPos) then
    DoPositionDragOver(SrcPos, TrgPos, Accept)
  else
    Accept := False;
  inherited DoSquareDragOver(Source, Target, Accept);
end;

procedure TCheckers.DoPositionDragOver(Source, Target: TPosition; var Accept: Boolean);
begin
  if Assigned(OnPositionDragOver) then
    OnPositionDragOver(Self, Source, Target, Accept);
end;

procedure TCheckers.DoSquareDragDrop(Source, Target: TSquare);
var
  SrcPos, TrgPos: TPosition;
begin
  if SquareToPos(Source, SrcPos) and SquareToPos(Target, TrgPos) then
    DoPositionDragDrop(SrcPos, TrgPos);
  inherited DoSquareDragDrop(Source, Target);
end;

procedure TCheckers.DoPositionDragDrop(Source, Target: TPosition);
begin
  if Assigned(OnPositionDragDrop) then
    OnPositionDragDrop(Self, Source, Target);
end;

procedure TCheckers.DoSquareDragStart;
begin
  DoPositionDragStart;
  inherited DoSquareDragStart;
end;

procedure TCheckers.DoPositionDragStart;
begin
  if Assigned(OnPositionDragStart) then
    OnPositionDragStart(Self);
end;

procedure TCheckers.DoSquareDragEnd;
begin
  DoPositionDragEnd;
  inherited DoSquareDragEnd;
end;

procedure TCheckers.DoPositionDragEnd;
begin
  if Assigned(OnPositionDragEnd) then
    OnPositionDragEnd(Self);
end;

procedure TCheckers.DoSimpleMove(SrcPos, DstPos: TPosition);
begin
  if Assigned(OnSimpleMove) then
    OnSimpleMove(Self, SrcPos, DstPos);
end;

procedure TCheckers.DoAttackMove(SrcPos, DstPos, JmpPos: TPosition);
begin
  if Assigned(OnAttackMove) then
    OnAttackMove(Self, SrcPos, DstPos, JmpPos);
end;

procedure TCheckers.DoPromote(Pos: TPosition);
begin
  if Assigned(OnPromote) then
    OnPromote(Self, Pos);
end;

procedure TCheckers.PiecePromote(Sender: TObject; Pos: TPosition);
begin
  DoPromote(Pos);
end;

procedure TCheckers.PieceSimpleMove(Sender: TObject; SrcPos, DstPos: TPosition);
begin
  DoSimpleMove(SrcPos, DstPos);
  StorePath([PosToSquare(SrcPos), PosToSquare(DstPos)]);
end;

procedure TCheckers.PieceAttackMove(Sender: TObject; SrcPos, DstPos, JmpPos: TPosition);
begin
  DoAttackMove(SrcPos, DstPos, JmpPos);
  StorePath([PosToSquare(SrcPos), PosToSquare(DstPos)]);
end;

procedure TCheckers.PiecesBitmapChange(Sender: TObject);
begin
  UpdatePiecesImages;
end;

function TCheckers.GetGameboardClass: TGameboardClass;
begin
  Result := TCheckersBoard;
end;

procedure TCheckers.SetImageIndexOfSquare(Square: TSquare; ImageIndex: Integer);
var
  Pos: TPosition;
begin
  if SquareToPos(Square, Pos) then
    case ImageIndex of
      0: Pieces[Pos] := BLACK;
      1: Pieces[Pos] := BLACK_KING;
      2: Pieces[Pos] := WHITE;
      3: Pieces[Pos] := WHITE_KING;
    else
      Pieces[Pos] := EMPTY;
    end;
end;

function TCheckers.GetImageIndexOfSquare(Square: TSquare): Integer;
var
  Pos: TPosition;
begin
  Result := -1;
  if SquareToPos(Square, Pos) then
    case Pieces[Pos] of
      BLACK:
        Result := 0;
      BLACK_KING:
        Result := 1;
      WHITE:
        Result := 2;
      WHITE_KING:
        Result := 3;
    end;
end;

function TCheckers.GetSquareLabel(Square: TSquare): String;
var
  Pos: TPosition;
begin
  if SquareToPos(Square, Pos) then
    Result := PosToStr(Pos, False)
  else
    Result := '';
end;

function TCheckers.IsPathComplete(const Path: array of TSquare): Boolean;
var
  Move: TMovement;
begin
  Result := False;
  if PathToMove(Path, Move) then
    Result := LegalMoves.MoveExists(Move);
end;

function TCheckers.CanPlayPath(const Path: array of TSquare): Boolean;
var
  Move: TMovement;
begin
  Result := False;
  if (Length(Path) >= 2) and PathToMove(Path, Move) then
    Result := (LegalMoves.NextMatchedMove(Move) >= 0);
end;

function TCheckers.PlayPath(const Path: array of TSquare): Boolean;
var
  Move: TMovement;
begin
  Result := False;
  if PathToMove(Path, Move) then
    Result := Play(Move);
end;

function TCheckers.CompletePath(var Path: TMovementPath): Boolean;
var
  Move: TMovement;
  Index: Integer;
begin
  Result := False;
  if PathToMove(Path, Move) then
  begin
    Index := Move.Count;
    Result := LegalMoves.CompleteMove(Move);
    SetLength(Path, Move.Count);
    while Index < Move.Count do
    begin
      Path[Index] := PosToSquare(Move.Positions[Index]);
      Inc(Index);
    end;
  end;
end;

function TCheckers.IsLegalSquare(Square: TSquare): Boolean;
begin
  Result := inherited IsLegalSquare(Square) and IsLogicalDarkSquare(Square);
end;

function TCheckers.CanSelectSquare(Square: TSquare): Boolean;
var
  Pos: TPosition;
  Move: TMovement;
begin
  Result := False;
  if inherited CanSelectSquare(Square) and SquareToPos(Square, Pos) and
     PathToMove(Selection, Move) and (Move.Count < Length(Move.Positions)) then
  begin
    Move.Positions[Move.Count] := Pos;
    Inc(Move.Count);
    Result := (LegalMoves.NextMatchedMove(Move) >= 0);
  end;
end;

function TCheckers.IsLegalMove(const Move: TMovement): Boolean;
begin
  Result := LegalMoves.MoveExists(Move)
end;

function TCheckers.CanPlayMove(const Move: TMovement): Boolean;
begin
  Result := (LegalMoves.NextMatchedMove(Move) >= 0);
end;

function TCheckers.Play(const Move: TMovement): Boolean;
begin
  Result := False;
  if CanPlayMove(Move) then
  begin
    InitializeMovement;
    try
      Result := CheckersBoard.ApplyMove(Move);
    finally
      FinalizeMovement;
    end;
  end;
end;

function TCheckers.MoveOf(Player: TPlayer; TurnNo: Integer): String;
var
  HistoryRec: TGameHistoryRec;
  Move: TMovement;
  Index: Integer;
begin
  Result := '';
  HistoryRec := HistoryOf(Player, TurnNo);
  if Assigned(HistoryRec) then
  begin
    Index := History.IndexOf(HistoryRec);
    if Index < 0 then
    begin
      PathToMove(HistoryRec.MovementPath, Move);
      Result := Result + MoveToStr(Move, TwoDigitsPositions);
      if not HistoryRec.Completed then
        Result := Result + '->';
      Index := History.Count - 1;
      while (Index >= 0) and not History[Index].Completed do
      begin
        PathToMove(History[Index].MovementPath, Move);
        Dec(Move.Count);
        Result := MoveToStr(Move, TwoDigitsPositions) + '-' + Result;
        Dec(Index);
      end;
    end
    else
    begin
      while (Index < History.Count - 1) and not HistoryRec.Completed do
      begin
        PathToMove(HistoryRec.MovementPath, Move);
        Dec(Move.Count);
        Result := Result + MoveToStr(Move, TwoDigitsPositions) + '-';
        Inc(Index);
        HistoryRec := History[Index];
      end;
      PathToMove(HistoryRec.MovementPath, Move);
      Result := Result + MoveToStr(Move, TwoDigitsPositions);
      if not HistoryRec.Completed then
        Result := Result + '->';
    end;
  end;
end;

function TCheckers.IsPositionSelected(Pos: TPosition): Boolean;
begin
  Result := IsSquareSelected(PosToSquare(Pos));
end;

function TCheckers.CanSelectPosition(Pos: TPosition): Boolean;
begin
  Result := CanSelectSquare(PosToSquare(Pos));
end;

function TCheckers.SelectPosition(Pos: TPosition): Boolean;
begin
  Result := SelectSquare(PosToSquare(Pos));
end;

function TCheckers.SelectMove(Index: Integer): Boolean;
begin
  Result := False;
  if (Index >= 0) and (Index < LegalMoves.Count) then
    Result := SelectMove(LegalMoves[Index]);
end;

function TCheckers.SelectMove(const Move: TMovement): Boolean;
begin
  Result := False;
  if LegalMoves.NextMatchedMove(Move) >= 0 then
  begin
    Selection := MoveToPath(Move);
    Result := True;
  end;
end;

function TCheckers.UpdatePiecesImages: Boolean;
var
  DefaultSize: Integer;
  BitmapOfPieces: TBitmap;
begin
  Result := False;
  if not Assigned(PiecesBitmap) or PiecesBitmap.Empty then
  begin
    case SquareSize of
      00..35: DefaultSize := 16;
      36..45: DefaultSize := 24;
      46..55: DefaultSize := 32;
      56..65: DefaultSize := 40;
    else
      DefaultSize := 48;
    end;
    if DefaultSize <> DefaultPieceSize then
    begin
      BitmapOfPieces := TBitmap.Create;
      try
        BitmapOfPieces.LoadFromResourceName(HInstance, 'CHECKERS_PIECES' + IntToStr(DefaultSize));
        ExtractPiecesImages(BitmapOfPieces, 1, 4);
      finally
        BitmapOfPieces.Free;
      end;
      DefaultPieceSize := DefaultSize;
      Result := True;
    end;
  end
  else
  begin
    DefaultPieceSize := 0;
    PiecesBitmap.OnChange := nil;
    try
      ExtractPiecesImages(PiecesBitmap, 1, 4);
    finally
      PiecesBitmap.OnChange := PiecesBitmapChange;
    end;
    Result := True;
  end;
end;

function TCheckers.ClientToPos(X, Y: Integer; out Pos: TPosition): Boolean;
begin
  Result := SquareToPos(ClientToSquare(X, Y), Pos);
end;

function TCheckers.SquareToPos(Square: TSquare; out Pos: TPosition): Boolean;
var
  Row, Col: Integer;
begin
  Result := SquareToCoordinate(Square, Row, Col) and CoordinateToPos(Row, Col, Pos);
end;

function TCheckers.PosToSquare(Pos: TPosition): TSquare;
var
  Row, Col: Integer;
begin
  PosToCoordinate(Pos, Row, Col);
  Result := CoordinateToSquare(Row, Col);
end;

function TCheckers.MoveToPath(const Move: TMovement): TMovementPath;
var
  I: Integer;
begin
  SetLength(Result, Move.Count);
  for I := 0 to Move.Count - 1 do
    Result[I] := PosToSquare(Move.Positions[I]);
end;

function TCheckers.PathToMove(const Path: array of TSquare; out Move: TMovement): Boolean;
var
  I: Integer;
begin
  Result := False;
  if Length(Path) <= Length(Move.Positions) then
  begin
    Move.Count := Length(Path);
    for I := 0 to Length(Path) - 1 do
      if not SquareToPos(Path[I], Move.Positions[I]) then
        Exit;
    Result := True;
  end;
end;

initialization
  InitZobritsNumbers(Zobrist, SizeOf(Zobrist) div SizeOf(TPositionKey));
  PositionTable := TPositionTable.Create;
finalization
  FreeAndNil(PositionTable);
end.

