//--------------------------------------------------------------------------
//  Gameboard - Abstract classes for Gameboards implementation
//  Copyright (C) 2005-2009 Kambiz R. Khojasteh
//--------------------------------------------------------------------------

{$B-} // Complete Boolean Evaluation Off
{$R-} // Range Checking Off
{$Q-} // Overflow Checking Off
{$V-} // Var String Checking Off
{$T-} // Typed @ operator Off
{$X+} // Extended Syntax On
{$P+} // Open String Params On
{$O+} // Optimization On
{$H+} // Use long strings by default

unit Gameboard;

interface

uses
  Windows, Messages, Classes, Contnrs, Controls, Graphics, ImgList,
  SysUtils, Forms;

const
  WM_SOLVER_START       = WM_USER + $0001;
  WM_SOLVER_STOP        = WM_USER + $0002;
  WM_ANIMATOR_START     = WM_USER + $0003;
  WM_ANIMATOR_STOP      = WM_USER + $0004;
  WM_THINKABOUTMOVE     = WM_USER + $0005;
  WM_BOARDCHANGE        = WM_USER + $0006;
  WM_TURNCHANGE         = WM_USER + $0007;
  WM_HISTORYCHANGE      = WM_USER + $0008;
  WM_SELECTIONCHANGE    = WM_USER + $0009;
  WM_BOARDMODECHANGE    = WM_USER + $000A;
  WM_PAUSESTATECHANGE   = WM_USER + $000B;
  WM_PLAYINGSTATECHANGE = WM_USER + $000C;
  WM_GAMEENDED          = WM_USER + $000D;
  WM_FLUSHSQUARE        = WM_USER + $000E;

const
  NONE                  = $00;
  LIGHT_PLAYER          = $01;
  DARK_PLAYER           = $02;
  BOTH_PLAYERS          = LIGHT_PLAYER or DARK_PLAYER;

const
  NO_SQUARE             = -1;
  MAX_VALUE             = 1000000;
  MAX_SEARCH_DEPTH      = 64;
  MAX_SCORE             = MAX_VALUE - MAX_SEARCH_DEPTH;
  PlayerZobristMask     = $456EBD66E88271CE;

type

  EInvalidGameFile = class(Exception);

  TPlayer = Byte;
  TPiece  = Byte;

  TSquare = Integer;
  TMovementPath = array of TSquare;

  TGameboard = class;
  TGameboardClass = class of TGameboard;

  TGameboardControl = class;

  TPositionKey = Int64;
  TPositionFlag = (tfNone, tfExact, tfUpper, tfLower);

  PPositionNode = ^TPositionNode;
  TPositionNode = packed record
    Key: TPositionKey;
    Value: Integer;
    Depth_Flag: Byte;
    BestMove: ShortInt;
  end;

  TPositionTable = class(TObject)
  private
    fTable: PPositionNode;
    fNumOfCells: Cardinal;
    procedure SetNumOfCells(Value: Cardinal);
    function GetNumOfUsedCells: Cardinal;
    function GetSize: Cardinal;
    procedure SetSize(Value: Cardinal);
    function GetUsedSize: Cardinal;
  protected
    procedure UpdateSize(AvgBranch: Cardinal; SearchDepth: Integer; MaxSize: Cardinal);
    function Find(const Key: TPositionKey; out Node: PPositionNode): Boolean; inline;
  public
    destructor Destroy; override;
    procedure Clear;
    procedure SaveToStream(Stream: TStream);
    procedure LoadFromStream(Stream: TStream);
    function BestMoveOf(const Key: TPositionKey): Integer;
    function Retrieve(const Key: TPositionKey; Depth: Integer;
      var Alpha, Beta: Integer; var Value: Integer;
      var BestMove: Integer): Boolean; inline;
    procedure Store(const Key: TPositionKey; Depth: Integer;
      Flag: TPositionFlag; const Value: Integer; BestMove: Integer); inline;
    property NumOfCells: Cardinal read fNumOfCells write SetNumOfCells;
    property NumOfUsedCells: Cardinal read GetNumOfUsedCells;
    property Size: Cardinal read GetSize write SetSize;
    property UsedSize: Cardinal read GetUsedSize;
  end;

  PSearchControl = ^TSearchControl;
  TSearchControl = (scContinue, scStop, scAbort);

  TGameEndState = (gesUnknown, gesDraw, gesLightWon, gesDarkWon);
  TGameStage = (gsStartGame, gsMidGame, gsEndGame);

  TGameboard = class(TPersistent)
  private
    fPlayer: TPlayer;
    fUpdateCount: Integer;
    fUpdated: Boolean;
    fOnChange: TNotifyEvent;
    fOnTurn: TNotifyEvent;
    procedure SetPlayer(Value: TPlayer);
  protected
    procedure DoChange; virtual;
    procedure DoTurn; virtual;
    procedure ChangeSide; virtual;
    function IsUpdateLocked: Boolean;
    procedure DisplayDebugInformation(const FileName: String;
      const Fmt: String; const Args: array of const);
  protected
    function GetDistinctPieceCount: Byte; virtual; abstract;
    function GetWinner: TPlayer; virtual;
    function GetLooser: TPlayer; virtual;
    function GetEndState: TGameEndState; virtual;
    function GetLegalMovesCount: Integer; virtual; abstract;
    function GetBoardAsText: String; virtual; abstract;
    procedure SetBoardAsText(const Value: String); virtual; abstract;
    property DistinctPieceCount: Byte read GetDistinctPieceCount;
  protected
    function GetAvgMoveCount: Cardinal; virtual; abstract;
    function GetKey: TPositionKey; virtual; abstract;
    function GetScore: Integer; virtual; abstract;
    class function GetPositionTable: TPositionTable; virtual; abstract;
    procedure Transpose(Gameboard: TGameboard; TargetMove: Integer); virtual; abstract;
    procedure InitializeSearch; virtual;
    procedure FinalizeSearch; virtual;
    function PerformSearch(SearchDepth: Integer; MaxMemory: Cardinal;
      var Control: TSearchControl): Integer; virtual;
    property Key: TPositionKey read GetKey;
    property Score: Integer read GetScore;
    property AvgMoveCount: Cardinal read GetAvgMoveCount;
    property ActivePlayer: TPlayer read fPlayer write fPlayer;
  public
    constructor Create; virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure GetTextPresentation(out Board, Moves: String); virtual; abstract;
    procedure Assign(Source: TPersistent); override;
    procedure BeginUpdate; virtual;
    procedure EndUpdate; virtual;
    procedure StartNewGame; virtual; abstract;
    procedure ClearBoard; virtual;
    function ApplyLegalMove(Index: Integer): Boolean; virtual; abstract;
    function SearchForBestMove(SearchDepth: Byte; MaxMemory: Cardinal;
      pControl: PSearchControl = nil): Integer; virtual;
    property Winner: TPlayer read GetWinner;
    property Looser: TPlayer read GetLooser;
    property EndState: TGameEndState read GetEndState;
    property LegalMovesCount: Integer read GetLegalMovesCount;
    property BoardAsText: String read GetBoardAsText write SetBoardAsText;
    property Player: TPlayer read fPlayer write SetPlayer;
    property OnTurn: TNotifyEvent read fOnTurn write fOnTurn;
    property OnChange: TNotifyEvent read fOnChange write fOnChange;
  end;

  TGameSolver = class(TThread)
  private
    IdleEvent: THandle;
    TaskEvent: THandle;
    Gameboard: TGameboard;
    fNotifyWnd: HWND;
    fSearchDepth: Byte;
    fMaxMemory: Cardinal;
    fState: TSearchControl;
  protected
    procedure Execute; override;
  public
    constructor Create; reintroduce;
    destructor Destroy; override;
    procedure Terminate;
    function IsBusy: Boolean;
    function WaitForIdle: Boolean;
    function Solve(AGameboard: TGameboard): Boolean;
    property NotifyWnd: HWND read fNotifyWnd write fNotifyWnd;
    property State: TSearchControl read fState write fState;
    property SearchDepth: Byte read fSearchDepth write fSearchDepth;
    property MaxMemory: Cardinal read fMaxMemory write fMaxMemory;
  end;

  TAnimator = class(TThread)
  private
    Control: TGameBoardControl;
    IdleEvent: THandle;
    TaskEvent: THandle;
    AnimationPath: TMovementPath;
    Cancelled: Boolean;
    fAnimateDelay: Byte;
    fAnimateStep: Byte;
  protected
    procedure Execute; override;
    procedure PerformAnimation(const Path: array of TSquare); virtual;
  public
    constructor Create(AControl: TGameboardControl); reintroduce;
    destructor Destroy; override;
    procedure Terminate;
    function IsBusy: Boolean;
    function WaitForIdle(BlockInputs: Boolean): Boolean;
    function Animate(const Path: array of TSquare; Reverse: Boolean): Boolean;
    procedure CancelAnimation;
    property AnimateDelay: Byte read fAnimateDelay write fAnimateDelay default 10;
    property AnimateStep: Byte read fAnimateStep write fAnimateStep default 1;
  end;

  TGameHistoryRec = class(TPersistent)
  private
    fBoardAfterMove: TMemoryStream;
    fMovementPath: TMovementPath;
    fElapsedTime: Cardinal;
    fCompleted: Boolean;
    procedure SetMovementPath(const Value: TMovementPath);
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Clear; virtual;
    procedure Assign(Source: TPersistent); override;
    procedure AssignTo(Dest: TPersistent); override;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure AppendMove(const Path: array of TSquare); virtual;
    function IsEmpty: Boolean; virtual;
    property BoardAfterMove: TMemoryStream read fBoardAfterMove;
    property MovementPath: TMovementPath read fMovementPath write SetMovementPath;
    property ElapsedTime: Cardinal read fElapsedTime write fElapsedTime;
    property Completed: Boolean read fCompleted write fCompleted;
  end;

  TGameHistory = class(TObjectList)
  private
    fCurrent: Integer;
    fInitialBoard: TGameboard;
    fClosed: Boolean;
    fOnChange: TNotifyEvent;
    fUpdateCount: Integer;
    function GetItems(Index: Integer): TGameHistoryRec;
    function GetPlayers(Index: Integer): TPlayer;
    procedure SetCurrent(Value: Integer);
    procedure InitialBoardChange(Sender: TObject);
  protected
    procedure DoChange; virtual;
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
    function IsUpdateLocked: Boolean;
    property Closed: Boolean read fClosed write fClosed;
  public
    constructor Create(AGameboardClass: TGameboardClass); reintroduce; virtual;
    destructor Destroy; override;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure Clear; override;
    procedure Add(AItem: TGameHistoryRec); reintroduce; virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    function IndexOf(AGameHistoryRec: TGameHistoryRec): Integer; overload;
    function IndexOf(Player: TPlayer; TurnNo: Integer): Integer; overload;
    function TurnOf(Player: TPlayer; Index: Integer): Integer;
    function TotalTimeOf(Player: TPlayer; TurnNo: Integer): Cardinal;
    function CurrentTurnOf(Player: TPlayer): Integer;
    function PlayedTurnsOf(Player: TPlayer): Integer;
    function PlayedTurns: Integer;
    function IsCurrentStillRunning: Boolean;
    property IsClosed: Boolean read fClosed;
    property InitialBoard: TGameboard read fInitialBoard;
    property Current: Integer read fCurrent write SetCurrent;
    property Players[Index: Integer]: TPlayer read GetPlayers;
    property Items[Index: Integer]: TGameHistoryRec read GetItems; default;
    property OnChange: TNotifyEvent read fOnChange write fOnChange;
  end;

  TSelectionOption = (soAutoComplete, soAutoPlayStep, soAutoPlayComplete);
  TSelectionOptions = set of TSelectionOption;

  TBoardLabel = (blLeft, blTop, blRight, blBottom);
  TBoardLabels = set of TBoardLabel;

  TSquareLabel = (slNone, slCenter, slLeftCenter, slLeftTop, slTopCenter,
    slRightTop, slRightCenter, slRightBottom, slBottomCenter, slLeftBottom);

  TGameboardMode = (gbmSetup, gbmPlay);

  TSquareDrawState = (sdsSelected, sdsMarked, sdsFocused, sdsNoCache,
    sdsReserved1, sdsReserved2, sdsReserved3);
  TSquareDrawStates = set of TSquareDrawState;

  TSquareDragOverEvent = procedure(Sender: TObject; Source, Target: TSquare; var Accept: Boolean) of object;
  TSquareDragDropEvent = procedure(Sender: TObject; Source, Target: TSquare) of object;
  TSquareInfoTipEvent = procedure(Sender: TObject; Square: TSquare; var InfoTip: String) of object;
  TSquareActionEvent = procedure(Sender: TObject; Square: TSquare; var Handled: Boolean) of object;
  TSquareEvent = procedure(Sender: TObject; Square: TSquare) of object;
  TMoveVarEvent = procedure(Sender: TObject; var Index: Integer) of object;
  TMoveEvent = procedure(Sender: TObject; Index: Integer) of object;

  TSquareDrawInfo = record
    States: TSquareDrawStates;
    ImageIndex: Integer;
  end;

  TGameboardControl = class(TCustomControl)
  private
    fBoardBorderColor: TColor;
    fBoardMarginColor: TColor;
    fSquareBorderColor: TColor;
    fSquareDarkSelectColor: TColor;
    fSquareDarkMarkColor: TColor;
    fSquareDarkColor: TColor;
    fSquareLightSelectColor: TColor;
    fSquareLightMarkColor: TColor;
    fSquareLightColor: TColor;
    fSquareDarkSelectPattern: TPicture;
    fSquareDarkMarkPattern: TPicture;
    fSquareDarkPattern: TPicture;
    fSquareLightSelectPattern: TPicture;
    fSquareLightMarkPattern: TPicture;
    fSquareLightPattern: TPicture;
    fBoardLabels: TBoardLabels;
    fBoardLabelsOffset: Integer;
    fSquareLabel: TSquareLabel;
    fSquareLabelOffset: Integer;
    fBoardAngle: Integer;
    fInvertSquares: Boolean;
    fCenter: Boolean;
    fMarkSelectableSquares: Boolean;
    fMarkerSize: Integer;
    fMarkerOffset: Integer;
    fSquareSize: Integer;
    fFixedSquareSize: Boolean;
    fBoardMarginLeft: Integer;
    fBoardMarginTop: Integer;
    fBoardMarginRight: Integer;
    fBoardMarginBottom: Integer;
    fCursor: TCursor;
    fThinkCursor: TCursor;
    fAnimateMoves: Boolean;
    fAnimating: Boolean;
    fComputerPlayer: TPlayer;
    fPlaying: Boolean;
    fMoving: Boolean;
    fThinking: Boolean;
    fDragedMove: Boolean;
    fPaused: Boolean;
    fPausedByHistory: Boolean;
    fBoardMode: TGameboardMode;
    fSelection: TMovementPath;
    fSelectionOptions: TSelectionOptions;
    fBoardBounds: TRect;
    fFullBoardBounds: TRect;
    fPieceImages: TImageList;
    fBoardBitmap: TBitmap;
    fCurrentHistoryRec: TGameHistoryRec;
    fHistory: TGameHistory;
    fGameboard: TGameboard;
    fSolver: TGameSolver;
    fAnimator: TAnimator;
    fDraggingSquare: Integer;
    fClickingSquare: Integer;
    fAutoDragDrop: Boolean;
    fUpdateCount: Integer;
    fIsFullBoardReady: Boolean;
    fIsBoardReady: Boolean;
    fMoveStartTime: Cardinal;
    fMoveElapsedTime: Cardinal;
    fModified: Boolean;
    fRecordFullMovesOnly: Boolean;
    fOnHistoryChange: TNotifyEvent;
    fOnTurnChange: TNotifyEvent;
    fOnBoardChange: TNotifyEvent;
    fOnAnimateBegin: TNotifyEvent;
    fOnAnimateEnd: TNotifyEvent;
    fOnThinkBegin: TNotifyEvent;
    fOnThinkEnd: TNotifyEvent;
    fOnComputerError: TNotifyEvent;
    fOnComputerMove: TMoveVarEvent;
    fOnMoveBegin: TNotifyEvent;
    fOnMoveEnd: TNotifyEvent;
    fOnGameEnded: TNotifyEvent;
    fOnSquareClick: TSquareActionEvent;
    fOnSquareDragStart: TNotifyEvent;
    fOnSquareDragEnd: TNotifyEvent;
    fOnSquareDragDrop: TSquareDragDropEvent;
    fOnSquareDragOver: TSquareDragOverEvent;
    fOnSquareInfoTip: TSquareInfoTipEvent;
    fOnSquareResize: TNotifyEvent;
    fOnSelectionChange: TNotifyEvent;
    fOnPlayingStateChange: TNotifyEvent;
    fOnBoardModeChange: TNotifyEvent;
    fOnPauseStateChange: TNotifyEvent;
    DragStartPt: TPoint;
    CheckDragStartPt: Boolean;
    LastDrawInfo: array of TSquareDrawInfo;
    procedure SetBoardBorderColor(Value: TColor);
    procedure SetBoardMarginColor(Value: TColor);
    procedure SetSquareBorderColor(Value: TColor);
    procedure SetSquareDarkColor(Value: TColor);
    procedure SetSquareDarkMarkColor(Value: TColor);
    procedure SetSquareDarkSelectColor(Value: TColor);
    procedure SetSquareLightColor(Value: TColor);
    procedure SetSquareLightMarkColor(Value: TColor);
    procedure SetSquareLightSelectColor(Value: TColor);
    procedure SetSquareDarkPattern(Value: TPicture);
    procedure SetSquareDarkMarkPattern(Value: TPicture);
    procedure SetSquareDarkSelectPattern(Value: TPicture);
    procedure SetSquareLightPattern(Value: TPicture);
    procedure SetSquareLightMarkPattern(Value: TPicture);
    procedure SetSquareLightSelectPattern(Value: TPicture);
    procedure SetBoardAngle(Value: Integer);
    procedure SetInvertSquares(Value: Boolean);
    procedure SetCenter(Value: Boolean);
    procedure SetMarkSelectableSquares(Value: Boolean);
    procedure SetMarkerSize(Value: Integer);
    procedure SetMarkerOffset(Value: Integer);
    procedure SetSquareSize(Value: Integer);
    procedure SetFixedSquareSize(Value: Boolean);
    procedure SetBoardMarginLeft(Value: Integer);
    procedure SetBoardMarginTop(Value: Integer);
    procedure SetBoardMarginRight(Value: Integer);
    procedure SetBoardMarginBottom(Value: Integer);
    procedure SetBoardLabels(Value: TBoardLabels);
    procedure SetBoardLabelsOffset(Value: Integer);
    procedure SetSquareLabel(Value: TSquareLabel);
    procedure SetSquareLabelOffset(Value: Integer);
    procedure SetPaused(Value: Boolean);
    procedure SetBoardMode(Value: TGameboardMode);
    procedure SetCursor(Value: TCursor);
    procedure SetThinkCursor(Value: TCursor);
    function GetBoardAsText: String;
    procedure SetBoardAsText(const Value: String);
    function IsGameboardStored: Boolean;
    procedure SetComputerPlayer(Value: TPlayer);
    procedure SetCurrentPlayer(Value: TPlayer);
    function GetCurrentPlayer: TPlayer;
    function GetWinnerPlayer: TPlayer;
    function GetLooserPlayer: TPlayer;
    function GetOpenerPlayer: TPlayer;
    function GetLegalMovesCount: Integer;
    function GetCurrentTurn: Integer;
    function GetCurrentTime: Cardinal;
    function GetElapsedTime: Cardinal;
    function GetPlayedTurns: Integer;
    function GetPlayedTime: Cardinal;
    function GetCurrentTurnOf(Player: TPlayer): Integer;
    function GetCurrentTimeOf(Player: TPlayer): Cardinal;
    function GetPlayedTurnsOf(Player: TPlayer): Integer;
    function GetPlayedTimeOf(Player: TPlayer): Cardinal;
    procedure SetThinkPriority(Value: TThreadPriority);
    function GetThinkPriority: TThreadPriority;
    procedure SetSearchDepth(Value: Byte);
    function GetSearchDepth: Byte;
    procedure SetMaxMemory(Value: Cardinal);
    function GetMaxMemory: Cardinal;
    procedure SetAnimateDelay(Value: Byte);
    function GetAnimateDelay: Byte;
    procedure SetAnimateStep(Value: Byte);
    function GetAnimateStep: Byte;
    function GetSquareBounds(Square: TSquare): TRect;
    procedure SetSelection(const Value: TMovementPath);
    function GetSelSquares(Index: Integer): TSquare;
    function GetSelSquareCount: Integer;
    function GetNumOfSquares: Integer;
    procedure GameChange(Sender: TObject);
    procedure GameTurnChange(Sender: TObject);
    procedure HistoryChange(Sender: TObject);
    procedure StyleChange(Sender: TObject);
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
  protected
    procedure WndProc(var Message: TMessage); override;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    procedure Loaded; override;
    procedure Click; override;
    procedure Paint; override;
    procedure Resize; override;
    procedure AdjustSize; override;
    function CanResize(var NewWidth, NewHeight: Integer): Boolean; override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  protected
    procedure DoSquareClick(Square: TSquare; var Handled: Boolean); virtual;
    procedure DoSquareInfoTip(Square: TSquare; var InfoTip: String); virtual;
    procedure DoSquareDragOver(Source, Target: TSquare; var Accept: Boolean); virtual;
    procedure DoSquareDragDrop(Source, Target: TSquare); virtual;
    procedure DoSquareDragStart; virtual;
    procedure DoSquareDragEnd; virtual;
    procedure DoSquareResize; virtual;
    procedure DoHistoryChange; virtual;
    procedure DoBoardChange; virtual;
    procedure DoTurnChange; virtual;
    procedure DoGameEnded; virtual;
    procedure DoAnimateBegin; virtual;
    procedure DoAnimateEnd; virtual;
    procedure DoThinkBegin; virtual;
    procedure DoThinkEnd; virtual;
    procedure DoComputerError; virtual;
    procedure DoComputerMove(var Index: Integer); virtual;
    procedure DoMoveBegin; virtual;
    procedure DoMoveEnd; virtual;
    procedure DoSelectionChange; virtual;
    procedure DoPlayingStateChange; virtual;
    procedure DoBoardModeChange; virtual;
    procedure DoPauseStateChange; virtual;
  protected
    function GetGameboardClass: TGameboardClass; virtual; abstract;
    function IsPathComplete(const Path: array of TSquare): Boolean; virtual; abstract;
    function CanPlayPath(const Path: array of TSquare): Boolean; virtual; abstract;
    function PlayPath(const Path: array of TSquare): Boolean; virtual; abstract;
    function CompletePath(var Path: TMovementPath): Boolean; virtual; abstract;
  protected
    function IsUpdateLocked: Boolean;
    function GetNumOfRows: Integer; virtual;
    function GetNumOfCols: Integer; virtual;
    function BestClientSize(ASquareSize: Integer): TSize; virtual;
    function BestSquareSize(AWidth, AHeight: Integer): Integer; virtual;
    function CoordinateToSquare(Row, Col: Integer): TSquare; virtual;
    function SquareToCoordinate(Square: TSquare; out Row, Col: Integer): Boolean; virtual;
    function GetRowLabel(Row: Integer): String; virtual;
    function GetColLabel(Col: Integer): String; virtual;
    function GetSquareLabel(Square: TSquare): String; virtual;
    procedure SetImageIndexOfSquare(Square: TSquare; ImageIndex: Integer); virtual;
    function GetImageIndexOfSquare(Square: TSquare): Integer; virtual;
    function GetStateOfSquare(Square: TSquare): TSquareDrawStates; virtual;
    function IsLegalSquare(Square: TSquare): Boolean; virtual;
    function IsLogicalDarkSquare(Square: TSquare): Boolean; virtual;
    function IsVisualDarkSquare(Square: TSquare): Boolean; virtual;
    function IsSquareSelected(Square: TSquare): Boolean; virtual;
    function CanSelectSquare(Square: TSquare): Boolean; virtual;
    function SelectSquare(Square: TSquare): Boolean; virtual;
    function DeselectSquare(Square: TSquare): Boolean; virtual;
    procedure AdjustBoardBounds(ForceAdjust: Boolean); virtual;
    procedure AdjustBoardOffset; virtual;
    procedure AdjustCoordinates(var Row, Col: Integer; Angle: Integer); virtual;
    function UpdateFullBoard: Boolean; virtual;
    function UpdateBoard(AllSquares: Boolean): Boolean; virtual;
    function UpdateSquare(Square: TSquare; ForceUpdate: Boolean): Boolean; virtual;
    procedure DrawBoard(Canvas: TCanvas; const Rect: TRect); virtual;
    procedure DrawBoardLabels(Canvas: TCanvas; const Rect: TRect); virtual;
    procedure DrawRowLabel(Canvas: TCanvas; Row: Integer; TextAlign: TAlign;
      const Rect: TRect); virtual;
    procedure DrawColLabel(Canvas: TCanvas; Col: Integer; TextAlign: TAlign;
      const Rect: TRect); virtual;
    procedure FlushSquare(Square: TSquare; ImageIndex: Integer); virtual;
    procedure DrawSquare(Canvas: TCanvas; Square: TSquare; Rect: TRect;
      ImageIndex: Integer; States: TSquareDrawStates); virtual;
    procedure DrawStates(Canvas: TCanvas; Square: TSquare; const Rect: TRect;
      ImageIndex: Integer; States: TSquareDrawStates); virtual;
    procedure DrawPiece(Canvas: TCanvas; Square: TSquare; const Rect: TRect;
      ImageIndex: Integer; States: TSquareDrawStates); virtual;
    procedure DrawLabel(Canvas: TCanvas; Square: TSquare; const Rect: TRect;
      States: TSquareDrawStates); virtual;
    procedure ExtractPiecesImages(BitmapOfPieces: TBitmap;
      Rows, Cols: Integer); virtual;
    procedure StorePath(const Path: array of TSquare); virtual;
    procedure ResetTimer(ResetOld, SaveCurrent: Boolean); virtual;
    function CheckPlaying(SaveTime: Boolean; StopAnyway: Boolean = False): Boolean; virtual;
    function CanThinkAboutMove: Boolean; virtual;
    function ThinkAboutMove: Boolean; virtual;
    function StopThinking(IgnoreSolution: Boolean): Boolean; virtual;
    function AnimateForward(const Path: array of TSquare): Boolean; virtual;
    function AnimateBackward(const Path: array of TSquare): Boolean; virtual;
    procedure StartNewSelection(Square: TSquare); virtual;
    function StartDrag(X, Y: Integer; TestOnly: Boolean): Boolean; virtual;
    function CanDrop(SourceSquare: TSquare; X, Y: Integer): Boolean; virtual;
    function StopDrag(X, Y: Integer; Cancelled: Boolean): Boolean; virtual;
    function ApplyComputerMove(Index: Integer): Boolean; virtual;
    function ApplyLegalMove(Index: Integer): Boolean; virtual;
    function StoreCurrentHistoryRecord: Boolean; virtual;
    function RetrieveHistory(Index: Integer): Boolean; virtual;
    function HistoryOf(Player: TPlayer; TurnNo: Integer): TGameHistoryRec; virtual;
    procedure InitializeMovement; virtual;
    procedure FinalizeMovement; virtual;
  protected
    property BoardBitmap: TBitmap read fBoardBitmap;
    property PieceImages: TImageList read fPieceImages;
    property Solver: TGameSolver read fSolver;
    property Animator: TAnimator read fAnimator;
    property GameboardClass: TGameboardClass read GeTGameboardClass;
    property Gameboard: TGameboard read fGameboard;
    property History: TGameHistory read fHistory;
    property CurrentHistoryRec: TGameHistoryRec read fCurrentHistoryRec;
    property NumOfRows: Integer read GetNumOfRows;
    property NumOfCols: Integer read GetNumOfCols;
    property NumOfSquares: Integer read GetNumOfSquares;
    property DragedMove: Boolean read fDragedMove;
    property PausedByHistory: Boolean read fPausedByHistory;
    property IsFullBoardReady: Boolean read fIsFullBoardReady;
    property IsBoardReady: Boolean read fIsBoardReady;
    property ClickingSquare: TSquare read fClickingSquare;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure StartNewGame; virtual;
    procedure ClearBoard; virtual;
    procedure BeginUpdate; virtual;
    procedure EndUpdate; virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToFile(const FileName: String;
      const Signature: String = ''); virtual;
    procedure LoadFromFile(const FileName: String;
      const Signature: String = ''); virtual;
    function IsGameEnded: Boolean; virtual;
    function IsGameDraw: Boolean; virtual;
    function CanMoveBackward: Boolean; virtual;
    function MoveBackward: Boolean; virtual;
    function MoveToFirst: Boolean; virtual;
    function CanMoveForward: Boolean; virtual;
    function MoveForward: Boolean; virtual;
    function MoveToLast: Boolean; virtual;
    function CanRetrieve(Player: TPlayer; TurnNo: Integer): Boolean; virtual;
    function Retrieve(Player: TPlayer; TurnNo: Integer): Boolean; virtual;
    procedure ClearHistory; virtual;
    function IsComputerTurn: Boolean; virtual;
    function PlayComputerTurnNow: Boolean; virtual;
    function CancelThinking: Boolean; virtual;
    function SuggestMove(ASearchDepth: ShortInt = -1): Integer; virtual;
    function ClientToSquare(X, Y: Integer): Integer; virtual;
    function ClearSelection: Boolean; virtual;
    function CompleteSelection: Boolean; virtual;
    function IsSelectionComplete: Boolean; virtual;
    function CanPlaySelection: Boolean; virtual;
    function PlaySelection: Boolean; virtual;
    function MoveOf(Player: TPlayer; TurnNo: Integer): String; virtual; abstract;
    function TimeOf(Player: TPlayer; TurnNo: Integer): Cardinal; virtual;
    function PlayerScore: Integer; virtual;
    function MoveScore(MoveIndex: Integer): Integer; virtual;
    property BoardBounds: TRect read fBoardBounds;
    property FullBoardBounds: TRect read fFullBoardBounds;
    property SquareBounds[Square: TSquare]: TRect read GetSquareBounds;
    property DraggingSquare: TSquare read fDraggingSquare;
    property Animating: Boolean read fAnimating;
    property Moving: Boolean read fMoving;
    property Thinking: Boolean read fThinking;
    property Playing: Boolean read fPlaying;
    property ElapsedTime: Cardinal read GetElapsedTime;
    property CurrentTurn: Integer read GetCurrentTurn;
    property CurrentTime: Cardinal read GetCurrentTime;
    property PlayedTurns: Integer read GetPlayedTurns;
    property PlayedTime: Cardinal read GetPlayedTime;
    property CurrentTurnOf[Player: TPlayer]: Integer read GetCurrentTurnOf;
    property CurrentTimeOf[Player: TPlayer]: Cardinal read GetCurrentTimeOf;
    property PlayedTurnsOf[Player: TPlayer]: Integer read GetPlayedTurnsOf;
    property PlayedTimeOf[Player: TPlayer]: Cardinal read GetPlayedTimeOf;
    property LegalMovesCount: Integer read GetLegalMovesCount;
    property WinnerPlayer: TPlayer read GetWinnerPlayer;
    property LooserPlayer: TPlayer read GetLooserPlayer;
    property OpenerPlayer: TPlayer read GetOpenerPlayer;
    property CurrentPlayer: TPlayer read GetCurrentPlayer write SetCurrentPlayer;
    property Selection: TMovementPath read fSelection write SetSelection;
    property SelSquares[Index: Integer]: TSquare read GetSelSquares;
    property SelSquareCount: Integer read GetSelSquareCount;
    property Modified: Boolean read fModified write fModified;
  protected
    property AnimateDelay: Byte read GetAnimateDelay write SetAnimateDelay default 10;
    property AnimateMoves: Boolean read fAnimateMoves write fAnimateMoves default True;
    property AnimateStep: Byte read GetAnimateStep write SetAnimateStep default 1;
    property AutoDragDrop: Boolean read fAutoDragDrop write fAutoDragDrop default True;
    property BoardAngle: Integer read fBoardAngle write SetBoardAngle default 0;
    property BoardAsText: String read GetBoardAsText write SetBoardAsText stored IsGameboardStored;
    property BoardBorderColor: TColor read fBoardBorderColor write SetBoardBorderColor default clBlack;
    property BoardLabels: TBoardLabels read fBoardLabels write SetBoardLabels default [];
    property BoardLabelsOffset: Integer read fBoardLabelsOffset write SetBoardLabelsOffset default 5;
    property BoardMarginBottom: Integer read fBoardMarginBottom write SetBoardMarginBottom default 0;
    property BoardMarginColor: TColor read fBoardMarginColor write SetBoardMarginColor default clWhite;
    property BoardMarginLeft: Integer read fBoardMarginLeft write SetBoardMarginLeft default 0;
    property BoardMarginRight: Integer read fBoardMarginRight write SetBoardMarginRight default 0;
    property BoardMarginTop: Integer read fBoardMarginTop write SetBoardMarginTop default 0;
    property BoardMode: TGameboardMode read fBoardMode write SetBoardMode default gbmSetup;
    property Center: Boolean read fCenter write SetCenter default True;
    property ComputerPlayer: TPlayer read fComputerPlayer write SetComputerPlayer default NONE;
    property Cursor: TCursor read fCursor write SetCursor default crDefault;
    property DragCursor default crHandPoint;
    property FixedSquareSize: Boolean read fFixedSquareSize write SetFixedSquareSize default False;
    property InvertSquares: Boolean read fInvertSquares write SetInvertSquares default False;
    property MarkerOffset: Integer read fMarkerOffset write SetMarkerOffset default 5;
    property MarkerSize: Integer read fMarkerSize write SetMarkerSize default 2;
    property MarkSelectableSquares: Boolean read fMarkSelectableSquares write SetMarkSelectableSquares default True;
    property MaxMemory: Cardinal read GetMaxMemory write SetMaxMemory default 1048576; {1MB}
    property Paused: Boolean read fPaused write SetPaused default False;
    property RecordFullMovesOnly: Boolean read fRecordFullMovesOnly write fRecordFullMovesOnly default False;
    property SearchDepth: Byte read GetSearchDepth write SetSearchDepth default 8;
    property SelectionOptions: TSelectionOptions read fSelectionOptions write fSelectionOptions
      default [soAutoComplete, soAutoPlayComplete];
    property SquareBorderColor: TColor read fSquareBorderColor write SetSquareBorderColor default clBlack;
    property SquareDarkColor: TColor read fSquareDarkColor write SetSquareDarkColor default clBlack;
    property SquareDarkMarkColor: TColor read fSquareDarkMarkColor write SetSquareDarkMarkColor default clRed;
    property SquareDarkMarkPattern: TPicture read fSquareDarkMarkPattern write SetSquareDarkMarkPattern;
    property SquareDarkPattern: TPicture read fSquareDarkPattern write SetSquareDarkPattern;
    property SquareDarkSelectColor: TColor read fSquareDarkSelectColor write SetSquareDarkSelectColor default clRed;
    property SquareDarkSelectPattern: TPicture read fSquareDarkSelectPattern write SetSquareDarkSelectPattern;
    property SquareLabel: TSquareLabel read fSquareLabel write SetSquareLabel default slNone;
    property SquareLabelOffset: Integer read fSquareLabelOffset write SetSquareLabelOffset default 4;
    property SquareLightColor: TColor read fSquareLightColor write SetSquareLightColor default clWhite;
    property SquareLightMarkColor: TColor read fSquareLightMarkColor write SetSquareLightMarkColor default clRed;
    property SquareLightMarkPattern: TPicture read fSquareLightMarkPattern write SetSquareLightMarkPattern;
    property SquareLightPattern: TPicture read fSquareLightPattern write SetSquareLightPattern;
    property SquareLightSelectColor: TColor read fSquareLightSelectColor write SetSquareLightSelectColor default clRed;
    property SquareLightSelectPattern: TPicture read fSquareLightSelectPattern write SetSquareLightSelectPattern;
    property SquareSize: Integer read fSquareSize write SetSquareSize stored fFixedSquareSize;
    property ThinkCursor: TCursor read fThinkCursor write SetThinkCursor default crAppStart;
    property ThinkPriority: TThreadPriority read GetThinkPriority write SetThinkPriority default tpNormal;
    property OnAnimateBegin: TNotifyEvent read fOnAnimateBegin write fOnAnimateBegin;
    property OnAnimateEnd: TNotifyEvent read fOnAnimateEnd write fOnAnimateEnd;
    property OnBoardChange: TNotifyEvent read fOnBoardChange write fOnBoardChange;
    property OnBoardModeChange: TNotifyEvent read fOnBoardModeChange write fOnBoardModeChange;
    property OnComputerError: TNotifyEvent read fOnComputerError write fOnComputerError;
    property OnComputerMove: TMoveVarEvent read fOnComputerMove write fOnComputerMove;
    property OnGameEnded: TNotifyEvent read fOnGameEnded write fOnGameEnded;
    property OnHistoryChange: TNotifyEvent read fOnHistoryChange write fOnHistoryChange;
    property OnMoveBegin: TNotifyEvent read fOnMoveBegin write fOnMoveEnd;
    property OnMoveEnd: TNotifyEvent read fOnMoveEnd write fOnMoveEnd;
    property OnPauseStateChange: TNotifyEvent read fOnPauseStateChange write fOnPauseStateChange;
    property OnPlayingStateChange: TNotifyEvent read fOnPlayingStateChange write fOnPlayingStateChange;
    property OnSelectionChange: TNotifyEvent read fOnSelectionChange write fOnSelectionChange;
    property OnSquareClick: TSquareActionEvent read fOnSquareClick write fOnSquareClick;
    property OnSquareDragDrop: TSquareDragDropEvent read fOnSquareDragDrop write fOnSquareDragDrop;
    property OnSquareDragEnd: TNotifyEvent read fOnSquareDragEnd write fOnSquareDragEnd;
    property OnSquareDragOver: TSquareDragOverEvent read fOnSquareDragOver write fOnSquareDragOver;
    property OnSquareDragStart: TNotifyEvent read fOnSquareDragStart write fOnSquareDragStart;
    property OnSquareInfoTip: TSquareInfoTipEvent read fOnSquareInfoTip write fOnSquareInfoTip;
    property OnSquareResize: TNotifyEvent read fOnSquareResize write fOnSquareResize;
    property OnThinkBegin: TNotifyEvent read fOnThinkBegin write fOnThinkBegin;
    property OnThinkEnd: TNotifyEvent read fOnThinkEnd write fOnThinkEnd;
    property OnTurnChange: TNotifyEvent read fOnTurnChange write fOnTurnChange;
  end;

function Opponent(Player: TPlayer): TPlayer; {$IFDEF COMPILER2005_UP} inline; {$ENDIF}

procedure InitZobritsNumbers(var ZobristNumbers; Count: Integer);

procedure SaveStringToStream(const Str: String; Stream: TStream);
function LoadStringFromStream(Stream: TStream): String;

function WaitForSyncObject(SyncObject: THandle; Timeout: Cardinal; BlockInput: Boolean): Cardinal;

function ProcessMessage: Boolean;
procedure ProcessMessages(Timeout: DWORD);

implementation

uses
  Math;

const
  FileVersion = '[3]';

{ Global Helper Functions }

function Opponent(Player: TPlayer): TPlayer;
begin
  Result := Player xor BOTH_PLAYERS;
end;

procedure InitZobritsNumbers(var ZobristNumbers; Count: Integer);
var
  Z: ^TPositionKey;
begin
  RandSeed := 2473;
  Z := @ZobristNumbers;
  while Count > 0 do
  begin
    Z^ := (TPositionKey(Random(32768)) shl 00)
      xor (TPositionKey(Random(32768)) shl 15)
      xor (TPositionKey(Random(32768)) shl 30)
      xor (TPositionKey(Random(32768)) shl 45)
      xor (TPositionKey(Random(32768)) shl 60);
    Inc(Z);
    Dec(Count);
  end;
  Randomize;
end;

procedure SaveStringToStream(const Str: String; Stream: TStream);
var
  Size: Integer;
begin
  Size := Length(Str);
  Stream.Write(Size, SizeOf(Size));
  Stream.Write(PChar(Str)^, Size * SizeOf(Char));
end;

function LoadStringFromStream(Stream: TStream): String;
var
  Size: Integer;
begin
  Stream.Read(Size, SizeOf(Size));
  SetString(Result, nil, Size);
  Stream.Read(PChar(Result)^, Size * SizeOf(Char));
end;

function WaitForSyncObject(SyncObject: THandle; Timeout: Cardinal; BlockInput: Boolean): Cardinal;
var
  Msg: TMsg;
  StartTime: Cardinal;
  EllapsedTime: Cardinal;
  EventMask: Cardinal;
  Handle: THandle;
begin
  EventMask := QS_ALLINPUT;
  if BlockInput then
    EventMask := EventMask and not QS_INPUT;
  Handle := SyncObject;
  if (SyncObject = GetCurrentThread) or (SyncObject = GetCurrentProcess) then
    DuplicateHandle(GetCurrentProcess, SyncObject, GetCurrentProcess, @Handle, SYNCHRONIZE, False, 0);
  repeat
    StartTime := GetTickCount;
    Result := MsgWaitForMultipleObjects(1, Handle, False, Timeout, EventMask);
    if Result = WAIT_OBJECT_0 + 1 then
    begin
      while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) do
      begin
        if not BlockInput or not (
         ((Msg.message >= WM_KEYFIRST) and (Msg.message <= WM_KEYLAST)) or
         ((Msg.message >= WM_MOUSEFIRST) and (Msg.message <= WM_MOUSELAST))) then
        begin
          TranslateMessage(Msg);
          DispatchMessage(Msg);
          if Msg.message = WM_QUIT then
          begin
            Result := WAIT_FAILED;
            Break;
          end;
        end;
      end;
      if Timeout <> INFINITE then
      begin
        EllapsedTime := GetTickCount - StartTime;
        if EllapsedTime < Timeout then
          Dec(Timeout, EllapsedTime)
        else
          Timeout := 0;
      end;
    end;
  until Result <> WAIT_OBJECT_0 + 1;
  if SyncObject <> Handle then
    CloseHandle(Handle);
end;

function ProcessMessage: Boolean;
var
  Msg: TMsg;
begin
  Result := False;
  if PeekMessage(Msg, 0, 0, 0, PM_REMOVE) then
  begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
    Result := True;
  end;
end;

procedure ProcessMessages(Timeout: DWORD);
var
  StopTime: DWORD;
begin
  StopTime := GetTickCount + Timeout;
  repeat
    ProcessMessage;
  until GetTickCount > StopTime;
end;

{ TGameboard }

constructor TGameboard.Create;
begin
end;

procedure TGameboard.SetPlayer(Value: TPlayer);
begin
  if Player <> Value then
  begin
    fPlayer := Value;
    DoChange;
  end;
end;

function TGameboard.GetEndState: TGameEndState;
begin
  Result := gesUnknown;
  if LegalMovesCount = 0 then
    case Player of
      LIGHT_PLAYER: Result := gesDarkWon;
      DARK_PLAYER: Result := gesLightWon;
    end;
end;

function TGameboard.GetWinner: TPlayer;
begin
  case EndState of
    gesLightWon: Result := LIGHT_PLAYER;
    gesDarkWon: Result := DARK_PLAYER;
  else
    Result := NONE;
  end;
end;

function TGameboard.GetLooser: TPlayer;
begin
  case EndState of
    gesLightWon: Result := DARK_PLAYER;
    gesDarkWon: Result := LIGHT_PLAYER;
  else
    Result := NONE;
  end;
end;

function TGameboard.IsUpdateLocked: Boolean;
begin
  Result := (fUpdateCount <> 0);
end;

procedure TGameboard.DoChange;
begin
  if fUpdateCount = 0 then
  begin
    fUpdated := False;
    if Assigned(OnChange) then
      OnChange(Self);
  end
  else
    fUpdated := True;
end;

procedure TGameboard.DoTurn;
begin
  if Assigned(OnTurn) then
    OnTurn(Self);
end;

procedure TGameboard.BeginUpdate;
begin
  Inc(fUpdateCount);
end;

procedure TGameboard.EndUpdate;
begin
  Dec(fUpdateCount);
  if (fUpdateCount = 0) and fUpdated then
    DoChange;
end;

procedure TGameboard.ClearBoard;
begin
  BeginUpdate;
  try
    BoardAsText := '';
    Player := NONE;
  finally
    EndUpdate;
  end;
end;

procedure TGameboard.ChangeSide;
begin
  Player := Opponent(Player);
  DoTurn;
end;

procedure TGameboard.SaveToStream(Stream: TStream);
begin
  Stream.Write(fPlayer, SizeOf(fPlayer));
end;

procedure TGameboard.LoadFromStream(Stream: TStream);
begin
  Stream.Read(fPlayer, SizeOf(fPlayer));
  DoChange;
end;

procedure TGameboard.Assign(Source: TPersistent);
var
  Stream: TMemoryStream;
begin
  if Source is TGameboard then
  begin
    Stream := TMemoryStream.Create;
    try
      TGameboard(Source).SaveToStream(Stream);
      Stream.Seek(0, soFromBeginning);
      Self.LoadFromStream(Stream);
    finally
      Stream.Free;
    end;
  end
  else
    inherited Assign(Source);
end;

procedure TGameboard.InitializeSearch;
begin
end;

procedure TGameboard.FinalizeSearch;
begin
end;

function TGameboard.SearchForBestMove(SearchDepth: Byte;
  MaxMemory: Cardinal; pControl: PSearchControl): Integer;
var
  Control: TSearchControl;
begin
  if SearchDepth = 0 then
    Result := -1
  else
  begin
    if SearchDepth > MAX_SEARCH_DEPTH then
      SearchDepth := MAX_SEARCH_DEPTH;
    InitializeSearch;
    try
      if not Assigned(pControl) then
      begin
        Control := scContinue;
        pControl := @Control;
      end;
      Result := PerformSearch(SearchDepth, MaxMemory, pControl^)
    finally
      FinalizeSearch;
    end;
  end;
end;

function TGameboard.PerformSearch(SearchDepth: Integer; MaxMemory: Cardinal;
  var Control: TSearchControl): Integer;
var
  Boards: array of TGameboard;
  PositionTable: TPositionTable;

  procedure AllocateDepthBoards;
  var
    I: Integer;
  begin
    SetLength(Boards, SearchDepth);
    for I := 0 to SearchDepth - 1 do
      Boards[I] := TGameboardClass(Self.ClassType).Create;
  end;

  procedure DeallocateDepthBoards;
  var
    I: Integer;
  begin
    for I := 0 to SearchDepth - 1 do
      Boards[I].Free;
    SetLength(Boards, 0);
  end;

  function AlphaBeta(CurrentBoard: TGameboard; Depth: Integer; Alpha, Beta: Integer): Integer;
  var
    CurrentEndState: TGameEndState;
    NextBoard: TGameboard;
    Flag: TPositionFlag;
    BestMove: Integer;
    BestValue: Integer;
    Value: Integer;
    MoveCount: Integer;
    Index: Integer;
  begin
    BestMove := -1;
    {$IFDEF BOARD_DEBUG}
    CurrentBoard.DisplayDebugInformation(Format('C:\*.log', [ClassType.ClassName]), 'Delpth = %d', [Depth]);
    {$ENDIF}
    if not PositionTable.Retrieve(CurrentBoard.Key, Depth, Alpha, Beta, BestValue, BestMove) then
    begin
      Flag := tfExact;
      CurrentEndState := CurrentBoard.EndState;
      if CurrentEndState <> gesUnknown then
      begin
        if CurrentEndState = gesDraw then
          BestValue := 0
        else if (CurrentEndState = gesDarkWon) xor (CurrentBoard.Player = LIGHT_PLAYER) then
          BestValue := +(MAX_SCORE + Depth)
        else
          BestValue := -(MAX_SCORE + Depth);
      end
      else if (Depth = 0) or (Control <> scContinue) then
      begin
        if Control = scAbort then Abort;
        BestValue := CurrentBoard.Score;
      end
      else
      begin
        NextBoard := Boards[Depth - 1];
        MoveCount := CurrentBoard.LegalMovesCount;
        if BestMove >= MoveCount then
          BestMove := -1;
        Flag := tfUpper;
        BestValue := Alpha;
        for Index := 0 to MoveCount - 1 do
        begin
          if (Index = 0) and (BestMove >= 0) then
            NextBoard.Transpose(CurrentBoard, BestMove)
          else if Index = BestMove then
            NextBoard.Transpose(CurrentBoard, 0)
          else
            NextBoard.Transpose(CurrentBoard, Index);
          Value := -AlphaBeta(NextBoard, Depth - 1, -Beta, -Alpha);
          if Value >= Beta then
          begin
            BestValue := Value;
            BestMove := Index;
            Flag := tfLower;
            Break;
          end;
          if Value > Alpha then
          begin
            Alpha := Value;
            BestValue := Value;
            BestMove := Index;
            Flag := tfExact;
          end;
        end;
      end;
      PositionTable.Store(CurrentBoard.Key, Depth, Flag, BestValue, BestMove);
    end;
    Result := BestValue;
  end;

var
  Index: Integer;
  NextBoard: TGameboard;
  Value, BestValue: Integer;
begin
  if Self.LegalMovesCount = 1 then
    Result := 0
  else
  begin
    Result := -1;
    BestValue := -MAX_VALUE;
    AllocateDepthBoards;
    try
      PositionTable := GetPositionTable;
      PositionTable.UpdateSize(AvgMoveCount, SearchDepth, MaxMemory);
      NextBoard := Boards[SearchDepth - 1];
      for Index := 0 to LegalMovesCount - 1 do
      begin
        NextBoard.Transpose(Self, Index);
        Value := -AlphaBeta(NextBoard, SearchDepth - 1, -MAX_VALUE, MAX_VALUE);
        if Value > BestValue then
        begin
          BestValue := Value;
          Result := Index;
        end
        else if (Value = BestValue) and Odd(GetTickCount) then
          Result := Index;
      end;
    finally
      DeallocateDepthBoards;
    end;
  end;
end;

procedure TGameboard.DisplayDebugInformation(const FileName: String;
  const Fmt: String; const Args: array of const);
var
  CurrentPlayer: String;
  CurrentBoard: String;
  CurrentMoves: String;
  F: TextFile;
begin
  case Player of
    DARK_PLAYER: CurrentPlayer := 'BLACK';
    LIGHT_PLAYER: CurrentPlayer := 'WHITE';
  else
    CurrentPlayer := 'NONE';
  end;
  GetTextPresentation(CurrentBoard, CurrentMoves);
  if CurrentMoves = '' then
    CurrentMoves := 'No Moves Available';
  AssignFile(F, FileName);
  if not FileExists(FileName) then
    Rewrite(F)
  else
    Append(F);
  Writeln(F);
  if Fmt <> '' then
  begin
    Writeln(F, Format(Fmt, Args));
    Writeln(F);
  end;
  Writeln(F, CurrentBoard);
  Writeln(F);
  Writeln(F, CurrentPlayer, ' Moves');
  Writeln(F);
  Writeln(F, CurrentMoves);
  Writeln(F);
  Writeln(F, StringOfChar('-', 80));
  CloseFile(F);
end;

{ TPositionTable }

destructor TPositionTable.Destroy;
begin
  ReallocMem(fTable, 0);
  inherited Destroy;
end;

procedure TPositionTable.SetNumOfCells(Value: Cardinal);
begin
  if NumOfCells <> Value then
  begin
    fNumOfCells := Value;
    ReallocMem(fTable, Size);
    FillChar(fTable^, Size, 0);
  end;
end;

function TPositionTable.GetNumOfUsedCells: Cardinal;
var
  Node: PPositionNode;
  I: Integer;
begin
  Result := 0;
  Node := fTable;
  if Assigned(Node) then
    for I := 1 to NumOfCells do
    begin
      if Node^.Key <> 0 then
        Inc(Result);
      Inc(Node);
    end;
end;

function TPositionTable.GetSize: Cardinal;
begin
  Result := NumOfCells * SizeOf(TPositionNode);
end;

procedure TPositionTable.SetSize(Value: Cardinal);
begin
  if (Value mod SizeOf(TPositionNode)) <> 0 then
    NumOfCells := Value div SizeOf(TPositionNode) + 1
  else
    NumOfCells := Value div SizeOf(TPositionNode)
end;

function TPositionTable.GetUsedSize: Cardinal;
begin
  Result := NumOfUsedCells * SizeOf(TPositionNode);
end;

procedure TPositionTable.UpdateSize(AvgBranch: Cardinal; SearchDepth: Integer; MaxSize: Cardinal);
var
  NewNumOfCells: Cardinal;
  NewSize: Cardinal;
  I: Integer;
begin
  NewNumOfCells := 1;
  for I := 1 to SearchDepth do
    NewNumOfCells := NewNumOfCells * AvgBranch;
  NewSize := NewNumOfCells * SizeOf(TPositionNode);
  if NewSize < MaxSize then
    Size := NewSize
  else
    Size := MaxSize;
end;

type
  TKeySplit = packed record
    Hi, Lo: DWORD;
  end;

function TPositionTable.Find(const Key: TPositionKey;
  out Node: PPositionNode): Boolean;
begin
  Node := fTable;
  with TKeySplit(Key) do Inc(Node, Cardinal(Hi xor Lo) mod NumOfCells);
  Result := (Node^.Key = Key);
end;

function TPositionTable.Retrieve(const Key: TPositionKey; Depth: Integer;
  var Alpha, Beta: Integer; var Value: Integer; var BestMove: Integer): Boolean;
var
  Node: PPositionNode;
begin
  Result := False;
  if Find(Key, Node) then
  begin
    BestMove := Node^.BestMove;
    if (Node^.Depth_Flag shr 2) >= Depth then
    begin
      case TPositionFlag(Node^.Depth_Flag and 3) of
        tfExact:
        begin
          Value := Node^.Value;
          Result := True;
        end;
        tfUpper:
          if Node^.Value <= Alpha then
          begin
            Value := Alpha;
            Result := True;
          end;
        tfLower:
          if Node^.Value >= Beta then
          begin
            Value := Beta;
            Result := True;
          end;
      end;
    end;
  end;
end;

procedure TPositionTable.Store(const Key: TPositionKey; Depth: Integer;
  Flag: TPositionFlag; const Value: Integer; BestMove: Integer);
var
  Node: PPositionNode;
begin
  Find(Key, Node);
  Node^.Key := Key;
  Node^.Depth_Flag := (Depth shl 2) or Ord(Flag);
  Node^.Value := Value;
  Node^.BestMove := BestMove;
end;

function TPositionTable.BestMoveOf(const Key: TPositionKey): Integer;
var
  Node: PPositionNode;
begin
  if Find(Key, Node) then
    Result := Node^.BestMove
  else
    Result := -1;
end;

procedure TPositionTable.Clear;
begin
  FillChar(fTable^, Size, 0);
end;

procedure TPositionTable.LoadFromStream(Stream: TStream);
var
  Key: TPositionKey;
  Node: PPositionNode;
  Count: Cardinal;
begin
  Stream.Read(Count, SizeOf(Count));
  while Count > 0 do
  begin
    Stream.Read(Key, SizeOf(TPositionKey));
    if not Find(Key, Node) then
      Node^.Key := Key;
    Stream.Read(Pointer(Cardinal(Node) + SizeOf(TPositionKey))^,
      SizeOf(TPositionNode) - SizeOf(TPositionKey));
    Dec(Count);
  end;
end;

procedure TPositionTable.SaveToStream(Stream: TStream);
var
  FirstFullNode, Node: PPositionNode;
  SkipingEmptyNodes: Boolean;
  BytesNeeded, BytesAvailable: Cardinal;
  Count: Cardinal;
  I: Integer;
begin
  Count := NumOfUsedCells;
  Stream.Write(Count, SizeOf(Count));
  if Count > 0 then
  begin
    BytesAvailable := Stream.Size - Stream.Position;
    BytesNeeded := Count * SizeOf(TPositionNode);
    if BytesNeeded > BytesAvailable then
      Stream.Size := Stream.Size + (BytesNeeded - BytesAvailable);
    Node := fTable;
    FirstFullNode := Node;
    SkipingEmptyNodes := (Node^.Key = 0);
    for I := 1 to NumOfCells do
    begin
      if SkipingEmptyNodes then
      begin
        if Node^.Key <> 0 then
        begin
          FirstFullNode := Node;
          SkipingEmptyNodes := False;
        end;
      end
      else
      begin
        if Node^.Key = 0 then
        begin
          Stream.Write(FirstFullNode^, Cardinal(Node) - Cardinal(FirstFullNode));
          SkipingEmptyNodes := True;
        end;
      end;
      Inc(Node);
    end;
    if not SkipingEmptyNodes and (Node <> FirstFullNode) then
      Stream.Write(FirstFullNode^, Cardinal(Node) - Cardinal(FirstFullNode));
  end;
end;

{ TGameSolver }

constructor TGameSolver.Create;
begin
  inherited Create(True);
  IdleEvent := CreateEvent(nil, True, True, nil);
  TaskEvent := CreateEvent(nil, False, False, nil);
  fMaxMemory := 1048576; {1MB};
  fSearchDepth := 8;
  Resume;
end;

destructor TGameSolver.Destroy;
begin
  Terminate;
  WaitFor;
  CloseHandle(IdleEvent);
  CloseHandle(TaskEvent);
  inherited Destroy;
end;

procedure TGameSolver.Terminate;
begin
  State := scAbort;
  Priority := tpTimeCritical;
  inherited Terminate;
  SetEvent(TaskEvent);
end;

function TGameSolver.IsBusy: Boolean;
begin
  Result := (WaitForSingleObject(IdleEvent, 0) = WAIT_TIMEOUT);
end;

function TGameSolver.WaitForIdle: Boolean;
begin
  Result := (WaitForSingleObject(IdleEvent, INFINITE) = WAIT_OBJECT_0);
end;

function TGameSolver.Solve(AGameboard: TGameboard): Boolean;
begin
  Result := False;
  if WaitForSingleObject(IdleEvent, 0) = WAIT_OBJECT_0 then
  begin
    ResetEvent(IdleEvent);
    State := scContinue;
    if Assigned(Gameboard) then
      FreeAndNil(Gameboard);
    Gameboard := TGameboardClass(AGameboard.ClassType).Create;
    Gameboard.Assign(AGameboard);
    SetEvent(TaskEvent);
    Result := True;
  end;
end;

procedure TGameSolver.Execute;
var
  BestMove: Integer;
begin
  while not Terminated do
  begin
    SetEvent(IdleEvent);
    WaitForSingleObject(TaskEvent, INFINITE);
    if not Terminated and Assigned(Gameboard) then
    begin
      try
        if not Terminated and (fNotifyWnd <> 0) then
          PostMessage(fNotifyWnd, WM_SOLVER_START, 0, 0);
        if not Terminated then
        begin
          try
            BestMove := Gameboard.SearchForBestMove(SearchDepth, MaxMemory, @fState);
          except
            BestMove := -9999;
          end;
          if not Terminated and (fNotifyWnd <> 0) then
            PostMessage(fNotifyWnd, WM_SOLVER_STOP, BestMove, Ord(State));
        end;
      finally
        FreeAndNil(Gameboard);
      end;
    end;
  end;
end;

{ TAnimator }

constructor TAnimator.Create(AControl: TGameBoardControl);
begin
  inherited Create(True);
  IdleEvent := CreateEvent(nil, True, False, nil);
  TaskEvent := CreateEvent(nil, False, False, nil);
  Control := AControl;
  fAnimateDelay := 10;
  fAnimateStep := 1;
  Resume;
end;

destructor TAnimator.Destroy;
begin
  Terminate;
  WaitFor;
  CloseHandle(IdleEvent);
  CloseHandle(TaskEvent);
  inherited Destroy;
end;

procedure TAnimator.Terminate;
begin
  Cancelled := True;
  Priority := tpTimeCritical;
  inherited Terminate;
  SetEvent(TaskEvent);
end;

function TAnimator.IsBusy: Boolean;
begin
  Result := (WaitForSingleObject(IdleEvent, 0) = WAIT_TIMEOUT);
end;

function TAnimator.WaitForIdle(BlockInputs: Boolean): Boolean;
begin
  Result := (WaitForSyncObject(IdleEvent, INFINITE, BlockInputs) = WAIT_OBJECT_0);
end;

function TAnimator.Animate(const Path: array of TSquare; Reverse: Boolean): Boolean;
var
  I: Integer;
begin
  Result := False;
  if Length(Path) > 1 then
  begin
    WaitForIdle(True);
    ResetEvent(IdleEvent);
    Cancelled := False;
    SetLength(AnimationPath, Length(Path));
    for I := 0 to Length(Path) - 1 do
      if Reverse then
        AnimationPath[I] := Path[Length(Path) - I - 1]
      else
        AnimationPath[I] := Path[I];
    SetEvent(TaskEvent);
    WaitForIdle(False);
    Result := True;
  end;
end;

procedure TAnimator.Execute;
begin
  while not Terminated do
  begin
    SetEvent(IdleEvent);
    WaitForSingleObject(TaskEvent, INFINITE);
    if not Terminated and (Length(AnimationPath) > 1) then
    begin
      PostMessage(Control.Handle, WM_ANIMATOR_START, 0, 0);
      PerformAnimation(AnimationPath);
      PostMessage(Control.Handle, WM_ANIMATOR_STOP, 0, Ord(Cancelled));
    end;
  end;
end;

procedure TAnimator.PerformAnimation(const Path: array of TSquare);

  procedure MovePiece(const Pt: TPoint);
  var
    TriggerTime: Cardinal;
  begin
    TriggerTime := GetTickCount + AnimateDelay;
    Control.PieceImages.DragMove(Pt.X, Pt.Y);
    while (GetTickCount < TriggerTime) and not Cancelled do
      Sleep(0);
  end;

  procedure AnimateDiagonally(var SrcPt: TPoint; const DstPt: TPoint; sX, sY: Integer);
  begin
    while ((SrcPt.X <> DstPt.X) or (SrcPt.Y <> DstPt.Y)) and not Cancelled do
    begin
      if SrcPt.X <> DstPt.X then Inc(SrcPt.X, sX);
      if SrcPt.Y <> DstPt.Y then Inc(SrcPt.Y, sY);
      MovePiece(SrcPt);
    end;
  end;

  procedure AnimateHorizontally(var SrcPt: TPoint; const DstPt: TPoint; sX: Integer);
  begin
    while (SrcPt.X <> DstPt.X) and not Cancelled do
    begin
      Inc(SrcPt.X, sX);
      MovePiece(SrcPt);
    end;
  end;

  procedure AnimateVertically(var SrcPt: TPoint; const DstPt: TPoint; sY: Integer);
  begin
    while (SrcPt.Y <> DstPt.Y) and not Cancelled do
    begin
      Inc(SrcPt.Y, sY);
      MovePiece(SrcPt);
    end;
  end;

  procedure InitializeAnimate(Square: TSquare; out ImageIndex: Integer; out Rect: TRect);
  begin
    ImageIndex := Control.GetImageIndexOfSquare(Square);
    Rect := Control.SquareBounds[Square];
    Control.FlushSquare(Square, -1);
    Control.PieceImages.DragCursor := crNone;
    Control.PieceImages.SetDragImage(ImageIndex,
     (Control.PieceImages.Width - Control.SquareSize) div 2,
     (Control.PieceImages.Height - Control.SquareSize) div 2);
    Control.PieceImages.BeginDrag(Control.Handle, Rect.Left, Rect.Top);
    Control.PieceImages.ShowDragImage;
  end;

  procedure FinalizeAnimate(Square: TSquare; ImageIndex: Integer; const Rect: TRect);
  begin
    Control.PieceImages.HideDragImage;
    Control.PieceImages.EndDrag;
    Control.FlushSquare(Square, ImageIndex);
  end;

var
  Current: Integer;
  ImageIndex: Integer;
  Rect, NxtRect: TRect;
  dX, dY, sX, sY: Integer;
begin
  InitializeAnimate(Path[Low(Path)], ImageIndex, Rect);
  try
    for Current := Succ(Low(Path)) to High(Path) do
    begin
      if Path[Current] < 0 then
      begin
        FinalizeAnimate(Path[Current - 1], ImageIndex, Rect);
        InitializeAnimate(Path[Current], ImageIndex, Rect);
      end;
      NxtRect := Control.SquareBounds[Path[Current]];
      if not Cancelled then
      begin
        dX := NxtRect.Left - Rect.Left;
        dY := NxtRect.Top - Rect.Top;
        if dX < 0 then
          sX := -AnimateStep
        else
          sX := +AnimateStep;
        if dY < 0 then
          sY := -AnimateStep
        else
          sY := +AnimateStep;
        Inc(Rect.Left, dX mod sX);
        Inc(Rect.Top, dY mod sY);
        if Abs(dX) = Abs(dY) then
          AnimateDiagonally(Rect.TopLeft, NxtRect.TopLeft, sX, sY)
        else if dX = 0 then
          AnimateVertically(Rect.TopLeft, NxtRect.TopLeft, sY)
        else if dY = 0 then
          AnimateHorizontally(Rect.TopLeft, NxtRect.TopLeft, sX)
        else
        begin
          if Odd(GetTickCount) then
          begin
            AnimateHorizontally(Rect.TopLeft, NxtRect.TopLeft, sX);
            AnimateVertically(Rect.TopLeft, NxtRect.TopLeft, sY);
          end
          else
          begin
            AnimateVertically(Rect.TopLeft, NxtRect.TopLeft, sY);
            AnimateHorizontally(Rect.TopLeft, NxtRect.TopLeft, sX);
          end;
        end;
      end;
      Rect := NxtRect;
    end;
  finally
    FinalizeAnimate(Path[High(Path)], ImageIndex, Rect);
  end;
end;

procedure TAnimator.CancelAnimation;
begin
  Cancelled := True;
end;

{ TGameHistoryRec }

constructor TGameHistoryRec.Create;
begin
  fBoardAfterMove := TMemoryStream.Create;
end;

destructor TGameHistoryRec.Destroy;
begin
  fBoardAfterMove.Free;
  fMovementPath := nil;
  inherited Destroy;
end;

procedure TGameHistoryRec.Clear;
begin
  fBoardAfterMove.Clear;
  fMovementPath := nil;
  fElapsedTime := 0;
  fCompleted := False;
end;

procedure TGameHistoryRec.Assign(Source: TPersistent);
begin
  if Source is TGameHistoryRec then
  begin
    fBoardAfterMove.Clear;
    fBoardAfterMove.CopyFrom(TGameHistoryRec(Source).BoardAfterMove, 0);
    MovementPath := TGameHistoryRec(Source).MovementPath;
    ElapsedTime := TGameHistoryRec(Source).ElapsedTime;
    Completed := TGameHistoryRec(Source).Completed;
  end
  else if Source is TGameboard then
  begin
    fBoardAfterMove.Position := 0;
    TGameboard(Source).SaveToStream(fBoardAfterMove);
    fBoardAfterMove.Size := fBoardAfterMove.Position;
  end
  else
    inherited Assign(Source);
end;

procedure TGameHistoryRec.AssignTo(Dest: TPersistent);
begin
  if Dest is TGameboard then
  begin
    fBoardAfterMove.Position := 0;
    TGameboard(Dest).LoadFromStream(fBoardAfterMove);
  end
  else
    Dest.Assign(Self);
end;

procedure TGameHistoryRec.LoadFromStream(Stream: TStream);
var
  Count: Integer;
begin
  Stream.Read(Count, SizeOf(Count));
  fBoardAfterMove.Size := Count;
  if Count > 0 then
    Stream.Read(fBoardAfterMove.Memory^, Count);
  Stream.Read(Count, SizeOf(Count));
  SetLength(fMovementPath, Count);
  if Count > 0 then
    Stream.Read(fMovementPath[0], Count * SizeOf(fMovementPath[0]));
  Stream.Read(fElapsedTime, SizeOf(fElapsedTime));
  Stream.Read(fCompleted, SizeOf(fCompleted));
end;

procedure TGameHistoryRec.SaveToStream(Stream: TStream);
var
  Count: Integer;
begin
  Count := fBoardAfterMove.Size;
  Stream.Write(Count, SizeOf(Count));
  if Count > 0 then
    Stream.Write(fBoardAfterMove.Memory^, Count);
  Count := Length(fMovementPath);
  Stream.Write(Count, SizeOf(Count));
  if Count > 0 then
    Stream.Write(fMovementPath[0], Count * SizeOf(fMovementPath[0]));
  Stream.Write(fElapsedTime, SizeOf(fElapsedTime));
  Stream.Write(fCompleted, SizeOf(fCompleted));
end;

procedure TGameHistoryRec.AppendMove(const Path: array of TSquare);
var
  I: Integer;
  Index: Integer;
begin
  if Length(Path) > 0 then
  begin
    Index := Length(fMovementPath);
    if (Index > 0) and (MovementPath[Index - 1] = Path[Low(Path)]) then
    begin
      SetLength(fMovementPath, Length(MovementPath) + Length(Path) - 1);
      Dec(Index);
    end
    else
      SetLength(fMovementPath, Length(MovementPath) + Length(Path));
    for I := Low(Path) to High(Path) do
    begin
      fMovementPath[Index] := Path[I];
      Inc(Index);
    end;
  end;
end;

function TGameHistoryRec.IsEmpty: Boolean;
begin
  Result := (Length(MovementPath) = 0);
end;

procedure TGameHistoryRec.SetMovementPath(const Value: TMovementPath);
begin
  fMovementPath := Copy(Value);
end;

{ TGameHistory }

constructor TGameHistory.Create(AGameboardClass: TGameboardClass);
begin
  inherited Create(True);
  fInitialBoard := AGameboardClass.Create;
  fInitialBoard.OnChange := InitialBoardChange;
  fCurrent := 0;
end;

destructor TGameHistory.Destroy;
begin
  fUpdateCount := -1;
  inherited Destroy;
end;

procedure TGameHistory.Notify(Ptr: Pointer; Action: TListNotification);
begin
  inherited Notify(Ptr, Action);
  DoChange;
end;

function TGameHistory.GetPlayers(Index: Integer): TPlayer;
var
  I: Integer;
begin
  Result := InitialBoard.Player;
  for I := 0 to Index - 1 do
    if Items[I].Completed then
      Result := Opponent(Result);
end;

function TGameHistory.GetItems(Index: Integer): TGameHistoryRec;
begin
  Result := TGameHistoryRec(Get(Index));
end;

procedure TGameHistory.SetCurrent(Value: Integer);
begin
  if Current <> Value then
  begin
    fCurrent := Value;
    DoChange;
  end;
end;

procedure TGameHistory.InitialBoardChange(Sender: TObject);
begin
  Clear;
end;

function TGameHistory.IsUpdateLocked: Boolean;
begin
  Result := (fUpdateCount <> 0);
end;

procedure TGameHistory.DoChange;
begin
  if fUpdateCount = 0 then
  begin
    if Assigned(OnChange) then
      OnChange(Self);
  end;
end;

procedure TGameHistory.BeginUpdate;
begin
  Inc(fUpdateCount);
end;

procedure TGameHistory.EndUpdate;
begin
  Dec(fUpdateCount);
  if fUpdateCount = 0 then
    DoChange;
end;

procedure TGameHistory.Clear;
begin
  Current := 0;
  inherited Clear;
end;

procedure TGameHistory.Add(AItem: TGameHistoryRec);
begin
  BeginUpdate;
  try
    while fCurrent < Count do
      Delete(Count - 1);
    fCurrent := inherited Add(AItem) + 1;
    fClosed := False;
  finally
    EndUpdate;
  end;
end;

procedure TGameHistory.LoadFromStream(Stream: TStream);
var
  Item: TGameHistoryRec;
  ItemCount: Integer;
  Index: Integer;
begin
  BeginUpdate;
  try
    Clear;
    InitialBoard.LoadFromStream(Stream);
    Stream.Read(ItemCount, SizeOf(ItemCount));
    for Index := 0 to ItemCount - 1 do
    begin
      Item := TGameHistoryRec.Create;
      Item.LoadFromStream(Stream);
      inherited Add(Item);
    end;
    Stream.Read(fCurrent, SizeOf(fCurrent));
    Stream.Read(fClosed, SizeOf(fClosed));
  finally
    EndUpdate;
  end;
end;

procedure TGameHistory.SaveToStream(Stream: TStream);
var
  ItemCount: Integer;
  Index: Integer;
begin
  fInitialBoard.SaveToStream(Stream);
  ItemCount := Count;
  Stream.Write(ItemCount, SizeOf(ItemCount));
  for Index := 0 to ItemCount - 1 do
    Items[Index].SaveToStream(Stream);
  Stream.Write(fCurrent, SizeOf(fCurrent));
  Stream.Write(fClosed, SizeOf(fClosed));
end;

function TGameHistory.IndexOf(AGameHistoryRec: TGameHistoryRec): Integer;
begin
  Result := inherited IndexOf(AGameHistoryRec);
end;

function TGameHistory.IndexOf(Player: TPlayer; TurnNo: Integer): Integer;
var
  I: Integer;
  TargetTurn: Integer;
  CurrentTurn: Integer;
begin
  Result := -1;
  if Player = InitialBoard.Player then
    TargetTurn := TurnNo * 2 - 1
  else
    TargetTurn := TurnNo * 2;
  CurrentTurn := 1;
  for I := 0 to Count - 1 do
  begin
    if CurrentTurn = TargetTurn then
    begin
      Result := I;
      Exit;
    end
    else if CurrentTurn > TargetTurn then
      Exit;
    if Items[I].Completed then
      Inc(CurrentTurn);
  end;
end;

function TGameHistory.TurnOf(Player: TPlayer; Index: Integer): Integer;
var
  I: Integer;
  IncludeThis: Boolean;
begin
  Result := 1;
  IncludeThis := (Player = InitialBoard.Player);
  for I := 0 to Index - 1 do
    if Items[I].Completed then
    begin
      if IncludeThis then Inc(Result);
      IncludeThis := not IncludeThis;
    end;
end;

function TGameHistory.TotalTimeOf(Player: TPlayer; TurnNo: Integer): Cardinal;
var
  I: Integer;
  IncludeThis: Boolean;
  CurrentTurn: Integer;
begin
  Result := 0;
  if Player = InitialBoard.Player then
  begin
    CurrentTurn := 1;
    IncludeThis := True;
  end
  else
  begin
    CurrentTurn := 0;
    IncludeThis := False;
  end;
  for I := 0 to Count - 1 do
  begin
    if CurrentTurn > TurnNo then
      Exit;
    with Items[I] do
    begin
      if IncludeThis then
        Inc(Result, ElapsedTime);
      if Completed then
      begin
        IncludeThis := not IncludeThis;
        if IncludeThis then
          Inc(CurrentTurn);
      end;
    end;
  end;
end;

function TGameHistory.CurrentTurnOf(Player: TPlayer): Integer;
begin
  Result := TurnOf(Player, Current);
end;

function TGameHistory.PlayedTurnsOf(Player: TPlayer): Integer;
begin
  Result := TurnOf(Player, Count) - 1;
end;

function TGameHistory.PlayedTurns: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := Count - 1 downto 0 do
    if Items[I].Completed then
      Inc(Result);
end;

function TGameHistory.IsCurrentStillRunning: Boolean;
begin
  Result := (Current > 0) and not Items[Current - 1].Completed;
end;

{ TGameboardControl }

constructor TGameboardControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque, csAcceptsControls];
  fGameboard := GameboardClass.Create;
  fGameboard.OnChange := GameChange;
  fGameboard.OnTurn := GameTurnChange;
  fBoardBitmap := TBitmap.Create;
  fPieceImages := TImageList.Create(nil);
  fPieceImages.OnChange := StyleChange;
  fCurrentHistoryRec := TGameHistoryRec.Create;
  fHistory := TGameHistory.Create(GameboardClass);
  fHistory.OnChange := HistoryChange;
  fSolver := TGameSolver.Create;
  fSolver.SearchDepth := 8;
  fAnimator := TAnimator.Create(Self);
  fAnimator.AnimateDelay := 10;
  fAnimator.AnimateStep := 1;
  fSquareDarkPattern := TPicture.Create;
  fSquareDarkPattern.OnChange := StyleChange;
  fSquareDarkSelectPattern := TPicture.Create;
  fSquareDarkSelectPattern.OnChange := StyleChange;
  fSquareDarkMarkPattern := TPicture.Create;
  fSquareDarkMarkPattern.OnChange := StyleChange;
  fSquareLightPattern := TPicture.Create;
  fSquareLightPattern.OnChange := StyleChange;
  fSquareLightSelectPattern := TPicture.Create;
  fSquareLightSelectPattern.OnChange := StyleChange;
  fSquareLightMarkPattern := TPicture.Create;
  fSquareLightMarkPattern.OnChange := StyleChange;
  fBoardBorderColor := clBlack;
  fBoardMarginColor := clWhite;
  fSquareBorderColor := clBlack;
  fSquareDarkColor := clBlack;
  fSquareDarkSelectColor := clRed;
  fSquareDarkMarkColor := clRed;
  fSquareLightColor := clWhite;
  fSquareLightSelectColor := clRed;
  fSquareLightMarkColor := clRed;
  fBoardAngle := 0;
  fInvertSquares := False;
  fCenter := True;
  fMarkSelectableSquares := True;
  fBoardLabels := [blLeft, blRight, blTop, blBottom];
  fBoardLabelsOffset := 5;
  fBoardMarginLeft := 0;
  fBoardMarginTop := 0;
  fBoardMarginRight := 0;
  fBoardMarginBottom := 0;
  fSquareLabel := slNone;
  fSquareLabelOffset := 4;
  fMarkerSize := 2;
  fMarkerOffset := 2;
  fAnimateMoves := True;
  fComputerPlayer := NONE;
  fFixedSquareSize := False;
  fRecordFullMovesOnly := False;
  fSelectionOptions := [soAutoComplete, soAutoPlayComplete];
  fAutoDragDrop := True;
  fDraggingSquare := NO_SQUARE;
  fClickingSquare := NO_SQUARE;
  fThinkCursor := crAppStart;
  fCursor := crDefault;
  DragCursor := crHandPoint;
  SetLength(LastDrawInfo, NumOfSquares);
  SetBounds(Left, Top, 400, 400);
end;

destructor TGameboardControl.Destroy;
begin
  fSolver.Free;
  fHistory.Free;
  fCurrentHistoryRec.Free;
  fPieceImages.Free;
  fSquareDarkPattern.Free;
  fSquareDarkMarkPattern.Free;
  fSquareDarkSelectPattern.Free;
  fSquareLightPattern.Free;
  fSquareLightMarkPattern.Free;
  fSquareLightSelectPattern.Free;
  fBoardBitmap.Free;
  SetLength(fSelection, 0);
  SetLength(LastDrawInfo, 0);
  inherited Destroy;
end;

procedure TGameboardControl.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TGameboardControl.CMFontChanged(var Message: TMessage);
begin
  UpdateFullBoard;
  inherited;
end;

procedure TGameboardControl.CMHintShow(var Message: TCMHintShow);
var
  Square: TSquare;
begin
  inherited;
  with Message.HintInfo^ do
  begin
    Square := ClientToSquare(CursorPos.X, CursorPos.Y);
    if IsLegalSquare(Square) then
    begin
      CursorRect := SquareBounds[Square];
      DoSquareInfoTip(Square, HintStr);
    end;
  end;
end;

procedure TGameboardControl.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_THINKABOUTMOVE:
      ThinkAboutMove;
    WM_SOLVER_START:
    begin
      fThinking := True;
      inherited Cursor := ThinkCursor;
      DoThinkBegin;
    end;
    WM_SOLVER_STOP:
    begin
      fThinking := False;
      inherited Cursor := Cursor;
      DoThinkEnd;
      if TSearchControl(Message.LParam) <> scAbort then
        ApplyComputerMove(Message.WParam);
    end;
    WM_ANIMATOR_START:
    begin
      fAnimating := True;
      DoAnimateBegin;
    end;
    WM_ANIMATOR_STOP:
    begin
      fAnimating := False;
      if not Moving then
      begin
        UpdateBoard(True);
        ResetTimer(False, False);
      end;
      DoAnimateEnd;
    end;
    WM_FLUSHSQUARE:
    begin
      FlushSquare(TSquare(Message.WParamLo), ShortInt(Message.WParamHi));
      if Message.LParam <> 0 then
        SetEvent(Message.LParam);
    end;
    WM_BOARDCHANGE:
      DoBoardChange;
    WM_TURNCHANGE:
      DoTurnChange;
    WM_HISTORYCHANGE:
      DoHistoryChange;
    WM_SELECTIONCHANGE:
      DoSelectionChange;
    WM_BOARDMODECHANGE:
      DoBoardModeChange;
    WM_PAUSESTATECHANGE:
      DoPauseStateChange;
    WM_PLAYINGSTATECHANGE:
      DoPlayingStateChange;
    WM_GAMEENDED:
      DoGameEnded;
  else
    inherited WndProc(Message);
  end;
end;

procedure TGameboardControl.CreateWnd;
begin
  inherited CreateWnd;
  Solver.NotifyWnd := WindowHandle;
  if not (csLoading in ComponentState) then
  begin
    AdjustSize;
    UpdateFullBoard;
    CheckPlaying(True)
  end;
end;

procedure TGameboardControl.DestroyWnd;
begin
  Animator.CancelAnimation;
  CheckPlaying(True, True);
  Solver.NotifyWnd := 0;
  inherited DestroyWnd;
end;

procedure TGameboardControl.Loaded;
begin
  inherited Loaded;
  if HandleAllocated then
  begin
    AdjustSize;
    UpdateFullBoard;
    CheckPlaying(False);
  end;
end;

procedure TGameboardControl.Click;
var
  Pt: TPoint;
  Square: Integer;
  Handled: Boolean;
  ImageIndex: Integer;
begin
  inherited Click;
  Pt := ScreenToClient(Mouse.CursorPos);
  Square := ClientToSquare(Pt.X, Pt.Y);
  if (Square = ClickingSquare) and IsLegalSquare(Square) then
  begin
    Handled := Animating or PieceImages.Dragging;
    DoSquareClick(Square, Handled);
    if not Handled then
      case BoardMode of
        gbmPlay:
          SelectSquare(Square);
        gbmSetup:
        begin
          ImageIndex := GetImageIndexOfSquare(Square);
          if ImageIndex = 2 * Gameboard.DistinctPieceCount then
            ImageIndex := -1
          else
            Inc(ImageIndex);
          SetImageIndexOfSquare(Square, ImageIndex);
        end;
      end;
  end;
  fClickingSquare := NO_SQUARE;
end;

procedure TGameboardControl.Paint;
begin
  if not Animating then
  begin
    Canvas.Brush.Color := Color;
    with FullBoardBounds do
    begin
      BoardBitmap.Canvas.Lock;
      try
        Canvas.Draw(Left, Top, BoardBitmap);
      finally
        BoardBitmap.Canvas.Unlock;
      end;
      ExcludeClipRect(Canvas.Handle, Left, Top, Right, Bottom);
      Canvas.FillRect(ClientRect);
      SelectClipRgn(Canvas.Handle, 0);
    end;
  end;
end;

procedure TGameboardControl.Resize;
begin
  if not (csLoading in ComponentState) then
    AdjustBoardBounds(False);
  inherited Resize;
end;

procedure TGameboardControl.AdjustSize;
begin
  if not (csLoading in ComponentState) and HandleAllocated then
  begin
    if AutoSize then
      with BestClientSize(SquareSize) do
        SetBounds(Left, Top, (Width - ClientWidth) + CX, (Height - ClientHeight) + CY);
    AdjustBoardBounds(False);
  end;
end;

function TGameboardControl.CanResize(var NewWidth, NewHeight: Integer): Boolean;
var
  NewSquareSize: Integer;
begin
  if AutoSize and not (csLoading in ComponentState) then
  begin
    Dec(NewWidth, Width - ClientWidth);
    Dec(NewHeight, Height - ClientHeight);
    if FixedSquareSize then
      NewSquareSize := SquareSize
    else
      NewSquareSize := BestSquareSize(NewWidth, NewHeight);
    with BestClientSize(NewSquareSize) do
    begin
      NewWidth := CX;
      NewHeight := CY;
    end;
    Inc(NewWidth, Width - ClientWidth);
    Inc(NewHeight, Height - ClientHeight);
  end;
  Result := inherited CanResize(NewWidth, NewHeight);
end;

procedure TGameboardControl.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Square: TSquare;
  CurrentCursor: TCursor;
begin
  inherited MouseMove(Shift, X, Y);
  if not (Animating or PieceImages.Dragging) then
  begin
    if Thinking then
      CurrentCursor := ThinkCursor
    else
      CurrentCursor := Cursor;
    if (Shift * [ssLeft, ssRight, ssMiddle]) = [] then
    begin
      if not AutoDragDrop then
      begin
        Square := ClientToSquare(X, Y);
        case BoardMode of
          gbmPlay:
            if CanSelectSquare(Square) or IsSquareSelected(Square) then
              CurrentCursor := DragCursor;
          gbmSetup:
            if IsLegalSquare(Square) then
              CurrentCursor := DragCursor;
        end;
      end
      else if StartDrag(X, Y, True) then
        CurrentCursor := DragCursor;
    end;
    if CurrentCursor <> inherited Cursor then
     inherited Cursor := CurrentCursor;
    if CheckDragStartPt and (Sqrt(Sqr(DragStartPt.X - X) + Sqr(DragStartPt.Y - Y)) >= 3) then
    begin
      fClickingSquare := NO_SQUARE;
      CheckDragStartPt := False;
      StartDrag(X, Y, False);
    end;
  end
  else if IsLegalSquare(DraggingSquare) then
  begin
    PieceImages.DragMove(X, Y);
    if CanDrop(DraggingSquare, X, Y) then
      PieceImages.DragCursor := DragCursor
    else
      PieceImages.DragCursor := crNoDrop;
  end;
end;

procedure TGameboardControl.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  CheckDragStartPt := False;
  fClickingSquare := NO_SQUARE;
  if not (Animating or PieceImages.Dragging) then
  begin
    if Button = mbLeft then
    begin
      fClickingSquare := ClientToSquare(X, Y);
      if AutoDragDrop and StartDrag(X, Y, True) then
      begin
        DragStartPt.X := X;
        DragStartPt.Y := Y;
        CheckDragStartPt := True;
      end;
    end;
  end
  else if IsLegalSquare(DraggingSquare) then
    StopDrag(X, Y, True);
end;

procedure TGameboardControl.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  CheckDragStartPt := False;
  fClickingSquare := NO_SQUARE;
  if PieceImages.Dragging and IsLegalSquare(DraggingSquare) then
    StopDrag(X, Y, Button = mbLeft);
end;

procedure TGameboardControl.DoSquareClick(Square: TSquare; var Handled: Boolean);
begin
  if Assigned(OnSquareClick) then
    OnSquareClick(Self, Square, Handled);
end;

procedure TGameboardControl.DoSquareInfoTip(Square: TSquare; var InfoTip: String);
begin
  if Assigned(OnSquareInfoTip) then
    OnSquareInfoTip(Self, Square, InfoTip);
end;

procedure TGameboardControl.DoSquareDragOver(Source, Target: TSquare; var Accept: Boolean);
begin
  if Assigned(OnSquareDragOver) then
    OnSquareDragOver(Self, Source, Target, Accept);
end;

procedure TGameboardControl.DoSquareDragDrop(Source, Target: TSquare);
begin
  if Assigned(OnSquareDragDrop) then
    OnSquareDragDrop(Self, Source, Target);
end;

procedure TGameboardControl.DoSquareDragStart;
begin
  if Assigned(OnSquareDragStart) then
    OnSquareDragStart(Self);
end;

procedure TGameboardControl.DoSquareDragEnd;
begin
  if Assigned(OnSquareDragEnd) then
    OnSquareDragEnd(Self);
end;

procedure TGameboardControl.DoSquareResize;
begin
  if Assigned(OnSquareResize) then
    OnSquareResize(Self);
end;

procedure TGameboardControl.DoHistoryChange;
begin
  if Assigned(OnHistoryChange) then
    OnHistoryChange(Self);
end;

procedure TGameboardControl.DoTurnChange;
begin
  if Assigned(OnTurnChange) then
    OnTurnChange(Self);
end;

procedure TGameboardControl.DoBoardChange;
begin
  if Assigned(OnBoardChange) then
    OnBoardChange(Self);
end;

procedure TGameboardControl.DoGameEnded;
begin
  if Assigned(OnGameEnded) then
    OnGameEnded(Self);
end;

procedure TGameboardControl.DoAnimateBegin;
begin
  if Assigned(OnAnimateBegin) then
    OnAnimateBegin(Self);
end;

procedure TGameboardControl.DoAnimateEnd;
begin
  if Assigned(OnAnimateEnd) then
    OnAnimateEnd(Self);
end;

procedure TGameboardControl.DoThinkBegin;
begin
  if Assigned(OnThinkBegin) then
    OnThinkBegin(Self);
end;

procedure TGameboardControl.DoThinkEnd;
begin
  if Assigned(OnThinkEnd) then
    OnThinkEnd(Self);
end;

procedure TGameboardControl.DoComputerError;
begin
  if Assigned(OnComputerError) then
    OnComputerError(Self)
  else
    ApplyLegalMove(Random(LegalMovesCount));
end;

procedure TGameboardControl.DoComputerMove(var Index: Integer);
begin
  if Assigned(OnComputerMove) then
    OnComputerMove(Self, Index);
end;

procedure TGameboardControl.DoMoveBegin;
begin
  if Assigned(OnMoveBegin) then
    OnMoveBegin(Self);
end;

procedure TGameboardControl.DoMoveEnd;
begin
  if Assigned(OnMoveEnd) then
    OnMoveEnd(Self);
end;

procedure TGameboardControl.DoSelectionChange;
begin
  if Assigned(OnSelectionChange) then
    OnSelectionChange(Self);
end;

procedure TGameboardControl.DoPlayingStateChange;
begin
  if Assigned(OnPlayingStateChange) then
    OnPlayingStateChange(Self);
end;

procedure TGameboardControl.DoBoardModeChange;
begin
  if Assigned(OnBoardModeChange) then
    OnBoardModeChange(Self);
end;

procedure TGameboardControl.DoPauseStateChange;
begin
  if Assigned(OnPauseStateChange) then
    OnPauseStateChange(Self);
end;

procedure TGameboardControl.SetBoardBorderColor(Value: TColor);
begin
  if BoardBorderColor <> Value then
  begin
    fBoardBorderColor := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetBoardMarginColor(Value: TColor);
begin
  if BoardMarginColor <> Value then
  begin
    fBoardMarginColor := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareBorderColor(Value: TColor);
var
  NeedResize: Boolean;
begin
  if SquareBorderColor <> Value then
  begin
    NeedResize := ((SquareBorderColor = clNone) or (Value = clNone));
    fSquareBorderColor := Value;
    if NeedResize then
      AdjustSize
    else
      UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareDarkSelectColor(Value: TColor);
begin
  if SquareDarkSelectColor <> Value then
  begin
    fSquareDarkSelectColor := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareLightSelectColor(Value: TColor);
begin
  if SquareLightSelectColor <> Value then
  begin
    fSquareLightSelectColor := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareDarkMarkColor(Value: TColor);
begin
  if SquareDarkMarkColor <> Value then
  begin
    fSquareDarkMarkColor := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareLightMarkColor(Value: TColor);
begin
  if SquareLightMarkColor <> Value then
  begin
    fSquareLightMarkColor := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareDarkColor(Value: TColor);
begin
  if SquareDarkColor <> Value then
  begin
    fSquareDarkColor := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareLightColor(Value: TColor);
begin
  if SquareLightColor <> Value then
  begin
    fSquareLightColor := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareDarkSelectPattern(Value: TPicture);
begin
  SquareDarkSelectPattern.Assign(Value);
end;

procedure TGameboardControl.SetSquareLightSelectPattern(Value: TPicture);
begin
  SquareLightSelectPattern.Assign(Value);
end;

procedure TGameboardControl.SetSquareDarkMarkPattern(Value: TPicture);
begin
  SquareDarkMarkPattern.Assign(Value);
end;

procedure TGameboardControl.SetSquareLightMarkPattern(Value: TPicture);
begin
  SquareLightMarkPattern.Assign(Value);
end;

procedure TGameboardControl.SetSquareDarkPattern(Value: TPicture);
begin
  SquareDarkPattern.Assign(Value);
end;

procedure TGameboardControl.SetSquareLightPattern(Value: TPicture);
begin
  SquareLightPattern.Assign(Value);
end;

procedure TGameboardControl.SetBoardAngle(Value: Integer);
begin
  Value := Value mod 360;
  if Value < 0 then Inc(Value, 360);
  if BoardAngle <> Value then
  begin
    fBoardAngle := Value;
    BeginUpdate;
    try
      AdjustSize;
      UpdateFullBoard;
    finally
      EndUpdate;
    end;
  end;
end;

procedure TGameboardControl.SetInvertSquares(Value: Boolean);
begin
  if InvertSquares <> Value then
  begin
    fInvertSquares := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetCenter(Value: Boolean);
begin
  if Center <> Value then
  begin
    fCenter := Value;
    AdjustBoardOffset;
  end;
end;

procedure TGameboardControl.SetMarkSelectableSquares(Value: Boolean);
begin
  if MarkSelectableSquares <> Value then
  begin
    fMarkSelectableSquares := Value;
    UpdateBoard(True);
  end;
end;

procedure TGameboardControl.SetBoardLabels(Value: TBoardLabels);
begin
  if BoardLabels <> Value then
  begin
    fBoardLabels := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetBoardLabelsOffset(Value: Integer);
begin
  if BoardLabelsOffset <> Value then
  begin
    fBoardLabelsOffset := Value;
    UpdateFullBoard;
  end;
end;

procedure TGameboardControl.SetSquareLabel(Value: TSquareLabel);
begin
  if SquareLabel <> Value then
  begin
    fSquareLabel := Value;
    UpdateBoard(True);
  end;
end;

procedure TGameboardControl.SetSquareLabelOffset(Value: Integer);
begin
  if SquareLabelOffset <> Value then
  begin
    fSquareLabelOffset := Value;
    UpdateBoard(True);
  end;
end;

procedure TGameboardControl.SetMarkerSize(Value: Integer);
begin
  if MarkerSize <> Value then
  begin
    fMarkerSize := Value;
    UpdateBoard(True);
  end;
end;

procedure TGameboardControl.SetMarkerOffset(Value: Integer);
begin
  if MarkerOffset <> Value then
  begin
    fMarkerOffset := Value;
    UpdateBoard(True);
  end;
end;

procedure TGameboardControl.SetSquareSize(Value: Integer);
begin
  if (SquareSize <> Value) and (Value > 0) then
  begin
    fSquareSize := Value;
    fFixedSquareSize := True;
    AdjustSize;
  end;
end;

procedure TGameboardControl.SetFixedSquareSize(Value: Boolean);
begin
  if FixedSquareSize <> Value then
  begin
    fFixedSquareSize := Value;
    AdjustSize;
  end;
end;

procedure TGameboardControl.SetBoardMarginLeft(Value: Integer);
begin
  if BoardMarginLeft <> Value then
  begin
    fBoardMarginLeft := Value;
    AdjustSize;
  end;
end;

procedure TGameboardControl.SetBoardMarginTop(Value: Integer);
begin
  if BoardMarginTop <> Value then
  begin
    fBoardMarginTop := Value;
    AdjustSize;
  end;
end;

procedure TGameboardControl.SetBoardMarginRight(Value: Integer);
begin
  if BoardMarginRight <> Value then
  begin
    fBoardMarginRight := Value;
    AdjustSize;
  end;
end;

procedure TGameboardControl.SetBoardMarginBottom(Value: Integer);
begin
  if BoardMarginBottom <> Value then
  begin
    fBoardMarginBottom := Value;
    AdjustSize;
  end;
end;

procedure TGameboardControl.SetPaused(Value: Boolean);
begin
  fPausedByHistory := False;
  if Paused <> Value then
  begin
    fPaused := Value;
    PostMessage(WindowHandle, WM_PAUSESTATECHANGE, 0, 0);
    CheckPlaying(True);
  end;
end;

procedure TGameboardControl.SetBoardMode(Value: TGameboardMode);
begin
  if BoardMode <> Value then
  begin
    fBoardMode := Value;
    ClearSelection;
    ClearHistory;
    PostMessage(WindowHandle, WM_BOARDMODECHANGE, 0, 0);
    if BoardMode = gbmPlay then
      PostMessage(WindowHandle, WM_TURNCHANGE, 0, 0);
    CheckPlaying(False);
  end;
end;

procedure TGameboardControl.SetCursor(Value: TCursor);
begin
  if Cursor <> Value then
  begin
    fCursor := Value;
    if not (Thinking or PieceImages.Dragging) then
      inherited Cursor := Cursor;
  end;
end;

procedure TGameboardControl.SetThinkCursor(Value: TCursor);
begin
  if ThinkCursor <> Value then
  begin
    fThinkCursor := Value;
    if Thinking and not PieceImages.Dragging then
      inherited Cursor := ThinkCursor;
  end;
end;

function TGameboardControl.GetBoardAsText: String;
begin
  Result := Gameboard.BoardAsText;
end;

function TGameboardControl.IsGameboardStored: Boolean;
begin
  Result := (Trim(BoardAsText) <> '');
end;

procedure TGameboardControl.SetBoardAsText(const Value: String);
begin
  Gameboard.BoardAsText := Value;
end;

procedure TGameboardControl.SetComputerPlayer(Value: TPlayer);
var
  WasComputerPlaying: Boolean;
  IsComputerPlaying: Boolean;
begin
  if (ComputerPlayer <> Value) and (Value in [NONE, DARK_PLAYER, LIGHT_PLAYER, BOTH_PLAYERS]) then
  begin
    WasComputerPlaying := ByteBool(CurrentPlayer and ComputerPlayer);
    fComputerPlayer := Value;
    IsComputerPlaying := ByteBool(CurrentPlayer and ComputerPlayer);
    if WasComputerPlaying <> IsComputerPlaying then
    begin
      ClearSelection;
      UpdateBoard(False);
      if IsComputerPlaying then
      begin
        ResetTimer(True, False);
        ThinkAboutMove;
      end
      else
        StopThinking(True);
    end;
  end;
end;

procedure TGameboardControl.SetCurrentPlayer(Value: TPlayer);
begin
  if (CurrentPlayer <> Value) and (Value in [DARK_PLAYER, LIGHT_PLAYER]) then
  begin
    Gameboard.Player := Value;
    ClearHistory;
  end;
end;

function TGameboardControl.GetCurrentPlayer: TPlayer;
begin
  Result := Gameboard.Player;
end;

function TGameboardControl.GetWinnerPlayer: TPlayer;
begin
  Result := Gameboard.Winner;
end;

function TGameboardControl.GetLooserPlayer: TPlayer;
begin
  Result := Gameboard.Looser;
end;

function TGameboardControl.GetOpenerPlayer: TPlayer;
begin
  Result := History.InitialBoard.Player;
end;

function TGameboardControl.GetLegalMovesCount: Integer;
begin
  Result := Gameboard.LegalMovesCount;
end;

function TGameboardControl.GetCurrentTurn: Integer;
begin
  Result := (History.Current div 2) + 1;
end;

function TGameboardControl.GetElapsedTime: Cardinal;
begin
  Result := 0;
  if Playing or not IsGameEnded then
  begin
    Result := fMoveElapsedTime;
    if Playing and (Moving or not Animating) then
      Inc(Result, GetTickCount - fMoveStartTime);
  end;
end;

function TGameboardControl.GetCurrentTime: Cardinal;
var
  Index: Integer;
begin
  Result := ElapsedTime;
  if Playing or not IsGameEnded then
    for Index := History.Current - 1 downto 0 do
      with History[Index] do
        if not Completed then
          Inc(Result, ElapsedTime)
        else
          Exit;
end;

function TGameboardControl.GetPlayedTurns: Integer;
begin
  Result := (History.Count + 1) div 2;
end;

function TGameboardControl.GetPlayedTime: Cardinal;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to History.Count - 1 do
    Inc(Result, History[I].ElapsedTime);
end;

function TGameboardControl.GetCurrentTurnOf(Player: TPlayer): Integer;
begin
  if Player in [LIGHT_PLAYER, DARK_PLAYER] then
    Result := History.CurrentTurnOf(Player)
  else
    Result := 0;
end;

function TGameboardControl.GetCurrentTimeOf(Player: TPlayer): Cardinal;
begin
  if Player in [LIGHT_PLAYER, DARK_PLAYER] then
  begin
    Result := History.TotalTimeOf(Player, CurrentTurnOf[Player] - 1);
    if Player = CurrentPlayer then
      Inc(Result, CurrentTime);
  end
  else
    Result := 0;
end;

function TGameboardControl.GetPlayedTurnsOf(Player: TPlayer): Integer;
begin
  if Player in [LIGHT_PLAYER, DARK_PLAYER] then
    Result := History.PlayedTurnsOf(Player)
  else
    Result := 0;
end;

function TGameboardControl.GetPlayedTimeOf(Player: TPlayer): Cardinal;
begin
  if Player in [LIGHT_PLAYER, DARK_PLAYER] then
    Result := History.TotalTimeOf(Player, PlayedTurnsOf[Player])
  else
    Result := 0;
end;

procedure TGameboardControl.SetThinkPriority(Value: TThreadPriority);
begin
  Solver.Priority := Value;
end;

function TGameboardControl.GetThinkPriority: TThreadPriority;
begin
  Result := Solver.Priority;
end;

procedure TGameboardControl.SetSearchDepth(Value: Byte);
begin
  if Value < 1 then
    Value := 1
  else if Value > MAX_SEARCH_DEPTH then
    Value := MAX_SEARCH_DEPTH;
  Solver.SearchDepth := Value;
end;

function TGameboardControl.GetSearchDepth: Byte;
begin
  Result := Solver.SearchDepth;
end;

procedure TGameboardControl.SetMaxMemory(Value: Cardinal);
begin
  Solver.MaxMemory := Value;
end;

function TGameboardControl.GetMaxMemory: Cardinal;
begin
  Result := Solver.MaxMemory;
end;

procedure TGameBoardControl.SetAnimateDelay(Value: Byte);
begin
  Animator.AnimateDelay := Value;
end;

function TGameBoardControl.GetAnimateDelay: Byte;
begin
  Result := Animator.AnimateDelay;
end;

procedure TGameBoardControl.SetAnimateStep(Value: Byte);
begin
  if Value = 0 then
    Value := 1;
  Animator.AnimateStep := Value;
end;

function TGameBoardControl.GetAnimateStep: Byte;
begin
  Result := Animator.AnimateStep;
end;

function TGameboardControl.GetSquareBounds(Square: TSquare): TRect;
var
  Row, Col: Integer;
begin
  SquareToCoordinate(Square, Row, Col);
  AdjustCoordinates(Row, Col, BoardAngle);
  SetRect(Result, 0, 0, SquareSize + 1, SquareSize + 1);
  with BoardBounds do
    OffsetRect(Result, Left + Col * SquareSize, Top + Row * SquareSize);
end;

procedure TGameboardControl.SetSelection(const Value: TMovementPath);
var
  I: Integer;
  Changed: Boolean;
begin
  Changed := (Length(fSelection) <> Length(Value));
  SetLength(fSelection, Length(Value));
  for I := 0 to Length(Value) - 1 do
  begin
    Changed := Changed or (fSelection[I] <> Value[I]);
    fSelection[I] := Value[I];
  end;
  if Changed then
  begin
    UpdateBoard(False);
    PostMessage(WindowHandle, WM_SELECTIONCHANGE, 0, 0);
  end;
end;

function TGameboardControl.GetSelSquares(Index: Integer): TSquare;
begin
  if (Index >= 0) and (Index < Length(fSelection)) then
    Result := fSelection[Index]
  else
    Result := NO_SQUARE;
end;

function TGameboardControl.GetSelSquareCount: Integer;
begin
  Result := Length(fSelection);
end;

function TGameboardControl.GetNumOfRows: Integer;
begin
  Result := 8;
end;

function TGameboardControl.GetNumOfCols: Integer;
begin
  Result := 8;
end;

function TGameboardControl.GetNumOfSquares: Integer;
begin
  Result := NumOfRows * NumOfCols;
end;

function TGameboardControl.IsUpdateLocked: Boolean;
begin
  Result := (fUpdateCount <> 0);
end;

procedure TGameboardControl.GameTurnChange(Sender: TObject);
begin
  if Playing and Moving then
  begin
    CurrentHistoryRec.Assign(Gameboard);
    CurrentHistoryRec.ElapsedTime := ElapsedTime;
    CurrentHistoryRec.Completed := True;
    StoreCurrentHistoryRecord;
    ClearSelection;
  end;
  PostMessage(WindowHandle, WM_TURNCHANGE, 0, 0);
end;

procedure TGameboardControl.GameChange(Sender: TObject);
begin
  StopThinking(True);
  Modified := True;
  if not Moving then
  begin
    ClearSelection;
    CurrentHistoryRec.Clear;
    ResetTimer(True, False);
  end;
  UpdateBoard(False);
  PostMessage(WindowHandle, WM_BOARDCHANGE, 0, 0);
  if (BoardMode = gbmPlay) and not Moving then
    PostMessage(WindowHandle, WM_TURNCHANGE, 0, 0);
end;

procedure TGameboardControl.HistoryChange(Sender: TObject);
begin
  ResetTimer(True, False);
  PostMessage(WindowHandle, WM_HISTORYCHANGE, 0, 0);
end;

procedure TGameboardControl.StyleChange(Sender: TObject);
begin
  if Sender = PieceImages then
    UpdateBoard(True)
  else
    UpdateFullBoard;
end;

function TGameboardControl.BestClientSize(ASquareSize: Integer): TSize;
var
  Adj: Integer;
begin
  if SquareBorderColor <> clNone then Adj := 1 else Adj := 0;
  Result.CX := (NumOfCols * ASquareSize) + BoardMarginLeft + BoardMarginRight + Adj;
  Result.CY := (NumOfRows * ASquareSize) + BoardMarginTop + BoardMarginBottom + Adj;
end;

function TGameboardControl.BestSquareSize(AWidth, AHeight: Integer): Integer;
var
  WSize, HSize: Integer;
  Adj: Integer;
begin
  if SquareBorderColor <> clNone then Adj := 1 else Adj := 0;
  WSize := (AWidth - BoardMarginLeft - BoardMarginRight - Adj) div NumOfCols;
  HSize := (AHeight - BoardMarginTop - BoardMarginBottom - adj) div NumOfRows;
  if WSize < HSize then
    Result := WSize
  else
    Result := HSize;
end;

function TGameboardControl.CoordinateToSquare(Row, Col: Integer): TSquare;
begin
  if (Row >= 0) and (Row < NumOfRows) and (Col >= 0) and (Col < NumOfCols) then
    Result := (Row * NumOfRows) + Col
  else
    Result := NO_SQUARE;
end;

function TGameboardControl.SquareToCoordinate(Square: TSquare; out Row, Col: Integer): Boolean;
begin
  Row := Square div NumOfRows;
  Col := Square mod NumOfCols;
  Result := (Row >= 0) and (Row < NumOfRows) and (Col >= 0) and (Col < NumOfCols);
end;

function TGameboardControl.ClientToSquare(X, Y: Integer): Integer;
var
  Row, Col: Integer;
begin
  Col := (X - BoardBounds.Left) div SquareSize;
  Row := (Y - BoardBounds.Top) div SquareSize;
  AdjustCoordinates(Row, Col, -BoardAngle);
  Result := CoordinateToSquare(Row, Col)
end;

function TGameboardControl.IsLegalSquare(Square: TSquare): Boolean;
begin
  Result := (Square >= 0) and (Square < NumOfSquares);
end;

function TGameboardControl.IsLogicalDarkSquare(Square: TSquare): Boolean;
begin
  Result := Odd(Square) xor Odd(Square div NumOfRows);
end;

function TGameboardControl.IsVisualDarkSquare(Square: TSquare): Boolean;
begin
  Result := IsLogicalDarkSquare(Square) xor InvertSquares;
end;

procedure TGameboardControl.SetImageIndexOfSquare(Square: TSquare; ImageIndex: Integer);
begin
end;

function TGameboardControl.GetImageIndexOfSquare(Square: TSquare): Integer;
begin
  Result := -1;
end;

function TGameboardControl.GetStateOfSquare(Square: TSquare): TSquareDrawStates;
begin
  Result := [];
  if Playing and not Moving and not (Animating or PieceImages.Dragging) then
    if IsSquareSelected(Square) then
      Include(Result, sdsSelected)
    else if MarkSelectableSquares and CanSelectSquare(Square) then
      Include(Result, sdsMarked);
end;

function TGameboardControl.GetSquareLabel(Square: TSquare): String;
var
  Row, Col: Integer;
begin
  if SquareToCoordinate(Square, Row, Col) then
    Result := GetColLabel(Col) + GetRowLabel(Row)
  else
    Result := '';
end;

function TGameboardControl.GetRowLabel(Row: Integer): String;
begin
  Result := IntToStr(Row + 1);
end;

function TGameboardControl.GetColLabel(Col: Integer): String;
const
  AlphaCount = Ord('Z') - Ord('A') + 1;
begin
  Result := Chr(Ord('A') + Col mod AlphaCount);
  if NumOfCols > AlphaCount then
    Result := Result + IntToStr((Col div AlphaCount) + 1);
end;

procedure TGameboardControl.AdjustBoardBounds(ForceAdjust: Boolean);
var
  NewSquareSize: Integer;
  NewSize: TSize;
begin
  if not FixedSquareSize then
    NewSquareSize := BestSquareSize(ClientWidth, ClientHeight)
  else
    NewSquareSize := SquareSize;
  NewSize := BestClientSize(SquareSize);
  if ForceAdjust or (SquareSize <> NewSquareSize) or
    (NewSize.CX <> BoardBitmap.Width) or (NewSize.CY <> BoardBitmap.Height) then
  begin
    Animator.CancelAnimation;
    BeginUpdate;
    try
      fSquareSize := NewSquareSize;
      SetRect(fFullBoardBounds, 0, 0, NewSize.CX, NewSize.CY);
      fBoardBounds.Left := fFullBoardBounds.Left + BoardMarginLeft;
      fBoardBounds.Top := fFullBoardBounds.Top + BoardMarginTop;
      fBoardBounds.Right := fFullBoardBounds.Right - BoardMarginRight;
      fBoardBounds.Bottom := fFullBoardBounds.Bottom - BoardMarginBottom;
      BoardBitmap.Width := NewSize.CX;
      BoardBitmap.Height := NewSize.CY;
      UpdateFullBoard;
      DoSquareResize;
    finally
      EndUpdate;
    end;
  end;
  AdjustBoardOffset;
end;

procedure TGameboardControl.AdjustBoardOffset;
var
  Offset: TPoint;
begin
  with FullBoardBounds do
    if Center then
    begin
      Offset.X := (ClientWidth - (Right - Left)) div 2 - Left;
      Offset.Y := (ClientHeight - (Bottom - Top)) div 2 - Top;
    end
    else
    begin
      Offset.X := -Left;
      Offset.Y := -Top;
    end;
  if (Offset.X <> 0) or (Offset.Y <> 0) then
  begin
    OffsetRect(fFullBoardBounds, Offset.X, Offset.Y);
    OffsetRect(fBoardBounds, Offset.X, Offset.Y);
    if not IsUpdateLocked and HandleAllocated then
      InvalidateRect(WindowHandle, nil, False);
  end;
end;

procedure TGameboardControl.AdjustCoordinates(var Row, Col: Integer; Angle: Integer);
var
  T: Integer;
begin
  case Angle mod 360 of
    -270, 90:
    begin
      T := Col;
      Col := NumOfRows - 1 - Row;
      Row := T;
    end;
    -180, 180:
    begin
      Row := NumOfRows - 1 - Row;
      Col := NumOfCols - 1 - Col;
    end;
    -90, 270:
    begin
      T := NumOfCols - 1 - Col;
      Col := Row;
      Row := T;
    end;
  end;
end;

function TGameboardControl.UpdateFullBoard: Boolean;
begin
  if not (IsUpdateLocked or Animating) and HandleAllocated and not (csLoading in ComponentState) then
  begin
    InvalidateRect(WindowHandle, @fFullBoardBounds, False);
    BoardBitmap.Canvas.Lock;
    try
      DrawBoard(BoardBitmap.Canvas, BoardBitmap.Canvas.ClipRect);
    finally
      BoardBitmap.Canvas.Unlock;
    end;
    Result := UpdateBoard(True);
    fIsFullBoardReady := True;
  end
  else
  begin
    fIsFullBoardReady := False;
    fIsBoardReady := False;
    Result := False;
  end;
end;

function TGameboardControl.UpdateBoard(AllSquares: Boolean): Boolean;
var
  Square: Integer;
begin
  Result := False;
  if not (IsUpdateLocked or Animating) and HandleAllocated and not (csLoading in ComponentState) then
  begin
    for Square := 0 to NumOfSquares - 1 do
      if UpdateSquare(Square, AllSquares) then
        Result := True;
    fIsBoardReady := True;
  end
  else
    fIsBoardReady := False;
end;

function TGameboardControl.UpdateSquare(Square: TSquare; ForceUpdate: Boolean): Boolean;
var
  Rect: TRect;
  ImageIndex: Integer;
  States: TSquareDrawStates;
begin
  Result := False;
  if not (IsUpdateLocked or Animating) then
  begin
    if Square <> DraggingSquare then
      ImageIndex := GetImageIndexOfSquare(Square)
    else
      ImageIndex := -1;
    States := GetStateOfSquare(Square);
    if ForceUpdate or (sdsNoCache in States) or
      (LastDrawInfo[Square].States <> States) or
      (LastDrawInfo[Square].ImageIndex <> ImageIndex) then
    begin
      Rect := SquareBounds[Square];
      InvalidateRect(WindowHandle, @Rect, False);
      OffsetRect(Rect, -FullBoardBounds.Left, -FullBoardBounds.Top);
      BoardBitmap.Canvas.Lock;
      try
        DrawSquare(BoardBitmap.Canvas, Square, Rect, ImageIndex, States);
      finally
        BoardBitmap.Canvas.Unlock;
      end;
      LastDrawInfo[Square].States := States;
      LastDrawInfo[Square].ImageIndex := ImageIndex;
    end;
  end;
end;

procedure TGameboardControl.DrawBoard(Canvas: TCanvas; const Rect: TRect);
begin
  with Canvas do
  begin
    Brush.Color := BoardMarginColor;
    Brush.Style := bsSolid;
    Pen.Mode := pmCopy;
    Pen.Width := 1;
    Pen.Style := psInsideFrame;
    Pen.Color := BoardBorderColor;
    if BoardBorderColor <> clNone then
      Rectangle(Rect)
    else
      FillRect(Rect);
  end;
  DrawBoardLabels(Canvas, Rect);
end;

procedure TGameboardControl.DrawBoardLabels(Canvas: TCanvas; const Rect: TRect);
var
  BoardLabel: TBoardLabel;
  Row, Col: Integer;
  LabelRect: TRect;
begin
  Canvas.Font := Font;
  for BoardLabel := Low(TBoardLabel) to High(TBoardLabel) do
    if BoardLabel in BoardLabels then
      case BoardLabel of
        blLeft:
          if BoardMarginLeft > BoardLabelsOffset then
          begin
            with LabelRect do
            begin
              Left := Rect.Left;
              Top := Rect.Top + BoardMarginTop;
              Right := Rect.Left + BoardMarginLeft;
              Bottom := Top + SquareSize;
            end;
            for Row := 0 to NumOfRows - 1 do
            begin
              DrawRowLabel(Canvas, Row, alRight, LabelRect);
              OffsetRect(LabelRect, 0, LabelRect.Bottom - LabelRect.Top);
            end;
          end;
        blRight:
          if BoardMarginRight > BoardLabelsOffset then
          begin
            with LabelRect do
            begin
              Left := Rect.Right - BoardMarginRight;
              Top := Rect.Top + BoardMarginTop;
              Right := Rect.Right;
              Bottom := Top + SquareSize;
            end;
            for Row := 0 to NumOfRows - 1 do
            begin
              DrawRowLabel(Canvas, Row, alLeft, LabelRect);
              OffsetRect(LabelRect, 0, LabelRect.Bottom - LabelRect.Top);
            end;
          end;
        blTop:
          if BoardMarginTop > BoardLabelsOffset then
          begin
            with LabelRect do
            begin
              Left := Rect.Left + BoardMarginLeft;
              Top := Rect.Top;
              Right := Left + SquareSize;
              Bottom := Rect.Top + BoardMarginTop;
            end;
            for Col := 0 to NumOfCols - 1 do
            begin
              DrawColLabel(Canvas, Col, alBottom, LabelRect);
              OffsetRect(LabelRect, LabelRect.Right - LabelRect.Left, 0);
            end;
          end;
        blBottom:
          if BoardMarginBottom > BoardLabelsOffset then
          begin
            with LabelRect do
            begin
              Left := Rect.Left + BoardMarginLeft;
              Top := Rect.Bottom - BoardMarginBottom;
              Right := Left + SquareSize;
              Bottom := Rect.Bottom;
            end;
            for Col := 0 to NumOfCols - 1 do
            begin
              DrawColLabel(Canvas, Col, alTop, LabelRect);
              OffsetRect(LabelRect, LabelRect.Right - LabelRect.Left, 0);
            end;
          end;
      end;
end;

procedure TGameboardControl.DrawRowLabel(Canvas: TCanvas; Row: Integer;
  TextAlign: TAlign; const Rect: TRect);
var
  Col: Integer;
  Text: String;
  X, Y: Integer;
  BkMode: Integer;
  Align: Integer;
begin
  Col := -1;
  AdjustCoordinates(Row, Col, -BoardAngle);
  if (Row >= 0) and (Row < NumOfRows) then
    Text := GetRowLabel(Row)
  else
    Text := GetColLabel(Col);
  if Text <> '' then
  begin
    if TextAlign = alRight then
      X := Rect.Right - Canvas.TextWidth(Text) - BoardLabelsOffset
    else
      X := Rect.Left + BoardLabelsOffset;
    Y := Rect.Top + ((Rect.Bottom - Rect.Top) - Canvas.TextHeight(Text)) div 2;
    Align := SetTextAlign(Canvas.Handle, TA_LEFT or TA_TOP);
    BkMode := SetBkMode(Canvas.Handle, TRANSPARENT);
    Canvas.TextOut(X, Y, Text);
    SetBkMode(Canvas.Handle, BkMode);
    SetTextAlign(Canvas.Handle, Align);
  end;
end;

procedure TGameboardControl.DrawColLabel(Canvas: TCanvas; Col: Integer;
  TextAlign: TAlign; const Rect: TRect);
var
  Row: Integer;
  Text: String;
  X, Y: Integer;
  BkMode: Integer;
  Align: Integer;
begin
  Row := -1;
  AdjustCoordinates(Row, Col, -BoardAngle);
  if (Col >= 0) and (Col < NumOfCols) then
    Text := GetColLabel(Col)
  else
    Text := GetRowLabel(Row);
  if Text <> '' then
  begin
    X := Rect.Left + ((Rect.Right - Rect.Left) - Canvas.TextWidth(Text)) div 2;
    if TextAlign = alBottom then
      Y := Rect.Bottom - Canvas.TextHeight(Text) - BoardLabelsOffset
    else
      Y := Rect.Top + BoardLabelsOffset;
    Align := SetTextAlign(Canvas.Handle, TA_LEFT or TA_TOP);
    BkMode := SetBkMode(Canvas.Handle, TRANSPARENT);
    Canvas.TextOut(X, Y, Text);
    SetBkMode(Canvas.Handle, BkMode);
    SetTextAlign(Canvas.Handle, Align);
  end;
end;

procedure TGameboardControl.FlushSquare(Square: TSquare; ImageIndex: Integer);
var
  WaitEvent: THandle;
begin
  if GetCurrentThreadId = MainThreadID then
  begin
    Canvas.Lock;
    try
      DrawSquare(Canvas, Square, SquareBounds[Square], ImageIndex, []);
    finally
      Canvas.Unlock;
    end;
  end
  else
  begin
    WaitEvent := CreateEvent(nil, True, False, nil);
    try
      PostMessage(Handle, WM_FLUSHSQUARE, MakeLong(Square, Word(ImageIndex)), WaitEvent);
      WaitForSingleObject(WaitEvent, INFINITE);
    finally
      CloseHandle(WaitEvent);
    end;
  end;
end;

procedure TGameboardControl.DrawSquare(Canvas: TCanvas; Square: TSquare;
  Rect: TRect; ImageIndex: Integer; States: TSquareDrawStates);
var
  Color: TColor;
  Pattern: TGraphic;
begin
  if IsVisualDarkSquare(Square) then
  begin
    Color := SquareDarkColor;
    Pattern := SquareDarkPattern.Graphic;
  end
  else
  begin
    Color := SquareLightColor;
    Pattern := SquareLightPattern.Graphic;
  end;
  with Canvas do
  begin
    Brush.Color := Color;
    Brush.Style := bsSolid;
    Pen.Mode := pmCopy;
    Pen.Width := 1;
    Pen.Style := psSolid;
    Pen.Color := SquareBorderColor;
    if SquareBorderColor <> clNone then
    begin
      Rectangle(Rect);
      InflateRect(Rect, -Pen.Width, -Pen.Width);
    end
    else
    begin
      Dec(Rect.Right);
      Dec(Rect.Bottom);
      FillRect(Rect);
    end;
    if Assigned(Pattern) then
      StretchDraw(Rect, Pattern);
  end;
  DrawStates(Canvas, Square, Rect, ImageIndex, States);
  DrawLabel(Canvas, Square, Rect, States);
  DrawPiece(Canvas, Square, Rect, ImageIndex, States);
end;

procedure TGameboardControl.DrawStates(Canvas: TCanvas; Square: TSquare;
  const Rect: TRect; ImageIndex: Integer; States: TSquareDrawStates);
var
  Color: TColor;
  Pattern: TGraphic;
begin
  if sdsSelected in States then
  begin
    if IsVisualDarkSquare(Square) then
    begin
      Color := SquareDarkSelectColor;
      Pattern := SquareDarkSelectPattern.Graphic;
    end
    else
    begin
      Color := SquareLightSelectColor;
      Pattern := SquareLightSelectPattern.Graphic;
    end;
    if Assigned(Pattern) then
      Canvas.StretchDraw(Rect, Pattern)
    else if Color <> clNone then
    begin
      Canvas.Pen.Mode := pmCopy;
      Canvas.Brush.Color := Color;
      Canvas.FillRect(Rect);
    end
  end
  else if sdsMarked in States then
  begin
    if IsVisualDarkSquare(Square) then
    begin
      Color := SquareDarkMarkColor;
      Pattern := SquareDarkMarkPattern.Graphic;
    end
    else
    begin
      Color := SquareLightMarkColor;
      Pattern := SquareLightMarkPattern.Graphic;
    end;
    if Assigned(Pattern) then
      Canvas.StretchDraw(Rect, Pattern)
    else if Color <> clNone then
    begin
      Canvas.Pen.Mode := pmCopy;
      Canvas.Pen.Width := MarkerSize;
      Canvas.Pen.Style := psSolid;
      Canvas.Pen.Color := Color;
      Canvas.Brush.Style := bsClear;
      Canvas.Rectangle(Rect.Left + MarkerOffset + 1, Rect.Top + MarkerOffset + 1,
        Rect.Right - MarkerOffset, Rect.Bottom - MarkerOffset);
    end
  end;
end;

procedure TGameboardControl.DrawPiece(Canvas: TCanvas; Square: TSquare;
  const Rect: TRect; ImageIndex: Integer; States: TSquareDrawStates);
var
  X, Y: Integer;
begin
  if ImageIndex >= 0 then
  begin
    X := (Rect.Left + Rect.Right - PieceImages.Width) div 2;
    Y := (Rect.Top + Rect.Bottom - PieceImages.Height) div 2;
    PieceImages.Draw(Canvas, X, Y, ImageIndex, dsNormal, itImage, True);
  end;
end;

procedure TGameboardControl.DrawLabel(Canvas: TCanvas; Square: TSquare;
  const Rect: TRect; States: TSquareDrawStates);
var
  Text: String;
  BkMode: Integer;
  Align: Integer;
  X, Y: Integer;
begin
  if SquareLabel <> slNone then
  begin
    Text := GetSquareLabel(Square);
    if Text <> '' then
    begin
      Canvas.Font := Font;
      case SquareLabel of
        slCenter:
        begin
          X := Rect.Left + ((Rect.Right - Rect.Left) - Canvas.TextWidth(Text)) div 2;
          Y := Rect.Top + ((Rect.Bottom - Rect.Top) - Canvas.TextHeight(Text)) div 2;
        end;
        slLeftCenter:
        begin
          X := Rect.Left + SquareLabelOffset;
          Y := Rect.Top + ((Rect.Bottom - Rect.Top) - Canvas.TextHeight(Text)) div 2;
        end;
        slLeftTop:
        begin
          X := Rect.Left + SquareLabelOffset;
          Y := Rect.Top + SquareLabelOffset;
        end;
        slTopCenter:
        begin
          X := Rect.Left + ((Rect.Right - Rect.Left) - Canvas.TextWidth(Text)) div 2;
          Y := Rect.Top + SquareLabelOffset;
        end;
        slRightTop:
        begin
          X := Rect.Right - SquareLabelOffset - Canvas.TextWidth(Text);
          Y := Rect.Top + SquareLabelOffset;
        end;
        slRightCenter:
        begin
          X := Rect.Right - SquareLabelOffset - Canvas.TextWidth(Text);
          Y := Rect.Top + ((Rect.Bottom - Rect.Top) - Canvas.TextHeight(Text)) div 2;
        end;
        slRightBottom:
        begin
          X := Rect.Right - SquareLabelOffset - Canvas.TextWidth(Text);
          Y := Rect.Bottom - SquareLabelOffset - Canvas.TextHeight(Text);
        end;
        slBottomCenter:
        begin
          X := Rect.Left + ((Rect.Right - Rect.Left) - Canvas.TextWidth(Text)) div 2;
          Y := Rect.Bottom - SquareLabelOffset - Canvas.TextHeight(Text);
        end;
        slLeftBottom:
        begin
          X := Rect.Left + SquareLabelOffset;
          Y := Rect.Bottom - SquareLabelOffset - Canvas.TextHeight(Text);
        end;
      else
        Exit;
      end;
      Align := SetTextAlign(Canvas.Handle, TA_LEFT or TA_TOP);
      BkMode := SetBkMode(Canvas.Handle, TRANSPARENT);
      Canvas.TextOut(X, Y, Text);
      SetBkMode(Canvas.Handle, BkMode);
      SetTextAlign(Canvas.Handle, Align);
    end;
  end;
end;

procedure TGameboardControl.ExtractPiecesImages(BitmapOfPieces: TBitmap;
  Rows, Cols: Integer);
var
  SrcRect, DstRect: TRect;
  PieceBitmap: TBitmap;
  R, C: Integer;
begin
  BeginUpdate;
  try
    PieceImages.Clear;
    PieceImages.Width := BitmapOfPieces.Width div Cols;
    PieceImages.Height := BitmapOfPieces.Height div Rows;
    PieceBitmap := TBitmap.Create;
    try
      PieceBitmap.Width := PieceImages.Width;
      PieceBitmap.Height := PieceImages.Height;
      SetRect(DstRect, 0, 0, PieceImages.Width, PieceImages.Height);
      SetRect(SrcRect, 0, 0, PieceImages.Width, PieceImages.Height);
      for R := 0 to Rows - 1 do
      begin
        for C := 0 to Cols - 1 do
        begin
          PieceBitmap.Canvas.CopyRect(DstRect, BitmapOfPieces.Canvas, SrcRect);
          PieceImages.AddMasked(PieceBitmap, clDefault);
          OffsetRect(SrcRect, PieceImages.Width, 0);
        end;
        OffsetRect(SrcRect, -Cols * PieceImages.Width, PieceImages.Height);
      end;
    finally
      PieceBitmap.Free;
    end;
  finally
    EndUpdate;
  end;
end;

procedure TGameboardControl.StorePath(const Path: Array of TSquare);
begin
  CurrentHistoryRec.AppendMove(Path);
  if not IsComputerTurn then
    StartNewSelection(Path[High(Path)]);
  if AnimateMoves and not DragedMove then
    AnimateForward(Path);
end;

function TGameboardControl.IsSquareSelected(Square: TSquare): Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to Length(fSelection) - 1 do
    if fSelection[I] = Square then
    begin
      Result := True;
      Exit;
    end;
end;

function TGameboardControl.CanSelectSquare(Square: TSquare): Boolean;
begin
  Result := IsLegalSquare(Square) and Playing and not IsComputerTurn;
end;

function TGameboardControl.SelectSquare(Square: TSquare): Boolean;
begin
  Result := False;
  if DeselectSquare(Square) then
    Result := True
  else if CanSelectSquare(Square) then
  begin
    SetLength(fSelection, Length(fSelection) + 1);
    fSelection[Length(fSelection) - 1] := Square;
    if soAutoComplete in SelectionOptions then
      CompletePath(fSelection);
    if ((soAutoPlayStep in SelectionOptions) and CanPlayPath(fSelection)) or
       ((soAutoPlayComplete in SelectionOptions) and IsPathComplete(fSelection))
    then
      PlayPath(fSelection)
    else
    begin
      UpdateBoard(False);
      PostMessage(WindowHandle, WM_SELECTIONCHANGE, 0, 0);
    end;
    Result := True;
  end;
end;

function TGameboardControl.DeselectSquare(Square: TSquare): Boolean;
begin
  Result := False;
  if (Length(fSelection) > 0) and (fSelection[Length(fSelection) - 1] = Square) then
  begin
    SetLength(fSelection, Length(fSelection) - 1);
    UpdateBoard(False);
    PostMessage(WindowHandle, WM_SELECTIONCHANGE, 0, 0);
    Result := True;
  end;
end;

procedure TGameboardControl.StartNewSelection(Square: TSquare);
begin
  if not IsLegalSquare(Square) then
    ClearSelection
  else if (Length(fSelection) <> 1) or (fSelection[0] <> Square) then
  begin
    SetLength(fSelection, 1);
    fSelection[0] := Square;
    PostMessage(WindowHandle, WM_SELECTIONCHANGE, 0, 0);
  end
end;

function TGameboardControl.ClearSelection: Boolean;
begin
  Result := False;
  if Length(fSelection) > 0 then
  begin
    SetLength(fSelection, 0);
    UpdateBoard(False);
    PostMessage(WindowHandle, WM_SELECTIONCHANGE, 0, 0);
    Result := True;
  end;
end;

function TGameboardControl.CompleteSelection: Boolean;
begin
  Result := CompletePath(fSelection);
end;

function TGameboardControl.IsSelectionComplete: Boolean;
begin
  Result := (Length(fSelection) > 0) and IsPathComplete(fSelection);
end;

function TGameboardControl.CanPlaySelection: Boolean;
begin
  Result := (Length(fSelection) > 0) and CanPlayPath(fSelection);
end;

function TGameboardControl.PlaySelection: Boolean;
begin
  Result := PlayPath(fSelection);
end;

procedure TGameboardControl.ResetTimer(ResetOld, SaveCurrent: Boolean);
begin
  if ResetOld then
    fMoveElapsedTime := 0;
  if SaveCurrent then
    Inc(fMoveElapsedTime, GetTickCount - fMoveStartTime);
  fMoveStartTime := GetTickCount;
end;

function TGameboardControl.CheckPlaying(SaveTime, StopAnyway: Boolean): Boolean;
var
  ShouldPlay: Boolean;
begin
  Result := False;
  ShouldPlay := not StopAnyway and not (csLoading in ComponentState) and
                HandleAllocated and not Paused and (BoardMode = gbmPlay) and
                not IsGameEnded;
  if Playing <> ShouldPlay then
  begin
    fPlaying := ShouldPlay;
    if Playing then
    begin
      PostMessage(WindowHandle, WM_PLAYINGSTATECHANGE, 0, 0);
      PostMessage(WindowHandle, WM_THINKABOUTMOVE, 0, 0);
      ResetTimer(not SaveTime, False);
    end
    else
    begin
      ResetTimer(not SaveTime, SaveTime);
      StopThinking(True);
      PostMessage(WindowHandle, WM_PLAYINGSTATECHANGE, 0, 0);
    end;
    UpdateBoard(False);
    Result := True;
  end;
end;

function TGameboardControl.CanThinkAboutMove: Boolean;
begin
  Result := (Playing and IsComputerTurn) and not Solver.IsBusy;
end;

function TGameboardControl.ThinkAboutMove: Boolean;
begin
  Result := CanThinkAboutMove and Solver.Solve(Gameboard)
end;

function TGameboardControl.StopThinking(IgnoreSolution: Boolean): Boolean;
begin
  Result := False;
  if Solver.IsBusy then
  begin
    if IgnoreSolution then
      Solver.State := scAbort
    else
      Solver.State := scStop;
    Solver.Suspended := False;
    Solver.WaitForIdle;
    Result := True;
  end;
end;

function TGameboardControl.AnimateForward(const Path: array of TSquare): Boolean;
begin
  Result := False;
  if not PieceImages.Dragging then
  begin
    Update;
    Result := Animator.Animate(Path, False);
  end;
end;

function TGameboardControl.AnimateBackward(const Path: array of TSquare): Boolean;
begin
  Result := False;
  if not PieceImages.Dragging then
  begin
    Update;
    Result := Animator.Animate(Path, True);
  end;
end;

function TGameboardControl.StartDrag(X, Y: Integer; TestOnly: Boolean): Boolean;
var
  Rect: TRect;
  Square: TSquare;
  ImageIndex: Integer;
begin
  Result := False;
  if not IsLegalSquare(DraggingSquare) and
    ((BoardMode = gbmSetup) or (Playing and not IsComputerTurn)) then
  begin
    Square := ClientToSquare(X, Y);
    if IsLegalSquare(Square) then
    begin
      ImageIndex := GetImageIndexOfSquare(Square);
      if ImageIndex >= 0 then
      begin
        Rect := SquareBounds[Square];
        InflateRect(Rect, (PieceImages.Width - SquareSize) div 2, (PieceImages.Height - SquareSize) div 2);
        if PtInRect(Rect, Point(X, Y)) then
        begin
          if not TestOnly then
          begin
            Application.CancelHint;
            fDraggingSquare := Square;
            PieceImages.SetDragImage(ImageIndex, X - Rect.Left, Y - Rect.Top);
            UpdateBoard(False);
            Update;
            PieceImages.BeginDrag(WindowHandle, X, Y);
            PieceImages.ShowDragImage;
            DoSquareDragStart;
          end;
          Result := True;
        end;
      end;
    end;
  end;
end;

function TGameboardControl.CanDrop(SourceSquare: TSquare; X, Y: Integer): Boolean;
var
  Square: TSquare;
begin
  Result := False;
  Square := ClientToSquare(X, Y);
  if SourceSquare = Square then
    Result := True
  else if IsLegalSquare(Square) then
    Result := (BoardMode = gbmSetup) or CanPlayPath([SourceSquare, Square]);
  DoSquareDragOver(SourceSquare, Square, Result);
end;

function TGameboardControl.StopDrag(X, Y: Integer; Cancelled: Boolean): Boolean;
var
  SrcSquare, DstSquare: TSquare;
begin
  Result := False;
  SrcSquare := DraggingSquare;
  if IsLegalSquare(DraggingSquare) then
  begin
    PieceImages.HideDragImage;
    PieceImages.EndDrag;
    fDraggingSquare := NO_SQUARE;
  end;
  if not Cancelled or CanDrop(SrcSquare, X, Y) then
  begin
    DstSquare := ClientToSquare(X, Y);
    if SrcSquare <> DstSquare then
    begin
      DoSquareDragDrop(SrcSquare, DstSquare);
      case BoardMode of
        gbmPlay:
        begin
          fDragedMove := True;
          try
            Result := PlayPath([SrcSquare, DstSquare]);
          finally
            fDragedMove := False;
          end;
        end;
        gbmSetup:
        begin
          BeginUpdate;
          try
            SetImageIndexOfSquare(DstSquare, GetImageIndexOfSquare(SrcSquare));
            SetImageIndexOfSquare(SrcSquare, -1);
            Result := True;
          finally
            EndUpdate;
          end;
        end;
      end;
    end;
  end;
  UpdateBoard(False);
  DoSquareDragEnd;
end;

function TGameboardControl.ApplyComputerMove(Index: Integer): Boolean;
begin
  Result := False;
  if not Paused and (BoardMode = gbmPlay) and IsComputerTurn then
  begin
    DoComputerMove(Index);
    if not ApplyLegalMove(Index) then
    begin
      DoComputerError;
      Result := False;
    end;
  end;
end;

function TGameboardControl.ApplyLegalMove(Index: Integer): Boolean;
begin
  Result := False;
  if (Index >= 0) and (Index < LegalMovesCount) then
  begin
    InitializeMovement;
    try
      Result := Gameboard.ApplyLegalMove(Index);
    finally
      FinalizeMovement;
    end;
  end;
end;

procedure TGameboardControl.InitializeMovement;
begin
  fMoving := True;
  UpdateBoard(False);
  DoMoveBegin;
end;

procedure TGameboardControl.FinalizeMovement;
begin
  fMoving := False;
  UpdateBoard(False);
  Update;
  if not CurrentHistoryRec.IsEmpty then
  begin
    CurrentHistoryRec.Assign(Gameboard);
    CurrentHistoryRec.ElapsedTime := ElapsedTime;
    if RecordFullMovesOnly then
      PostMessage(WindowHandle, WM_HISTORYCHANGE, 0, 0)
    else
      StoreCurrentHistoryRecord;
  end;
  DoMoveEnd;
  if PausedByHistory then
    Paused := False
  else
    CheckPlaying(False);
  if IsGameEnded then
  begin
    History.Closed := True;
    PostMessage(WindowHandle, WM_GAMEENDED, 0, 0);
  end
  else
    PostMessage(WindowHandle, WM_THINKABOUTMOVE, 0, 0);
end;

function TGameboardControl.StoreCurrentHistoryRecord: Boolean;
var
  HistoryRec: TGameHistoryRec;
begin
  Result := False;
  if not CurrentHistoryRec.IsEmpty then
  begin
    HistoryRec := TGameHistoryRec.Create;
    HistoryRec.Assign(CurrentHistoryRec);
    CurrentHistoryRec.Clear;
    History.Add(HistoryRec);
    Result := True;
  end;
end;

function TGameboardControl.RetrieveHistory(Index: Integer): Boolean;
var
  OldPlayer: TPlayer;
begin
  Result := False;
  if (Index >= 0) and (Index <= History.Count) and
     ((Index <> History.Current) or not CurrentHistoryRec.IsEmpty) then
  begin
    if not Paused then
    begin
      Paused := True;
      fPausedByHistory := True;
    end;
    BeginUpdate;
    try
      OldPlayer := CurrentPlayer;
      if AnimateMoves then
      begin
        if History.Current = Index then
          AnimateBackward(CurrentHistoryRec.MovementPath)
        else if History.Current = Index + 1 then
          AnimateBackward(History[Index].MovementPath)
        else if History.Current = Index - 1 then
          AnimateForward(History[Index - 1].MovementPath);
      end;
      if Index > 0 then
        Gameboard.Assign(History[Index - 1])
      else
        Gameboard.Assign(History.InitialBoard);
      if OldPlayer <> CurrentPlayer then
        PostMessage(WindowHandle, WM_TURNCHANGE, 0, 0);
      if History.Current <> Index then
        History.Current := Index
      else
        PostMessage(WindowHandle, WM_HISTORYCHANGE, 0, 0);
    finally
      EndUpdate;
    end;
    if PausedByHistory and (not IsComputerTurn or (History.Current = History.Count)) then
      Paused := False;
    Result := True;
  end;
end;

function TGameboardControl.HistoryOf(Player: TPlayer; TurnNo: Integer): TGameHistoryRec;
var
  Index: Integer;
begin
  Index := History.IndexOf(Player, TurnNo);
  if Index >= 0 then
    Result := History[Index]
  else if (Player = CurrentPlayer) and not CurrentHistoryRec.IsEmpty and (TurnNo = CurrentTurn) then
    Result := CurrentHistoryRec
  else
    Result := nil;
end;

function TGameboardControl.TimeOf(Player: TPlayer; TurnNo: Integer): Cardinal;
var
  HistoryRec: TGameHistoryRec;
  Index: Integer;
begin
  Result := 0;
  HistoryRec := HistoryOf(Player, TurnNo);
  if Assigned(HistoryRec) then
  begin
    Index := History.IndexOf(HistoryRec);
    if Index < 0 then
    begin
      Inc(Result, HistoryRec.ElapsedTime);
      Index := History.Count - 1;
      while (Index >= 0) and not History[Index].Completed do
      begin
        Inc(Result, History[Index].ElapsedTime);
        Dec(Index);
      end;
    end
    else
    begin
      while (Index < History.Count - 1) and not HistoryRec.Completed do
      begin
        Inc(Result, HistoryRec.ElapsedTime);
        Inc(Index);
        HistoryRec := History[Index];
      end;
      Inc(Result, HistoryRec.ElapsedTime);
    end;
  end;
end;

procedure TGameboardControl.StartNewGame;
begin
  Gameboard.StartNewGame;
  ClearHistory;
  BoardMode := gbmPlay;
  Modified := False;
  CheckPlaying(False);
end;

procedure TGameboardControl.ClearBoard;
begin
  BoardMode := gbmSetup;
  Gameboard.ClearBoard;
  Modified := False;
end;

procedure TGameboardControl.BeginUpdate;
begin
  Inc(fUpdateCount);
  Gameboard.BeginUpdate;
end;

procedure TGameboardControl.EndUpdate;
begin
  Gameboard.EndUpdate;
  Dec(fUpdateCount);
  if fUpdateCount = 0 then
  begin
    if not IsFullBoardReady then
      UpdateFullBoard
    else if not IsBoardReady then
      UpdateBoard(True);
  end;
end;

procedure TGameboardControl.SaveToStream(Stream: TStream);
var
  IsSolverSuspended: Boolean;
begin
  IsSolverSuspended := Solver.Suspended;
  Solver.Suspended := True;
  try
    History.SaveToStream(Stream);
    Modified := False;
  finally
    Solver.Suspended := IsSolverSuspended;
  end;
end;

procedure TGameboardControl.LoadFromStream(Stream: TStream);
var
  CurState: TMemoryStream;
  IsPaused: Boolean;
begin
  IsPaused := Paused;
  Paused := True;
  try
    History.BeginUpdate;
    try
      CurState := TMemoryStream.Create;
      try
        SaveToStream(CurState);
        try
          History.LoadFromStream(Stream);
          if History.Current > 0 then
            Gameboard.Assign(History[History.Current - 1])
          else
            Gameboard.Assign(History.InitialBoard);
        except
          LoadFromStream(CurState);
          raise;
        end;
      finally
        CurState.Free;
      end;
    finally
      History.EndUpdate;
    end;
    Modified := False;
  finally
    if CurrentPlayer = NONE then
      BoardMode := gbmSetup;
    Paused := IsPaused;
  end;
end;

procedure TGameboardControl.SaveToFile(const FileName: String;
  const Signature: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate or fmShareExclusive);
  try
    SaveStringToStream(Signature + FileVersion, Stream);
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TGameboardControl.LoadFromFile(const FileName: String;
  const Signature: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    if Signature + FileVersion <> LoadStringFromStream(Stream) then
      raise EInvalidGameFile.CreateFmt('Invalid %s File', [Signature]);
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TGameboardControl.IsGameEnded: Boolean;
begin
  Result := (Gameboard.EndState <> gesUnknown);
end;

function TGameboardControl.IsGameDraw: Boolean;
begin
  Result := (Gameboard.EndState = gesDraw);
end;

function TGameboardControl.CanMoveBackward: Boolean;
begin
  Result := (History.Current > 0) or not CurrentHistoryRec.IsEmpty;
end;

function TGameboardControl.MoveBackward: Boolean;
begin
  if not CurrentHistoryRec.IsEmpty then
    Result := RetrieveHistory(History.Current)
  else
    Result := RetrieveHistory(History.Current - 1);
end;

function TGameboardControl.MoveToFirst: Boolean;
begin
  Result := RetrieveHistory(0);
end;

function TGameboardControl.CanMoveForward: Boolean;
begin
  Result := (History.Current < History.Count);
end;

function TGameboardControl.MoveForward: Boolean;
begin
  Result := RetrieveHistory(History.Current + 1);
end;

function TGameboardControl.MoveToLast: Boolean;
begin
  Result := RetrieveHistory(History.Count);
end;

function TGameboardControl.CanRetrieve(Player: TPlayer; TurnNo: Integer): Boolean;
begin
  Result := (History.IndexOf(Player, TurnNo) >= 0)
end;

function TGameboardControl.Retrieve(Player: TPlayer; TurnNo: Integer): Boolean;
begin
  Result := RetrieveHistory(History.IndexOf(Player, TurnNo));
end;

procedure TGameboardControl.ClearHistory;
begin
  CurrentHistoryRec.Clear;
  History.BeginUpdate;
  try
    History.Clear;
    History.InitialBoard.Assign(Gameboard);
  finally
    History.EndUpdate;
  end;
end;

function TGameboardControl.IsComputerTurn: Boolean;
begin
  Result := ByteBool(ComputerPlayer and CurrentPlayer);
end;

function TGameboardControl.PlayComputerTurnNow: Boolean;
begin
  Result := Playing and IsComputerTurn and StopThinking(False);
end;

function TGameboardControl.CancelThinking: Boolean;
begin
  Result := Playing and IsComputerTurn and StopThinking(True);
end;

function TGameboardControl.SuggestMove(ASearchDepth: ShortInt): Integer;
begin
  Result := -1;
  if not (IsComputerTurn or IsGameEnded) then
  begin
    if ASearchDepth < 0 then
      ASearchDepth := SearchDepth
    else if ASearchDepth > MAX_SEARCH_DEPTH then
      ASearchDepth := MAX_SEARCH_DEPTH;
    Result := Gameboard.SearchForBestMove(ASearchDepth, MaxMemory);
  end;
end;

function TGameboardControl.PlayerScore: Integer;
begin
  Result := Gameboard.Score;
end;

function TGameboardControl.MoveScore(MoveIndex: Integer): Integer;
var
  GB: TGameboard;
begin
  Result := 0;
  if (MoveIndex >= 0) and (MoveIndex < LegalMovesCount) then
  begin
    GB := GameboardClass.Create;
    try
      GB.Transpose(Gameboard, MoveIndex);
      Result := -GB.Score;
    finally
      GB.Free;
    end;
  end;
end;

end.
