//--------------------------------------------------------------------------
//  Othello Gameboad (Reversi)
//  Copyright (C) 2005-2009 Kambiz R. Khojasteh
//--------------------------------------------------------------------------

{$B-} // Complete Boolean Evaluation Off
{$R-} // Range Checking Off
{$Q-} // Overflow Checking Off
{$V-} // Var String Checking Off
{$T-} // Typed @ operator Off
{$X+} // Extended Syntax On
{$P+} // Open String Params On
{$O+} // Optimization On
{$H+} // Use long strings by default

unit OthelloBoard;

interface

uses
  Windows, Messages, Classes, Graphics, Controls, SysUtils, Gameboard;

const
  EMPTY      = $00;
  BLACK      = DARK_PLAYER;
  WHITE      = LIGHT_PLAYER;

type

  TPosition = (A1, B1, C1, D1, E1, F1, G1, H1,
               A2, B2, C2, D2, E2, F2, G2, H2,
               A3, B3, C3, D3, E3, F3, G3, H3,
               A4, B4, C4, D4, E4, F4, G4, H4,
               A5, B5, C5, D5, E5, F5, G5, H5,
               A6, B6, C6, D6, E6, F6, G6, H6,
               A7, B7, C7, D7, E7, F7, G7, H7,
               A8, B8, C8, D8, E8, F8, G8, H8);

  QWORD = Int64;

  TBoard = record // LSB: A1; MSB: H8
    Blacks: QWORD;
    Whites: QWORD;
  end;

  PMovement = ^TMovement;
  TMovement = record
    NullMove: Boolean;
    Pos: TPosition;
  end;

  TMoveList = class(TObject)
  private
    fList: PMovement;
    fCount: Integer;
    fCapacity: Integer;
    procedure SetCapacity(Value: Integer);
    function GetMovePtr(Index: Integer): PMovement; inline;
    function GetMoves(Index: Integer): TMovement;
    procedure Grow; inline;
  protected
    procedure Clear; inline;
    procedure AddNullMove; inline;
    procedure Add(Pos: TPosition); overload; inline;
    procedure Add(Mobility: QWORD); overload; inline;
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    property Capacity: Integer read fCapacity write SetCapacity;
    property MovePtr[Index: Integer]: PMovement read GetMovePtr;
    property FirstMove: PMovement read fList;
  public
    destructor Destroy; override;
    function IndexOf(const Move: TMovement): Integer;
    function MoveExists(const Move: TMovement): Boolean; overload;
    function MoveExists(Pos: TPosition): Boolean; overload;
    property Moves[Index: Integer]: TMovement read GetMoves; default;
    property Count: Integer read fCount;
  end;

  TPositionInfoTipEvent = procedure(Sender: TObject; Pos: TPosition; var InfoTip: String) of object;
  TPositionActionEvent = procedure(Sender: TObject; Pos: TPosition; var Handled: Boolean) of object;
  TPositionEvent = procedure(Sender: TObject; Pos: TPosition) of object;

  TOthelloBoard = class(TGameboard)
  private
    fBoard: TBoard;
    fLegalMoves: TMoveList;
    fBlackPieceCount: Byte;
    fWhitePieceCount: Byte;
    fKey: TPositionKey;
    fPrepared: Boolean;
    fOnCapture: TPositionEvent;
    fOnMove: TPositionEvent;
    fOnNullMove: TNotifyEvent;
    procedure SetBoard(const Value: TBoard);
    function GetPieces(Pos: TPosition): TPiece;
    procedure SetPieces(Pos: TPosition; Value: TPiece);
    function GetLegalMoves: TMoveList;
  protected
    procedure DoTurn; override;
    procedure DoChange; override;
    function GetDistinctPieceCount: Byte; override;
    function GetEndState: TGameEndState; override;
    function GetLegalMovesCount: Integer; override;
    function GetBoardAsText: String; override;
    procedure SetBoardAsText(const Value: String); override;
  protected
    function CalculateKey: TPositionKey;
    function GetKey: TPositionKey; override;
    function GetScore: Integer; override;
    function GetAvgMoveCount: Cardinal; override;
    class function GetPositionTable: TPositionTable; override;
    procedure InitializeSearch; override;
    procedure Transpose(Gameboard: TGameboard; TargetMove: Integer); override;
  protected
    procedure DoNullMove; virtual;
    procedure DoMove(Pos: TPosition); virtual;
    procedure DoCapture(Pos: TPosition); virtual;
    procedure PrepareLegalMoves;
    property Prepared: Boolean read fPrepared;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure SaveToStream(Stream: TStream); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure GetTextPresentation(out Board, Moves: String); override;
    procedure StartNewGame; override;
    function ApplyLegalMove(Index: Integer): Boolean; override;
    function ApplyMove(const Move: TMovement): Boolean; overload;
    function ApplyMove(Pos: TPosition): Boolean; overload;
    function ApplyNullMove: Boolean;
    property BlackPieceCount: Byte read fBlackPieceCount;
    property WhitePieceCount: Byte read fWhitePieceCount;
    property LegalMoves: TMoveList read GetLegalMoves;
    property Board: TBoard read fBoard write SetBoard;
    property Pieces[Pos: TPosition]: TPiece read GetPieces write SetPieces; default;
    property OnNullMove: TNotifyEvent read fOnNullMove write fOnNullMove;
    property OnMove: TPositionEvent read fOnMove write fOnMove;
    property OnCapture: TPositionEvent read fOnCapture write fOnCapture;
  end;

  TOthello = class(TGameboardControl)
  private
    fPiecesBitmap: TBitmap;
    fOnPositionClick: TPositionActionEvent;
    fOnPositionInfoTip: TPositionInfoTipEvent;
    fOnMove: TPositionEvent;
    fOnNullMove: TNotifyEvent;
    fOnCapture: TPositionEvent;
    DefaultPieceSize: Integer;
    function GetBoard: TBoard;
    procedure SetBoard(const Value: TBoard);
    function GetPieces(Pos: TPosition): TPiece;
    procedure SetPieces(Pos: TPosition; Value: TPiece);
    procedure SetPiecesBitmap(Value: TBitmap);
    function GetPosRect(Pos: TPosition): TRect;
    function GetSelectedMove: TMovement;
    function GetLegalMoves: TMoveList;
    function GetOthelloBoard: TOthelloBoard;
    function GetCoordinatesDisplay: TBoardLabels;
    procedure SetCoordinatesDisplay(Value: TBoardLabels);
    function GetPositionScoreDisplay: TSquareLabel;
    procedure SetPositionScoreDisplay(Value: TSquareLabel);
    function GetCoordinatesOffset: Integer;
    procedure SetCoordinatesOffset(Value: Integer);
    function GetSquareColor: TColor;
    procedure SetSquareColor(Value: TColor);
    function GetSquareMarkColor: TColor;
    procedure SetSquareMarkColor(Value: TColor);
    function GetSquareMarkPattern: TPicture;
    procedure SetSquareMarkPattern(Value: TPicture);
    function GetSquareSelectColor: TColor;
    procedure SetSquareSelectColor(Value: TColor);
    function GetSquareSelectPattern: TPicture;
    procedure SetSquareSelectPattern(Value: TPicture);
    function GetSquarePatern: TPicture;
    procedure SetSquarePattern(Value: TPicture);
    function GetPieceCount(Player: TPlayer): Byte;
    procedure PiecesBitmapChange(Sender: TObject);
    procedure PieceNullMove(Sender: TObject);
    procedure PieceMove(Sender: TObject; Pos: TPosition);
    procedure PieceCapture(Sender: TObject; Pos: TPosition);
  protected
    procedure DoSquareResize; override;
    procedure DoSquareClick(Square: TSquare; var Handled: Boolean); override;
    procedure DoPositionClick(Pos: TPosition; var Handled: Boolean); virtual;
    procedure DoSquareInfoTip(Square: TSquare; var InfoTip: String); override;
    procedure DoPositionInfoTip(Pos: TPosition; var InfoTip: String); virtual;
    procedure DoNullMove; virtual;
    procedure DoMove(Pos: TPosition); virtual;
    procedure DoCapture(Pos: TPosition); virtual;
    function GetGameboardClass: TGameboardClass; override;
    procedure SetImageIndexOfSquare(Square: TSquare; ImageIndex: Integer); override;
    function GetImageIndexOfSquare(Square: TSquare): Integer; override;
    function GetSquareLabel(Square: TSquare): String; override;
    function GetStateOfSquare(Square: TSquare): TSquareDrawStates; override;
    function IsPathComplete(const Path: array of TSquare): Boolean; override;
    function CanPlayPath(const Path: array of TSquare): Boolean; override;
    function PlayPath(const Path: array of TSquare): Boolean; override;
    function CompletePath(var Path: TMovementPath): Boolean; override;
    function IsVisualDarkSquare(Square: TSquare): Boolean; override;
    function CanSelectSquare(Square: TSquare): Boolean; override;
    function UpdatePiecesImages: Boolean; virtual;
    function PosToSquare(Pos: TPosition): TSquare;
    function SquareToPos(Square: TSquare; out Pos: TPosition): Boolean;
    function MoveToPath(const Move: TMovement): TMovementPath;
    function PathToMove(const Path: array of TSquare; out Move: TMovement): Boolean;
    property OthelloBoard: TOthelloBoard read GetOthelloBoard;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function SelectMove(Index: Integer): Boolean; overload;
    function SelectMove(const Move: TMovement): Boolean; overload;
    function SelectPosition(Pos: TPosition): Boolean;
    function CanSelectPosition(Pos: TPosition): Boolean;
    function IsPositionSelected(Pos: TPosition): Boolean;
    function IsLegalMove(const Move: TMovement): Boolean;
    function Play(const Move: TMovement): Boolean;
    function PlayNullMove: Boolean;
    function MustPlayNull: Boolean;
    function MoveOf(Player: TPlayer; TurnNo: Integer): String; override;
    function ClientToPos(X, Y: Integer; out Pos: TPosition): Boolean;
    property PieceCount[Player: TPlayer]: Byte read GetPieceCount;
    property PosRect[Pos: TPosition]: TRect read GetPosRect;
    property SelectedMove: TMovement read GetSelectedMove;
    property LegalMoves: TMoveList read GetLegalMoves;
    property Board: TBoard read GetBoard write SetBoard;
    property Pieces[Pos: TPosition]: TPiece read GetPieces write SetPieces;
  published
    property Align;
    property Anchors;
    property AutoSize;
    property BoardAngle;
    property BoardAsText;
    property BoardBorderColor;
    property BoardMarginBottom default 20;
    property BoardMarginColor;
    property BoardMarginLeft default 20;
    property BoardMarginRight default 20;
    property BoardMarginTop default 20;
    property BoardMode;
    property Center;
    property Color;
    property ComputerPlayer default WHITE;
    property Constraints;
    property CoordinatesDisplay: TBoardLabels read GetCoordinatesDisplay write SetCoordinatesDisplay default [blLeft, blTop];
    property CoordinatesOffset: Integer read GetCoordinatesOffset write SetCoordinatesOffset default 4;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FixedSquareSize;
    property Font;
    property Hint;
    property MarkerOffset;
    property MarkerSize;
    property MarkSelectableSquares;
    property MaxMemory;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property Paused;
    property PiecesBitmap: TBitmap read fPiecesBitmap write SetPiecesBitmap;
    property PositionScoreDisplay: TSquareLabel read GetPositionScoreDisplay write SetPositionScoreDisplay;
    property SearchDepth;
    property ShowHint;
    property SquareBorderColor;
    property SquareColor: TColor read GetSquareColor write SetSquareColor default $00008000;
    property SquareMarkColor: TColor read GetSquareMarkColor write SetSquareMarkColor default clLime;
    property SquareMarkPattern: TPicture read GetSquareMarkPattern write SetSquareMarkPattern;
    property SquareSelectColor: TColor read GetSquareSelectColor write SetSquareSelectColor default clLime;
    property SquareSelectPattern: TPicture read GetSquareSelectPattern write SetSquareSelectPattern;
    property SquarePattern: TPicture read GetSquarePatern write SetSquarePattern;
    property SquareSize;
    property TabOrder;
    property TabStop;
    property ThinkCursor;
    property ThinkPriority;
    property Visible;
    property OnBoardChange;
    property OnBoardModeChange;
    property OnCanResize;
    property OnCapture: TPositionEvent read fOnCapture write fOnCapture;
    property OnClick;
    property OnComputerError;
    property OnComputerMove;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGameEnded;
    property OnHistoryChange;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMove: TPositionEvent read fOnMove write fOnMove;
    property OnMoveBegin;
    property OnMoveEnd;
    property OnNullMove: TNotifyEvent read fOnNullMove write fOnNullMove;
    property OnPauseStateChange;
    property OnPlayingStateChange;
    property OnPositionClick: TPositionActionEvent read fOnPositionClick write fOnPositionClick;
    property OnPositionInfoTip: TPositionInfoTipEvent read fOnPositionInfoTip write fOnPositionInfoTip;
    property OnResize;
    property OnSelectionChange;
    property OnSquareResize;
    property OnStartDock;
    property OnStartDrag;
    property OnThinkBegin;
    property OnThinkEnd;
    property OnTurnChange;
  end;

function PosToStr(Pos: TPosition): String;
function StrToPos(const S: String; out Pos: TPosition): Boolean;

function MoveToStr(const Move: TMovement): String;
function StrToMove(const S: String; out Move: TMovement): Boolean;

procedure PosToCoordinate(Pos: TPosition; out Row, Col: Integer);
function CoordinateToPos(Row, Col: Integer; out Pos: TPosition): Boolean;

function HasLegalMove(const Board: TBoard; Player: TPlayer): Boolean;
procedure FindLegalMoves(Moves: TMoveList; const Board: TBoard; Player: TPlayer);

implementation

{$R *.res}

var
  PositionTable: TPositionTable;

var
  Zobrist: array[TPosition] of array[WHITE..BLACK] of TPositionKey;

type
  TFlipInfo = record
    TopLeftSide: Boolean;
    Direction: ShortInt;
    InitMask: QWORD;
    Mask: QWORD;
  end;

const
  FlipInfo: array[1..8] of TFlipInfo = (
    { Delta Row =  0; Delta Column = -1 }
    (TopLeftSide: True;  Direction: 1; InitMask: QWORD($3F3F3F3F3F3F3F3F); Mask: QWORD($7F7F7F7F7F7F7F7F)),
    { Delta Row =  0; Delta Column = +1 }
    (TopLeftSide: False; Direction: 1; InitMask: QWORD($FCFCFCFCFCFCFCFC); Mask: QWORD($FEFEFEFEFEFEFEFE)),
    { Delta Row = -1; Delta Column =  0 }
    (TopLeftSide: True;  Direction: 8; InitMask: QWORD($0000FFFFFFFFFFFF); Mask: QWORD($00FFFFFFFFFFFFFF)),
    { Delta Row = +1; Delta Column =  0 }
    (TopLeftSide: False; Direction: 8; InitMask: QWORD($FFFFFFFFFFFF0000); Mask: QWORD($FFFFFFFFFFFFFF00)),
    { Delta Row = -1; Delta Column = -1 }
    (TopLeftSide: True;  Direction: 9; InitMask: QWORD($00003F3F3F3F3F3F); Mask: QWORD($007F7F7F7F7F7F7F)),
    { Delta Row = +1; Delta Column = +1 }
    (TopLeftSide: False; Direction: 9; InitMask: QWORD($FCFCFCFCFCFC0000); Mask: QWORD($FEFEFEFEFEFEFE00)),
    { Delta Row = -1; Delta Column = +1 }
    (TopLeftSide: True;  Direction: 7; InitMask: QWORD($0000FCFCFCFCFCFC); Mask: QWORD($00FEFEFEFEFEFEFE)),
    { Delta Row = +1; Delta Column = -1 }
    (TopLeftSide: False; Direction: 7; InitMask: QWORD($3F3F3F3F3F3F0000); Mask: QWORD($7F7F7F7F7F7F7F00)));

{ Helper Functions }

function PosToStr(Pos: TPosition): String;
begin
  Result := Chr(Ord('A') + (Ord(Pos) mod 8)) + IntToStr((Ord(Pos) div 8) + 1);
end;

function StrToPos(const S: String; out Pos: TPosition): Boolean;
var
  R, C: Integer;
begin
  Result := False;
  if Length(S) = 2 then
  begin
    C := Ord(UpCase(S[1])) - Ord('A');
    R := Ord(UpCase(S[2])) - Ord('1');
    Result := CoordinateToPos(R, C, Pos);
  end;
end;

function MoveToStr(const Move: TMovement): String;
begin
  if not Move.NullMove then
    Result := PosToStr(Move.Pos)
  else
    Result := '--';
end;

function StrToMove(const S: String; out Move: TMovement): Boolean;
begin
  if S = '--' then
  begin
    Move.NullMove := True;
    Result := True;
  end
  else
  begin
    Move.NullMove := False;
    Result := StrToPos(S, Move.Pos);
  end;
end;

procedure PosToCoordinate(Pos: TPosition; out Row, Col: Integer);
begin
  Row := Ord(Pos) shr 3;
  Col := Ord(Pos) and 7;
end;

function CoordinateToPos(Row, Col: Integer; out Pos: TPosition): Boolean;
begin
  if (Cardinal(Row) <= 7) and (Cardinal(Col) <= 7) then
  begin
    Pos := TPosition(Row shl 3 + Col);
    Result := True;
  end
  else
    Result := False;
end;

function BitCountOf(Value: QWORD): Integer; inline;
begin
  Result := 0;
  while Value <> 0 do
  begin
    while not ByteBool(Value) do
      Value := Value shr 8;
    if ByteBool(Value and 1) then
      Inc(Result);
    Value := Value shr 1;
  end;
end;

function HasLegalMove(const Board: TBoard; Player: TPlayer): Boolean;

  function HasMobility(SideCells, EnemyCells: QWORD): Boolean;
  var
    EmptyCells: QWORD;
    Cells: QWORD;
    I: Integer;
  begin
    Result := True;
    EmptyCells := not (SideCells or EnemyCells);
    for I := Low(FlipInfo) to High(FlipInfo) do
      with FlipInfo[I] do
      begin
        if TopLeftSide then
        begin
          Cells := ((SideCells and InitMask) shl Direction) and EnemyCells;
          while Cells <> 0 do
          begin
            Cells := (Cells shl Direction) and not SideCells;
            if Cells = 0 then Break;
            if (Cells and EmptyCells) <> 0 then Exit;
            Cells := Cells and Mask;
          end;
        end
        else
        begin
          Cells := ((SideCells and InitMask) shr Direction) and EnemyCells;
          while Cells <> 0 do
          begin
            Cells := (Cells shr Direction) and not SideCells;
            if Cells = 0 then Break;
            if (Cells and EmptyCells) <> 0 then Exit;
            Cells := Cells and Mask;
          end;
        end
      end;
    Result := False;
  end;

begin
  with Board do
  begin
    case Player of
      BLACK: Result := HasMobility(Blacks, Whites);
      WHITE: Result := HasMobility(Whites, Blacks);
    else
      Result := False;
    end;
  end;
end;

procedure FindLegalMoves(Moves: TMoveList; const Board: TBoard; Player: TPlayer);

  function FindMobility(SideCells, EnemyCells: QWORD): QWORD;
  var
    EmptyCells: QWORD;
    Cells: QWORD;
    I: Integer;
  begin
    Result := 0;
    EmptyCells := not (SideCells or EnemyCells);
    for I := Low(FlipInfo) to High(FlipInfo) do
      with FlipInfo[I] do
      begin
        if TopLeftSide then
        begin
          Cells := ((SideCells and InitMask) shl Direction) and EnemyCells;
          while Cells <> 0 do
          begin
            Cells := (Cells shl Direction) and not SideCells;
            if Cells = 0 then Break;
            Result := Result or (Cells and EmptyCells);
            Cells := not Result and Cells and Mask;
          end;
        end
        else
        begin
          Cells := ((SideCells and InitMask) shr Direction) and EnemyCells;
          while Cells <> 0 do
          begin
            Cells := (Cells shr Direction) and not SideCells;
            if Cells = 0 then Break;
            Result := Result or (Cells and EmptyCells);
            Cells := not Result and Cells and Mask;
          end;
        end
      end;
  end;

begin
  Moves.Clear;
  with Board do
  begin
    case Player of
      BLACK: Moves.Add(FindMobility(Blacks, Whites));
      WHITE: Moves.Add(FindMobility(Whites, Blacks));
    end;
  end;
end;

{ TMoveList }

destructor TMoveList.Destroy;
begin
  SetCapacity(0);
  inherited Destroy;
end;

procedure TMoveList.SetCapacity(Value: Integer);
begin
  if fCapacity <> Value then
  begin
    fCapacity := Value;
    ReallocMem(fList, fCapacity * SizeOf(TMovement));
    if fCount > fCapacity then
      fCount := fCapacity;
  end;
end;

function TMoveList.GetMovePtr(Index: Integer): PMovement;
begin
  Result := fList;
  Inc(Result, Index);
end;

function TMoveList.GetMoves(Index: Integer): TMovement;
begin
  if (Index >= 0) and (Index < Count) then
    Result := MovePtr[Index]^
  else
    Result.NullMove := True;
end;

procedure TMoveList.Grow;
begin
  Inc(fCapacity, 32);
  ReallocMem(fList, fCapacity * SizeOf(TMovement));
end;

procedure TMoveList.Clear;
begin
  fCount := 0;
end;

procedure TMoveList.AddNullMove;
var
  Item: PMovement;
begin
  if fCount = fCapacity then Grow;
  Item := fList;
  Inc(Item, fCount);
  Item^.NullMove := True;
  Inc(fCount);
end;

procedure TMoveList.Add(Pos: TPosition);
var
  Item: PMovement;
begin
  if fCount = fCapacity then Grow;
  Item := fList;
  Inc(Item, fCount);
  Item^.NullMove := False;
  Item^.Pos := Pos;
  Inc(fCount);
end;

procedure TMoveList.Add(Mobility: QWORD);
var
  Pos: TPosition;
begin
  Pos := Low(TPosition);
  while Mobility <> 0 do
  begin
    while Byte(Mobility) = 0 do
    begin
      Mobility := Mobility shr 8;
      Inc(Pos, 8);
    end;
    if (Mobility and 1) <> 0 then
      Add(Pos);
    Mobility := Mobility shr 1;
    Inc(Pos);
  end;
end;

function TMoveList.IndexOf(const Move: TMovement): Integer;
var
  Item: PMovement;
  Index: Integer;
begin
  Result := -1;
  Item := fList;
  for Index := 0 to fCount - 1 do
  begin
    if (Item^.NullMove = Move.NullMove) and (Move.NullMove or (Item^.Pos = Move.Pos)) then
    begin
      Result := Index;
      Exit;
    end;
    Inc(Item);
  end;
end;

function TMoveList.MoveExists(const Move: TMovement): Boolean;
begin
  Result := (IndexOf(Move) >= 0);
end;

function TMoveList.MoveExists(Pos: TPosition): Boolean;
var
  Move: TMovement;
begin
  Move.NullMove := False;
  Move.Pos := Pos;
  Result := MoveExists(Move);
end;

procedure TMoveList.LoadFromStream(Stream: TStream);
begin
  Stream.Read(fCount, SizeOf(fCount));
  if fCount > 0 then
  begin
    if fCount > fCapacity then
      SetCapacity(fCount);
    Stream.Read(fList^, fCount * SizeOf(TPosition));
  end;
end;

procedure TMoveList.SaveToStream(Stream: TStream);
begin
  Stream.Write(fCount, SizeOf(fCount));
  if fCount > 0 then
    Stream.Write(fList^, fCount * SizeOf(TMovement));
end;

{ TOthelloBoard }

constructor TOthelloBoard.Create;
begin
  inherited Create;
  fLegalMoves := TMoveList.Create;
end;

destructor TOthelloBoard.Destroy;
begin
  fLegalMoves.Free;
  inherited Destroy;
end;

procedure TOthelloBoard.SetBoard(const Value: TBoard);
begin
  if not CompareMem(@fBoard, @Value, SizeOf(TBoard)) then
  begin
    fBoard := Value;
    fBlackPieceCount := BitCountOf(Value.Blacks);
    fWhitePieceCount := BitCountOf(Value.Whites);
    DoChange;
  end;
end;

function TOthelloBoard.GetPieces(Pos: TPosition): TPiece;
var
  Mask: QWORD;
begin
  Mask := QWORD(1) shl Ord(Pos);
  with fBoard do
  begin
    if (Blacks and Mask) <> 0 then
      Result := BLACK
    else if (Whites and Mask) <> 0 then
      Result := WHITE
    else
      Result := EMPTY;
  end;
end;

procedure TOthelloBoard.SetPieces(Pos: TPosition; Value: TPiece);
var
  Mask: QWORD;
  OldValue: TPiece;
begin
  OldValue := Pieces[Pos];
  if OldValue <> Value then
  begin
    Mask := QWORD(1) shl Ord(Pos);
    with fBoard do
    begin
      case OldValue of
        BLACK:
        begin
          Dec(fBlackPieceCount);
          Blacks := Blacks and not Mask;
        end;
        WHITE:
        begin
          Dec(fWhitePieceCount);
          Whites := Whites and not Mask;
        end;
      end;
      case Value of
        BLACK:
        begin
          Inc(fBlackPieceCount);
          Blacks := Blacks or Mask;
        end;
        WHITE:
        begin
          Inc(fWhitePieceCount);
          Whites := Whites or Mask;
        end;
      end;
    end;
    DoChange;
  end;
end;

function TOthelloBoard.GetLegalMoves: TMoveList;
begin
  if not Prepared then
    PrepareLegalMoves;
  Result := fLegalMoves;
end;

function TOthelloBoard.GetLegalMovesCount: Integer;
begin
  Result := LegalMoves.Count;
end;

function TOthelloBoard.GetBoardAsText: String;
var
  Pos: TPosition;
  S: PChar;
begin
  SetLength(Result, 64);
  S := PChar(Result);
  for Pos := Low(TPosition) to High(TPosition) do
  begin
    case Pieces[Pos] of
      WHITE:
        S^ := 'W';
      BLACK:
        S^ := 'B';
    else
      S^ := ' ';
    end;
    Inc(S);
  end;
end;

procedure TOthelloBoard.SetBoardAsText(const Value: String);
var
  Pos: TPosition;
  S: PChar;
begin
  BeginUpdate;
  try
    S := PChar(Value);
    for Pos := Low(TPosition) to High(TPosition) do
      if S^ <> #0 then
      begin
        case S^ of
          'W': Pieces[Pos] := WHITE;
          'B': Pieces[Pos] := BLACK;
        else
          Pieces[Pos] := EMPTY;
        end;
        Inc(S);
      end
      else
        Pieces[Pos] := EMPTY;
  finally
    EndUpdate;
  end;
end;

function TOthelloBoard.GetEndState: TGameEndState;
begin
  Result := gesUnknown;
  if LegalMoves.Count = 0 then
  begin
    if BlackPieceCount > WhitePieceCount then
      Result := gesDarkWon
    else if BlackPieceCount < WhitePieceCount then
      Result := gesLightWon
    else
      Result := gesDraw;
  end;
end;

function TOthelloBoard.GetDistinctPieceCount: Byte;
begin
  Result := 1;
end;

function TOthelloBoard.GetAvgMoveCount: Cardinal;
begin
  Result := 10;
end;

function TOthelloBoard.GetKey: TPositionKey;
begin
  Result := fKey;
end;

function TOthelloBoard.GetScore: Integer;
var
  Mask: QWORD;
  Pos: TPosition;
  N: Integer;
begin
  Result := 0;
  Mask := 1;
  for Pos := Low(TPosition) to High(TPosition) do
  begin
    with fBoard do
    begin
      if (Blacks and Mask) <> 0 then
      begin
        N := 1;
        if (Mask and $FF000000000000FF) <> 0 then
          N := N shl 2;
        if (Mask and $8181818181818181) <> 0 then
          N := N shl 2;
        Inc(Result, N);
      end
      else if (Whites and Mask) <> 0 then
      begin
        N := 1;
        if (Mask and $FF000000000000FF) <> 0 then
          N := N shl 2;
        if (Mask and $8181818181818181) <> 0 then
          N := N shl 2;
        Dec(Result, N);
      end;
    end;
    Mask := Mask shl 1;
  end;
  if Player = WHITE then
    Result := -Result;
end;

class function TOthelloBoard.GetPositionTable: TPositionTable;
begin
  Result := PositionTable;
end;

procedure TOthelloBoard.Transpose(Gameboard: TGameboard; TargetMove: Integer);

  function DoFlip(Pos: TPosition; const SideMask: QWORD; Side, Enemy: TPlayer;
    var SideCells, EnemyCells: QWORD): Integer;
  var
    EnemyPos: TPosition;
    EnemyMask: QWORD;
    I: Integer;
  begin
    Result := 0;
    for I := Low(FlipInfo) to High(FlipInfo) do
      with FlipInfo[I] do
        if TopLeftSide then
        begin
          EnemyPos := Pos;
          EnemyMask := (SideMask and Mask) shl Direction;
          while (EnemyCells and EnemyMask and Mask) <> 0 do
          begin
            EnemyMask := EnemyMask shl Direction;
            Inc(EnemyPos, Direction);
          end;
          if (EnemyPos <> Pos) and ((SideCells and EnemyMask) <> 0) then
            repeat
              EnemyMask := EnemyMask shr Direction;
              EnemyCells := EnemyCells and not EnemyMask;
              SideCells := SideCells or EnemyMask;
              Inc(Result);
              fKey := fKey xor Zobrist[EnemyPos][Enemy] xor Zobrist[EnemyPos][Side];
              Dec(EnemyPos, Direction);
            until EnemyPos = Pos;
        end
        else
        begin
          EnemyPos := Pos;
          EnemyMask := (SideMask and Mask) shr Direction;
          while (EnemyCells and EnemyMask and Mask) <> 0 do
          begin
            EnemyMask := EnemyMask shr Direction;
            Dec(EnemyPos, Direction);
          end;
          if (EnemyPos <> Pos) and ((SideCells and EnemyMask) <> 0) then
            repeat
              EnemyMask := EnemyMask shl Direction;
              EnemyCells := EnemyCells and not EnemyMask;
              SideCells := SideCells or EnemyMask;
              Inc(Result);
              fKey := fKey xor Zobrist[EnemyPos][Enemy] xor Zobrist[EnemyPos][Side];
              Inc(EnemyPos, Direction);
            until EnemyPos = Pos;
        end;
  end;

var
  Move: PMovement;
  Mask: QWORD;
  FlipCount: Integer;
begin
  fPrepared := False;
  with TOthelloBoard(Gameboard) do
  begin
    Self.fBoard := fBoard;
    Self.fBlackPieceCount := fBlackPieceCount;
    Self.fWhitePieceCount := fWhitePieceCount;
    Self.ActivePlayer := Player xor BOTH_PLAYERS;
    Self.fKey := fKey xor PlayerZobristMask;
    Move := fLegalMoves.MovePtr[TargetMove];
  end;
  with Move^, fBoard do
    if not NullMove then
    begin
      Mask := QWORD(1) shl Ord(Pos);
      if ActivePlayer = WHITE then
      begin
        Blacks := Blacks or Mask;
        fKey := fKey xor Zobrist[Pos][BLACK];
        Inc(fBlackPieceCount);
        FlipCount := DoFlip(Pos, Mask, BLACK, WHITE, Blacks, Whites);
        Dec(fWhitePieceCount, FlipCount);
        Inc(fBlackPieceCount, FlipCount);
      end
      else
      begin
        Whites := Whites or Mask;
        fKey := fKey xor Zobrist[Pos][WHITE];
        Inc(fWhitePieceCount);
        FlipCount := DoFlip(Pos, Mask, WHITE, BLACK, Whites, Blacks);
        Dec(fBlackPieceCount, FlipCount);
        Inc(fWhitePieceCount, FlipCount);
      end;
    end;
end;

function TOthelloBoard.CalculateKey: TPositionKey;
var
  Pos: TPosition;
  Mask: QWORD;
begin
  Result := 0;
  Mask := 1;
  for Pos := Low(TPosition) to High(TPosition) do
  begin
    with fBoard do
    begin
      if (Blacks and Mask) <> 0 then
        fKey := fKey xor Zobrist[Pos][BLACK]
      else if (Whites and Mask) <> 0 then
        fKey := fKey xor Zobrist[Pos][WHITE];
      Mask := Mask shl 1;
    end;
  end;
  if Player = BLACK then
    fKey := fKey xor PlayerZobristMask;
end;

procedure TOthelloBoard.InitializeSearch;
begin
  fKey := CalculateKey;
  inherited InitializeSearch;
end;

procedure TOthelloBoard.GetTextPresentation(out Board, Moves: String);
var
  R, C, M: Integer;
  Pos: TPosition;
begin
  Board := '  A B C D E F G H' + #13#10;
  for R := 0 to 7 do
  begin
    if R <> 0 then
      Board := Board + #13#10;
    Board := Board + IntToStr(R + 1);
    for C := 0 to 7 do
    begin
      Board := Board + ' ';
      CoordinateToPos(R, C, Pos);
      case Pieces[Pos] of
        BLACK: Board := Board + 'B';
        WHITE: Board := Board + 'W';
      else
        Board := Board + '-';
      end;
    end;
  end;
  Moves := '';
  for M := 0 to LegalMoves.Count - 1 do
  begin
    if M <> 0 then
      Moves := Moves + #13#10;
    Moves := Moves + Format('%2d) %s', [M + 1, MoveToStr(LegalMoves[M])]);
  end;
end;

procedure TOthelloBoard.DoTurn;
begin
  fPrepared := False;
  inherited DoTurn;
end;

procedure TOthelloBoard.DoChange;
begin
  fPrepared := False;
  inherited DoChange;
end;

procedure TOthelloBoard.DoNullMove;
begin
  if Assigned(OnNullMove) then
    OnNullMove(Self);
end;

procedure TOthelloBoard.DoMove(Pos: TPosition);
begin
  if Assigned(OnMove) then
    OnMove(Self, Pos);
end;

procedure TOthelloBoard.DoCapture(Pos: TPosition);
begin
  if Assigned(OnCapture) then
    OnCapture(Self, Pos);
end;

procedure TOthelloBoard.PrepareLegalMoves;
begin
  if not Prepared then
  begin
    fPrepared := True;
    FindLegalMoves(fLegalMoves, fBoard, Player);
    if (fLegalMoves.Count = 0) and HasLegalMove(fBoard, Opponent(Player)) then
      fLegalMoves.AddNullMove;
  end;
end;

procedure TOthelloBoard.StartNewGame;
var
  Pos: TPosition;
begin
  BeginUpdate;
  try
    Player := BLACK;
    for Pos := Low(TPosition) to High(TPosition) do
      case Pos of
        E4, D5: Pieces[Pos] := BLACK;
        D4, E5: Pieces[Pos] := WHITE;
      else
        Pieces[Pos] := EMPTY;
      end;
  finally
    EndUpdate;
  end;
end;

procedure TOthelloBoard.SaveToStream(Stream: TStream);
begin
  inherited SaveToStream(Stream);
  Stream.Write(fBoard, SizeOf(fBoard));
  Stream.Write(fBlackPieceCount, SizeOf(fBlackPieceCount));
  Stream.Write(fWhitePieceCount, SizeOf(fWhitePieceCount));
end;

procedure TOthelloBoard.LoadFromStream(Stream: TStream);
begin
  BeginUpdate;
  try
    inherited LoadFromStream(Stream);
    Stream.Read(fBoard, SizeOf(fBoard));
    Stream.Read(fBlackPieceCount, SizeOf(fBlackPieceCount));
    Stream.Read(fWhitePieceCount, SizeOf(fWhitePieceCount));
    DoChange;
  finally
    EndUpdate;
  end;
end;

function TOthelloBoard.ApplyLegalMove(Index: Integer): Boolean;
begin
  Result := False;
  if (Index >= 0) and (Index < LegalMoves.Count) then
    Result := ApplyMove(LegalMoves[Index]);
end;

function TOthelloBoard.ApplyMove(const Move: TMovement): Boolean;
begin
  if Move.NullMove then
    Result := ApplyNullMove
  else
    Result := ApplyMove(Move.Pos);
end;

function TOthelloBoard.ApplyMove(Pos: TPosition): Boolean;

  function DoFlip(const SideCells, EnemyCells: QWORD): Integer;
  var
    EnemyPos: TPosition;
    EnemyMask: QWORD;
    I: Integer;
  begin
    Result := 0;
    for I := Low(FlipInfo) to High(FlipInfo) do
      with FlipInfo[I] do
        if TopLeftSide then
        begin
          EnemyPos := Pos;
          EnemyMask := ((QWORD(1) shl Ord(Pos)) and Mask) shl Direction;
          while (EnemyCells and EnemyMask and Mask) <> 0 do
          begin
            EnemyMask := EnemyMask shl Direction;
            Inc(EnemyPos, Direction);
          end;
          if (EnemyPos <> Pos) and ((SideCells and EnemyMask) <> 0) then
            repeat
              EnemyMask := EnemyMask shr Direction;
              Pieces[EnemyPos] := Player;
              DoCapture(EnemyPos);
              Inc(Result);
              Dec(EnemyPos, Direction);
            until EnemyPos = Pos;
        end
        else
        begin
          EnemyPos := Pos;
          EnemyMask := ((QWORD(1) shl Ord(Pos)) and Mask) shr Direction;
          while (EnemyCells and EnemyMask and Mask) <> 0 do
          begin
            EnemyMask := EnemyMask shr Direction;
            Dec(EnemyPos, Direction);
          end;
          if (EnemyPos <> Pos) and ((SideCells and EnemyMask) <> 0) then
            repeat
              EnemyMask := EnemyMask shl Direction;
              Pieces[EnemyPos] := Player;
              DoCapture(EnemyPos);
              Inc(Result);
              Inc(EnemyPos, Direction);
            until EnemyPos = Pos;
        end;
  end;

begin
  Result := False;
  if LegalMoves.MoveExists(Pos) then
  begin
    DoMove(Pos);
    BeginUpdate;
    try
      case Player of
        BLACK: DoFlip(Board.Blacks, Board.Whites);
        WHITE: DoFlip(Board.Whites, Board.Blacks);
      end;
      Pieces[Pos] := Player;
      ChangeSide;
    finally
      EndUpdate;
    end;
    Result := True;
  end;
end;

function TOthelloBoard.ApplyNullMove: Boolean;
begin
  Result := False;
  if (LegalMoves.Count = 1) and LegalMoves[0].NullMove then
  begin
    DoNullMove;
    ChangeSide;
    Result := True;
  end;
end;

{ TOthello }

constructor TOthello.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  OthelloBoard.OnNullMove := PieceNullMove;
  OthelloBoard.OnMove := PieceMove;
  OthelloBoard.OnCapture := PieceCapture;
  fPiecesBitmap := TBitmap.Create;
  fPiecesBitmap.OnChange := PiecesBitmapChange;
  ComputerPlayer := WHITE;
  SquareColor := $00008000;
  SquareSelectColor := clLime;
  SquareMarkColor := clLime;
  AnimateMoves := False;
  AutoDragDrop := False;
  SelectionOptions := [soAutoPlayStep];
  BoardLabels := [blLeft, blTop];
  BoardLabelsOffset := 4;
  BoardMarginLeft := 20;
  BoardMarginTop := 20;
  BoardMarginRight := 20;
  BoardMarginBottom := 20;
end;

destructor TOthello.Destroy;
begin
  fPiecesBitmap.Free;
  inherited Destroy;
end;

function TOthello.GetOthelloBoard: TOthelloBoard;
begin
  Result := TOthelloBoard(Gameboard);
end;

function TOthello.GetBoard: TBoard;
begin
  Result := OthelloBoard.Board;
end;

procedure TOthello.SetBoard(const Value: TBoard);
begin
  OthelloBoard.Board := Value;
end;

function TOthello.GetPieces(Pos: TPosition): TPiece;
begin
  Result := OthelloBoard.Pieces[Pos];
end;

procedure TOthello.SetPieces(Pos: TPosition; Value: TPiece);
begin
  OthelloBoard.Pieces[Pos] := Value;
end;

function TOthello.GetLegalMoves: TMoveList;
begin
  Result := OthelloBoard.LegalMoves;
end;

procedure TOthello.SetPiecesBitmap(Value: TBitmap);
begin
  PiecesBitmap.Assign(Value);
end;

function TOthello.GetPosRect(Pos: TPosition): TRect;
begin
  Result := SquareBounds[PosToSquare(Pos)];
end;

function TOthello.GetSelectedMove: TMovement;
begin
  if not PathToMove(Selection, Result) then
    Result.NullMove := True;
end;

function TOthello.GetPositionScoreDisplay: TSquareLabel;
begin
  Result := SquareLabel;
end;

procedure TOthello.SetPositionScoreDisplay(Value: TSquareLabel);
begin
  SquareLabel := Value;
end;

function TOthello.GetCoordinatesDisplay: TBoardLabels;
begin
  Result := BoardLabels;
end;

procedure TOthello.SetCoordinatesDisplay(Value: TBoardLabels);
begin
  BoardLabels := Value;
end;

function TOthello.GetCoordinatesOffset: Integer;
begin
  Result := BoardLabelsOffset;
end;

procedure TOthello.SetCoordinatesOffset(Value: Integer);
begin
  BoardLabelsOffset := Value;
end;

function TOthello.GetSquareColor: TColor;
begin
  Result := SquareDarkColor;
end;

procedure TOthello.SetSquareColor(Value: TColor);
begin
  SquareDarkColor := Value;
end;

function TOthello.GetSquareMarkColor: TColor;
begin
  Result := SquareDarkMarkColor;
end;

procedure TOthello.SetSquareMarkColor(Value: TColor);
begin
  SquareDarkMarkColor := Value;
end;

function TOthello.GetSquareMarkPattern: TPicture;
begin
  Result := SquareDarkMarkPattern;
end;

procedure TOthello.SetSquareMarkPattern(Value: TPicture);
begin
  SquareDarkMarkPattern := Value;
end;

function TOthello.GetSquareSelectColor: TColor;
begin
  Result := SquareDarkSelectColor;
end;

procedure TOthello.SetSquareSelectColor(Value: TColor);
begin
  SquareDarkSelectColor := Value;
end;

function TOthello.GetSquareSelectPattern: TPicture;
begin
  Result := SquareDarkSelectPattern;
end;

procedure TOthello.SetSquareSelectPattern(Value: TPicture);
begin
  SquareDarkSelectPattern := Value;
end;

function TOthello.GetSquarePatern: TPicture;
begin
  Result := SquareDarkPattern;
end;

procedure TOthello.SetSquarePattern(Value: TPicture);
begin
  SquareDarkPattern := Value;
end;

function TOthello.GetPieceCount(Player: TPlayer): Byte;
begin
  case Player of
    BLACK: Result := OthelloBoard.BlackPieceCount;
    WHITE: Result := OthelloBoard.WhitePieceCount;
  else
    Result := 0;
  end;
end;

procedure TOthello.DoSquareResize;
begin
  if PiecesBitmap.Empty then
    UpdatePiecesImages;
  inherited DoSquareResize;
end;

procedure TOthello.DoSquareClick(Square: TSquare; var Handled: Boolean);
var
  Pos: TPosition;
begin
  if SquareToPos(Square, Pos) then
    DoPositionClick(Pos, Handled)
  else
    Handled := True;
  inherited DoSquareClick(Square, Handled);
end;

procedure TOthello.DoPositionClick(Pos: TPosition; var Handled: Boolean);
begin
  if Assigned(OnPositionClick) then
    OnPositionClick(Self, Pos, Handled);
end;

procedure TOthello.DoSquareInfoTip(Square: TSquare; var InfoTip: String);
var
  Pos: TPosition;
begin
  if SquareToPos(Square, Pos) then
    DoPositionInfoTip(Pos, InfoTip);
  inherited DoSquareInfoTip(Square, InfoTip);
end;

procedure TOthello.DoPositionInfoTip(Pos: TPosition; var InfoTip: String);
begin
  if Assigned(OnPositionInfoTip) then
    OnPositionInfoTip(Self, Pos, InfoTip);
end;

procedure TOthello.DoNullMove;
begin
  if Assigned(OnNullMove) then
    OnNullMove(Self);
end;

procedure TOthello.DoMove(Pos: TPosition);
begin
  if Assigned(OnMove) then
    OnMove(Self, Pos);
end;

procedure TOthello.DoCapture(Pos: TPosition);
begin
  if Assigned(OnCapture) then
    OnCapture(Self, Pos);
end;

procedure TOthello.PieceNullMove(Sender: TObject);
begin
  DoNullMove;
  StorePath([NO_SQUARE]);
end;

procedure TOthello.PieceMove(Sender: TObject; Pos: TPosition);
begin
  DoMove(Pos);
  StorePath([PosToSquare(Pos)]);
end;

procedure TOthello.PieceCapture(Sender: TObject; Pos: TPosition);
begin
  DoCapture(Pos);
end;

procedure TOthello.PiecesBitmapChange(Sender: TObject);
begin
  UpdatePiecesImages;
end;

function TOthello.GetGameboardClass: TGameboardClass;
begin
  Result := TOthelloBoard;
end;

procedure TOthello.SetImageIndexOfSquare(Square: TSquare; ImageIndex: Integer);
var
  Pos: TPosition;
begin
  if SquareToPos(Square, Pos) then
    case ImageIndex of
      0: Pieces[Pos] := BLACK;
      1: Pieces[Pos] := WHITE;
    else
      Pieces[Pos] := EMPTY;
    end;
end;

function TOthello.GetImageIndexOfSquare(Square: TSquare): Integer;
var
  Pos: TPosition;
begin
  Result := -1;
  if SquareToPos(Square, Pos) then
    case Pieces[Pos] of
      BLACK: Result := 0;
      WHITE: Result := 1;
    end;
end;

function TOthello.GetSquareLabel(Square: TSquare): String;
var
  Move: TMovement;
  MoveIndex: Integer;
  Score: Single;
begin
  Result := '';
  Move.NullMove := not SquareToPos(Square, Move.Pos);
  MoveIndex := LegalMoves.IndexOf(Move);
  if MoveIndex >= 0 then
  begin
    Score := MoveScore(MoveIndex);
    Result := FormatFloat('+.00;-.00;.00', Score);
  end;
end;

function TOthello.GetStateOfSquare(Square: TSquare): TSquareDrawStates;
begin
  Result := inherited GetStateOfSquare(Square);
  if SquareLabel <> slNone then Include(Result, sdsNoCache);
end;

function TOthello.IsPathComplete(const Path: array of TSquare): Boolean;
var
  Move: TMovement;
begin
  Result := False;
  if PathToMove(Path, Move) then
    Result := LegalMoves.MoveExists(Move);
end;

function TOthello.CanPlayPath(const Path: array of TSquare): Boolean;
begin
  Result := IsPathComplete(Path);
end;

function TOthello.PlayPath(const Path: array of TSquare): Boolean;
var
  Move: TMovement;
begin
  Result := False;
  if PathToMove(Path, Move) then
    Result := Play(Move);
end;

function TOthello.CompletePath(var Path: TMovementPath): Boolean;
begin
  Result := IsPathComplete(Path);
end;

function TOthello.IsVisualDarkSquare(Square: TSquare): Boolean;
begin
  Result := True;
end;

function TOthello.CanSelectSquare(Square: TSquare): Boolean;
var
  Pos: TPosition;
begin
  Result := inherited CanSelectSquare(Square) and SquareToPos(Square, Pos) and
            LegalMoves.MoveExists(Pos);
end;

function TOthello.IsLegalMove(const Move: TMovement): Boolean;
begin
  Result := LegalMoves.MoveExists(Move)
end;

function TOthello.Play(const Move: TMovement): Boolean;
begin
  Result := False;
  if IsLegalMove(Move) then
  begin
    InitializeMovement;
    try
      Result := OthelloBoard.ApplyMove(Move);
    finally
      FinalizeMovement;
    end;
  end;
end;

function TOthello.PlayNullMove: Boolean;
var
  Move: TMovement;
begin
  Move.NullMove := True;
  Result := Play(Move);
end;

function TOthello.MustPlayNull: Boolean;
begin
  Result := (LegalMoves.Count = 1) and LegalMoves[0].NullMove;
end;

function TOthello.MoveOf(Player: TPlayer; TurnNo: Integer): String;
var
  HistoryRec: TGameHistoryRec;
  Move: TMovement;
  Index: Integer;
begin
  Result := '';
  HistoryRec := HistoryOf(Player, TurnNo);
  if Assigned(HistoryRec) then
  begin
    Index := History.IndexOf(HistoryRec);
    if Index < 0 then
    begin
      PathToMove(HistoryRec.MovementPath, Move);
      Result := Result + MoveToStr(Move);
      if not HistoryRec.Completed then
        Result := Result + '->';
      Index := History.Count - 1;
      while (Index >= 0) and not History[Index].Completed do
      begin
        PathToMove(History[Index].MovementPath, Move);
        Result := MoveToStr(Move) + ',' + Result;
        Dec(Index);
      end;
    end
    else
    begin
      while (Index < History.Count - 1) and not HistoryRec.Completed do
      begin
        PathToMove(HistoryRec.MovementPath, Move);
        Result := Result + MoveToStr(Move) + ',';
        Inc(Index);
        HistoryRec := History[Index];
      end;
      PathToMove(HistoryRec.MovementPath, Move);
      Result := Result + MoveToStr(Move);
      if not HistoryRec.Completed then
        Result := Result + '->';
    end;
  end;
end;

function TOthello.IsPositionSelected(Pos: TPosition): Boolean;
begin
  Result := IsSquareSelected(PosToSquare(Pos));
end;

function TOthello.CanSelectPosition(Pos: TPosition): Boolean;
begin
  Result := CanSelectSquare(PosToSquare(Pos));
end;

function TOthello.SelectPosition(Pos: TPosition): Boolean;
begin
  Result := SelectSquare(PosToSquare(Pos));
end;

function TOthello.SelectMove(Index: Integer): Boolean;
begin
  Result := False;
  if (Index >= 0) and (Index < LegalMoves.Count) then
    Result := SelectMove(LegalMoves[Index]);
end;

function TOthello.SelectMove(const Move: TMovement): Boolean;
begin
  Result := False;
  if IsLegalMove(Move) then
  begin
    Selection := MoveToPath(Move);
    Result := True;
  end;
end;

function TOthello.UpdatePiecesImages: Boolean;
var
  DefaultSize: Integer;
  BitmapOfPieces: TBitmap;
begin
  Result := False;
  if not Assigned(PiecesBitmap) or PiecesBitmap.Empty then
  begin
    case SquareSize of
      00..27: DefaultSize := 16;
      28..35: DefaultSize := 24;
      36..43: DefaultSize := 32;
      44..51: DefaultSize := 40;
    else
      DefaultSize := 48;
    end;
    if DefaultSize <> DefaultPieceSize then
    begin
      BitmapOfPieces := TBitmap.Create;
      try
        BitmapOfPieces.LoadFromResourceName(HInstance, 'OTHELLO_PIECES' + IntToStr(DefaultSize));
        ExtractPiecesImages(BitmapOfPieces, 1, 2);
      finally
        BitmapOfPieces.Free;
      end;
      DefaultPieceSize := DefaultSize;
      Result := True;
    end;
  end
  else
  begin
    DefaultPieceSize := 0;
    PiecesBitmap.OnChange := nil;
    try
      ExtractPiecesImages(PiecesBitmap, 1, 2)
    finally
      PiecesBitmap.OnChange := PiecesBitmapChange;
    end;
    Result := True;
  end;
end;

function TOthello.ClientToPos(X, Y: Integer; out Pos: TPosition): Boolean;
begin
  Result := SquareToPos(ClientToSquare(X, Y), Pos);
end;

function TOthello.SquareToPos(Square: TSquare; out Pos: TPosition): Boolean;
var
  Row, Col: Integer;
begin
  Result := SquareToCoordinate(Square, Row, Col) and CoordinateToPos(Row, Col, Pos);
end;

function TOthello.PosToSquare(Pos: TPosition): TSquare;
var
  Row, Col: Integer;
begin
  PosToCoordinate(Pos, Row, Col);
  Result := CoordinateToSquare(Row, Col);
end;

function TOthello.MoveToPath(const Move: TMovement): TMovementPath;
begin
  SetLength(Result, 1);
  if Move.NullMove then
    Result[0] := NO_SQUARE
  else
    Result[0] := PosToSquare(Move.Pos);
end;

function TOthello.PathToMove(const Path: array of TSquare; out Move: TMovement): Boolean;
begin
  Result := False;
  if Length(Path) <= 1 then
  begin
    Move.NullMove := (Length(Path) = 0) or not SquareToPos(Path[0], Move.Pos);
    Result := True;
  end;
end;

initialization
  InitZobritsNumbers(Zobrist, SizeOf(Zobrist) div SizeOf(TPositionKey));
  PositionTable := TPositionTable.Create;
finalization
  FreeAndNil(PositionTable);
end.
